/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

/* ao_status() copied from LAGII, (C) 1999 XoXus 
 * modified my me to allow for selection of items (FLAG[13]) */

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "general.h"
#include "main.h"
#include "driver.h"
#include "graphics.h"
#include "event.h"
#include "menu.h"
#include "resource.h"
#include "picture.h"
#include "gameid.h"
#include "text.h"
#include "vm.h"
#include "status.h"
#include "object.h"
#include "view.h"
#include "check.h"
#include "save.h"
#include "message.h"
#include "sound.h"

/* AGI INSTRUCTIONS THAT MANIPULATE VARIABLES */
void ao_increment(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	if(VAR[arg] != 255) VAR[arg]++;
}

void ao_decrement(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	if(VAR[arg] != 0) VAR[arg]--;
}

void ao_assignn(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] = arg1;
}

void ao_assignv(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] = VAR[arg1];
}

void ao_addn(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] += arg1;
}

void ao_addv(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] += VAR[arg1];
}

void ao_subn(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] -= arg1;
}

void ao_subv(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] -= VAR[arg1];
}

void ao_lindirectv(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[VAR[arg0]] = VAR[arg1];
}

void ao_lindirectn(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[VAR[arg0]] = arg1;
}

void ao_rindirect(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] = VAR[VAR[arg1]];
}

void ao_muln(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] *= arg1;
}

void ao_mulv(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] *= VAR[arg1];
}

void ao_divn(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] /= arg1;
}

void ao_divv(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg0] /= VAR[arg1];
}

void ao_random(vm_t *l)
{
	UINT8 arg0, arg1, arg2;
	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	/* VAR[arg2] = random number from arg0 to arg1 */
	VAR[arg2] = arg0 + (rand() % (arg1 - arg0 + 1));
}

/* AGI INSTRUCTIONS THAT MANIPULATE FLAGS */
void ao_set(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	FLAG[arg] = true;
}

void ao_setv(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	FLAG[VAR[arg]] = true;
}

void ao_reset(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	FLAG[arg] = false;
}

void ao_resetv(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	FLAG[VAR[arg]] = false;
}

void ao_toggle(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	FLAG[arg] = !FLAG[arg];
}

void ao_togglev(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	FLAG[VAR[arg]] = !FLAG[VAR[arg]];
}

/* AGI COMMANDS TO LOAD AND UNLOAD RESOURCES */
void ao_load_logic(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	res_load(LOGIC[arg]);
}

void ao_load_logic_v(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	res_load(LOGIC[VAR[arg]]);
}

void ao_load_pic(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	res_load(PICTURE[VAR[arg]]);
}

void ao_load_view(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	res_load(VIEW[arg]);
}

void ao_load_view_v(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	res_load(VIEW[VAR[arg]]);
}

void ao_load_sound(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	res_load(SOUND[arg]);
}

void ao_discard_pic(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	res_unload(PICTURE[VAR[arg]]);
}

void ao_discard_view(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	res_unload(VIEW[arg]);
}

void ao_discard_view_v(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	res_unload(VIEW[VAR[arg]]);
}

/* AGI PROGRAM CONTROL COMMANDS */
void ao_new_room(vm_t *l)
{
	UINT8 arg;
	int i;
	
	arg = vm_pop();

	for(i = 0; i < 256; i++) {
		if(i && LOGIC[i] && LOGIC[i]->data) res_unload(LOGIC[i]);
		
		if(PICTURE[i] && PICTURE[i]->data) res_unload(PICTURE[i]);
		if(VIEW[i] && VIEW[i]->data) res_unload(VIEW[i]);
		if(SOUND[i] && SOUND[i]->data) res_unload(SOUND[i]);
		
/*		object[i].flags &= ~VO_ON_SCREEN;*/
	}

	/* stop.update(all) : handled by unanimate */
	ao_stop_sound(l);
	ao_unanimate_all(l);
	ao_player_control(l);

	/* this should probably not be here */
	ao_accept_input(l);
	
	/* unblock() */
	vm_push(36);
	ao_set_horizon(l);
	VAR[1] = VAR[0]; /* old_room = current_room */
	VAR[0] = arg;    /* room = new_room */
	VAR[4] = VAR[5] = 0; /* no objects touched border */
	VAR[16] = object[0].view; /* ego_view */

	res_load(LOGIC[arg]);

	/* Adjust EGO's coords */
	switch(VAR[2]) { /* border touched by ego */
		case 1:	/* Touched horizon */
			object[0].y = 167;
			break;
		case 2: /* Touched right */
			object[0].x = 1;
			break;
		case 3: /* Touched bottom */
			object[0].y = 37; /* default horizon + 1 */
			break;
		case 4: /* Touched left */
			object[0].x = 160 - object[0].width;
			break;
	}

	VAR[2] = 0; /* ego touched no border */
	FLAG[0] = 0; /* not on water */
	FLAG[3] = 0; /* not on trip wire */

	FLAG[5] = 1; /* new_room */
	FLAG[2] = 0; /* player has not issued command line */
	FLAG[4] = 0;

	driver_kb_update();

	new_room = true;
	vm_reset = true; /* exit all logics now! */
}

void ao_new_room_v(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	vm_push(VAR[arg]);
	ao_new_room(l); /* redirect to new_room(); */
}

void ao_call(vm_t *l)
{
	UINT8 arg;
	boolean weloaded = false;

	arg = vm_pop();

	if(!LOGIC[arg]->data) {
		weloaded = true;
		res_load(LOGIC[arg]);
	}

	vm_exec(arg);

	if(weloaded) res_unload(LOGIC[arg]);
}

void ao_call_v(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	vm_push(VAR[arg]);
	ao_call(l);
}

void ao_set_scan_start(vm_t *l)
{
	l->res->ssIP = l->IP;
}

void ao_reset_scan_start(vm_t *l)
{
	l->res->ssIP = 0;
}

/* AGI OBJECT CONTROL COMMANDS */
void ao_animate_obj(vm_t *l)
{
	UINT8 arg;
	view_obj * o;
	
	arg = vm_pop();
	o = &object[arg];
	
	/* set these */
	o->flags |= (VO_ANIMATE | VO_UPDATE | VO_CYCLE | VO_BLOCK | VO_MOTION);

	/* unset these */
	o->flags &= ~(VO_DRAWN | VO_REVERSE | VO_FIX_PRI | VO_FIX_LOOP);
	o->flags &= ~(VO_AUTO_MOVE | VO_WANDER | VO_FOLLOW_EGO);
	o->flags &= ~(VO_NOLAND | VO_NOWATER | VO_OBJECTS);
	
	if(o != &object[0]) o->direction = 0;
	
	o->cycle_time = 1; o->cycle_count = 1;
	o->step_time = 1; o->step_count = 1;
	o->step_size = 1;
	o->do_once_flag = 0;
}

void ao_unanimate_all(vm_t *l)
{
	int i;
	for(i = 0; i < 256; i++) {
		object[i].flags |= (VO_HORIZON | VO_BLOCK);
		object[i].flags &= ~(VO_ANIMATE | VO_DRAWN);
	}
}

void ao_set_view(vm_t *l)
{
	UINT8 arg0, arg1;
	view_obj *o;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	o = &object[arg0];
	o->view = arg1;
/*	o->loop = 0;
	o->cel = 0;*/
	
	view_recalc(o);
}

void ao_set_view_v(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();

	vm_push(arg0);
	vm_push(VAR[arg1]);
	
	ao_set_view(l);	/* redirect */
}

void ao_set_loop(vm_t *l)
{
	UINT8 arg0, arg1;
	view_obj *o;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	o = &object[arg0];
	o->loop = arg1;
/*	o->cel = 0;*/

	view_recalc(o);
}

void ao_set_loop_v(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();

	vm_push(arg0);
	vm_push(VAR[arg1]);
	
	ao_set_loop(l);	/* redirect */
}

void ao_fix_loop(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].flags |= VO_FIX_LOOP;
}

void ao_release_loop(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].flags &= ~VO_FIX_LOOP;
}

void ao_set_cel(vm_t *l)
{
	UINT8 arg0, arg1;
	view_obj *o;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	o = &object[arg0];
	o->cel = arg1;
	
	view_recalc(o);
}

void ao_set_cel_v(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();

	vm_push(arg0);
	vm_push(VAR[arg1]);
	
	ao_set_cel(l);	/* redirect */
}

void ao_last_cel(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	VAR[arg1] = object[arg0].nocels - 1;
}

void ao_number_of_loops(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	VAR[arg1] = object[arg0].noloops;
}

void ao_current_cel(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	VAR[arg1] = object[arg0].cel;
}

void ao_current_loop(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	VAR[arg1] = object[arg0].loop;
}

void ao_current_view(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	VAR[arg1] = object[arg0].view;
}

void ao_set_priority(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	if(arg1 < 4) arg1 = 4;
	object[arg0].priority = arg1;
	object[arg0].flags |= VO_FIX_PRI;
}

void ao_set_priority_v(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	object[arg0].priority = VAR[arg1];
	if(object[arg0].priority < 4) object[arg0].priority = 4;
	object[arg0].flags |= VO_FIX_PRI;
}

void ao_release_priority(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();

	object[arg].flags &= ~VO_FIX_PRI;
}

void ao_get_priority(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	VAR[arg1] = object[arg0].priority;
}

void ao_position(vm_t *l)
{
	UINT8 arg0, arg1, arg2;

	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	object[arg0].x = arg1;
	object[arg0].y = arg2;
/*	check_fix_position(&object[arg0]);*/
}

void ao_position_v(vm_t *l)
{
	UINT8 arg0, arg1, arg2;

	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	object[arg0].x = VAR[arg1];
	object[arg0].y = VAR[arg2];
/*	check_fix_position(&object[arg0]);*/
}

void ao_draw(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();

	object[arg].flags |= VO_DRAWN | VO_UPDATE;
	check_fix_position(&object[arg]);
}

void ao_erase(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();

/*	view_erase(&object[arg]);*/
	object[arg].flags &= ~(VO_DRAWN | VO_UPDATE);
}

void ao_get_posn(vm_t *l)
{
	UINT8 arg0, arg1, arg2;

	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	VAR[arg1] = object[arg0].x;
	VAR[arg2] = object[arg0].y;
}

void ao_start_cycling(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();

	object[arg].flags |= VO_CYCLE | VO_UPDATE;
}

void ao_stop_cycling(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();

	object[arg].flags &= ~VO_CYCLE;
}

void ao_normal_cycle(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();

	object[arg].flags |= VO_CYCLE | VO_UPDATE;
	object[arg].flags &= ~VO_REVERSE;
}

void ao_reverse_cycle(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();

	object[arg].flags |= VO_CYCLE | VO_REVERSE | VO_UPDATE;
}

void ao_end_of_loop(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	object[arg0].flags |= VO_CYCLE | VO_UPDATE;
	object[arg0].flags &= ~VO_REVERSE;
	object[arg0].do_once_flag = arg1;

	/* I'm sure this can't be right, but it fixes the dumpster sequence
	 * in Larry */
	object[arg0].cycle_time = 1;	
	object[arg0].cycle_count = 1;	

	FLAG[arg1] = false;
}

void ao_reverse_loop(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	object[arg0].flags |= VO_CYCLE | VO_UPDATE | VO_REVERSE;
	object[arg0].do_once_flag = arg1;
	FLAG[arg1] = false;
}

void ao_cycle_time(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	object[arg0].cycle_time = VAR[arg1];
	object[arg0].cycle_count = VAR[arg1];
}

/* AGI OBJECT MOTION CONTROL COMMANDS */
void ao_set_horizon(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	status.horizon = arg;
}

void ao_ignore_horizon(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].flags &= ~VO_HORIZON;
}

void ao_observe_horizon(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].flags |= VO_HORIZON;
}

void ao_ignore_blocks(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].flags &= ~VO_BLOCK;
}

void ao_observe_blocks(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].flags |= VO_BLOCK;
}

void ao_ignore_objs(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].flags &= ~VO_OBJECTS;
}

void ao_observe_objs(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].flags |= VO_OBJECTS;
}

void ao_player_control(vm_t *l)
{
	status.program_control = false;
}

void ao_program_control(vm_t *l)
{
	status.program_control = true;
	object[0].flags &= ~VO_MOTION;
	object[0].direction = 0;
	VAR[6] = 0;
}

void ao_stop_motion(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].direction = 0;
	object[arg].flags &= ~(VO_MOTION | VO_AUTO_MOVE | VO_WANDER | VO_FOLLOW_EGO);
	if(arg == 0) VAR[6] = 0;/*ao_program_control(l);*/
}

void ao_start_motion(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].flags |= VO_MOTION;
/*	if(arg == 0) ao_player_control(l);*/
}

void ao_step_size(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	object[arg0].step_size = VAR[arg1];
}

void ao_step_time(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	object[arg0].step_time = VAR[arg1];
	object[arg0].step_count = VAR[arg1];
}

void ao_move_obj(vm_t *l)
{
	UINT8 arg0, arg1, arg2, arg3, arg4;
	view_obj *o;
	
	arg4 = vm_pop();
	arg3 = vm_pop();
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	o = &object[arg0];
	o->flags |= (VO_AUTO_MOVE | VO_UPDATE | VO_MOTION);
	o->flags &= ~(VO_FOLLOW_EGO | VO_WANDER | VO_NOT_MOVED);
	o->tx = arg1;
	o->ty = arg2;
	if(arg3 != 0) o->step_size = arg3;
/*	else o->step_size = 1;*/
	o->target_flag = arg4;
	FLAG[arg4] = false;
	if(arg0 == 0) ao_program_control(l);
}

void ao_move_obj_v(vm_t *l)
{
	UINT8 arg0, arg1, arg2, arg3, arg4;
	
	arg4 = vm_pop();
	arg3 = vm_pop();
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	vm_push(arg0);
	vm_push(VAR[arg1]);
	vm_push(VAR[arg2]);
	vm_push(VAR[arg3]);
	vm_push(arg4);
	
	ao_move_obj(l);
}

void ao_follow_ego(vm_t *l)
{
	UINT8 arg0, arg1, arg2;
	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	object[arg0].flags |= (VO_FOLLOW_EGO | VO_AUTO_MOVE | VO_UPDATE | VO_MOTION);
	object[arg0].flags &= ~VO_WANDER;
	if(VAR[arg1] != 0) object[arg0].step_size = VAR[arg1];
	object[arg0].target_flag = arg2;
	FLAG[arg2] = false;
}


void ao_wander(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	
	object[arg].flags |= (VO_AUTO_MOVE | VO_WANDER | VO_UPDATE | VO_MOTION);
	object[arg].flags &= ~VO_FOLLOW_EGO;
}

void ao_normal_motion(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	
	object[arg].flags &= ~(VO_FOLLOW_EGO | VO_AUTO_MOVE | VO_WANDER);
}

void ao_set_dir(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();

	object[arg0].direction = VAR[arg1] % 9;
	if(arg0 == 0) VAR[6] = object[0].direction;
}

void ao_get_dir(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	VAR[arg1] = object[arg0].direction;
}

void ao_object_on_land(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	
	object[arg].flags |= VO_NOWATER;
	object[arg].flags &= ~VO_NOLAND;
	check_fix_position(&object[arg]);
}

void ao_object_on_water(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	
	object[arg].flags |= VO_NOLAND;
	object[arg].flags &= ~VO_NOWATER;
	check_fix_position(&object[arg]);
}

void ao_object_on_anything(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	
	object[arg].flags &= ~(VO_NOLAND | VO_NOWATER);
}

void ao_reposition(vm_t *l)
{
	UINT8 arg0, arg1, arg2;
	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	object[arg0].x += VAR[arg1];
	object[arg0].y += VAR[arg2];
	check_fix_position(&object[arg0]);
}

void ao_stop_update(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	if(!object[arg].do_once_flag)
		object[arg].flags &= ~VO_UPDATE;
}

void ao_start_update(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();
	
	object[arg].flags |= VO_UPDATE;
}

void ao_distance(vm_t *l)
{
	UINT8 arg0, arg1, arg2;
	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	if((~object[arg0].flags & (VO_ANIMATE | VO_DRAWN)) ||
	   (~object[arg1].flags & (VO_ANIMATE | VO_DRAWN))) {
	   	VAR[arg2] = 255;
		return;
	}

	VAR[arg2] = abs(object[arg0].x - object[arg1].x) +
		    abs(object[arg0].y - object[arg1].y);
}

/* AGI INVENTORY ITEM MANAGEMENT COMMANDS */
void ao_get(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	
	inv_item[arg].room = 255;
}

void ao_get_v(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	
	inv_item[VAR[arg]].room = 255;
}

void ao_drop(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	
	inv_item[arg].room = 0;
}

void ao_put(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();	

	inv_item[arg0].room = VAR[arg1];
}

void ao_put_v(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();	

	inv_item[VAR[arg0]].room = VAR[arg1];
}

void ao_get_room_v(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();	

	VAR[arg1] = inv_item[VAR[arg0]].room;
}

void ao_status(vm_t *l) /* Copied from LAGII, (C) 1999 XoXus */
{
	int i, num;
	struct inv_menu_item_s {
		inv_item_t * item;
		int itemno;
		int x;
		int y;
	} inv_menu_item[50];
	
	struct inv_menu_item_s * c = NULL;
	
	/* Clear screen */
	memcpy (textmodebuffer, video_buffer, 64000);
	memset (video_buffer, 15, 64000);

	graph_drawstring(video_buffer, 92, 0, "You are carrying:", 0, 15);

	/* Print items and build inv_menu_item array */
	num = 0;
	for (i = 0; i < 256; i++) {
		if(inv_item[i].room != 0xFF) continue;
		if(inv_item[i].name[0] == 0) continue;
		num++;

		c = &inv_menu_item[num];
		c->itemno = i;
		c->item = &inv_item[i];
		c->y = 9 * ((num + 3) >> 1);
		if(num % 2) c->x = 8;
		else c->x = 312 - 8 * strlen (c->item->name);

		graph_drawstring (video_buffer, c->x, c->y, c->item->name, 0, 15);
	}

	if((num != 0) && FLAG[13]) { /* Select item enabled, and have items */
		int k;
		i = 1;
		c = &inv_menu_item[i]; /* Draw cursor in first position */
		graph_drawstring (video_buffer, c->x, c->y, c->item->name, 15, 0);

		graph_drawstring (video_buffer, 16, 192,
			"Press ENTER to select, ESC to cancel", 0, 15);
			
		while(1) {
			driver_showbuffer(video_buffer);
			k = event_waitforkey();
				/* ESC or ENTER pressed */
			if((k == 0x0D) || (k == '\n') || (k == 27)) break;
				/* Erase old cursor */			
			graph_drawstring (video_buffer, c->x, c->y, c->item->name, 0, 15);

			switch(k) {	/* Move cursor */
				case AGIL_KEY_CURSORRIGHT:
					if(i % 2) i++;
					break;
				case AGIL_KEY_CURSORLEFT:
					if(!(i % 2)) i--;
					break;
				case AGIL_KEY_CURSORUP:
					i -= 2;
					break;
				case AGIL_KEY_CURSORDOWN:
					i += 2;
					break;
			}

			if(i < 1) i = 1;
			if(i > num) i = num;
			c = &inv_menu_item[i]; /* draw cursor in new pos */
			graph_drawstring (video_buffer, c->x, c->y, c->item->name, 15, 0);
		}
		if(k == 27) {
			VAR[25] = 0xFF; /* canceled */
			FLAG[13] = false;
		} else VAR[25] = c->itemno;		
	} else { /* select item disabled or have no items */
		if (num == 0) 
			graph_drawstring (video_buffer, 132, 16, "nothing", 0, 15);
		
		graph_drawstring (video_buffer, 28, 192,
			"Press a key to return to the game", 0, 15);

		/* Show inventory screen & wait for key */
		driver_showbuffer (video_buffer);
		event_waitforkey ();

		if(FLAG[13]) {
			FLAG[13] = false;
			VAR[25] = 0xFF; /* no item selected, canceled */
		}
	}
	/* Restore screen contents */
	memcpy (video_buffer, textmodebuffer, 64000);
	driver_showbuffer (video_buffer);
}


/* AGI PICTURE RESOURCE MANAGEMENT COMMANDS */
void ao_draw_pic(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	if(PICTURE[VAR[arg]]) 
		pic_draw(PICTURE[VAR[arg]], true);
}

void ao_overlay_pic(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	if(PICTURE[VAR[arg]])
		pic_draw(PICTURE[VAR[arg]], false);
}

void ao_show_pic(vm_t *l)
{
	int i;
	for(i = 0; i < 256; i++) object[i].flags &= ~VO_ON_SCREEN;
	memcpy(workpribuffer + (320 * 8), backpribuffer, 320 * 168);
	memcpy(workvisbuffer + (320 * 8), backvisbuffer, 320 * 168);
}

void ao_add_to_pic(vm_t *l)
{
	UINT8 a, b, c, d, e, f, g;

	g = vm_pop();	
	f = vm_pop();
	e = vm_pop();
	d = vm_pop();
	c = vm_pop();
	b = vm_pop();
	a = vm_pop();

	view_add_to_pic(a, b, c, d, e, f, g);
}

void ao_add_to_pic_v(vm_t *l)
{
	UINT8 a, b, c, d, e, f, g;

	g = vm_pop();	
	f = vm_pop();
	e = vm_pop();
	d = vm_pop();
	c = vm_pop();
	b = vm_pop();
	a = vm_pop();

	view_add_to_pic(VAR[a], VAR[b], VAR[c], VAR[d], VAR[e], VAR[f], VAR[g]);
}

/* AGI SOUND RESOURCE MANAGEMENT COMMANDS */
/* NOT YET IMPLEMENTED */
/* Preliminary sound driver (LollypopMan 21-12-2001) */

void ao_sound(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();

	FLAG[arg1] = false;
	sound_play(SOUND[arg0]->data, arg1);
}

void ao_stop_sound(vm_t *l)
{
/*	driver_stop_sound();*/
	sound_stop();
}

/* AGI TEXT MANAGEMENT COMMANDS */

void ao_prevent_input(vm_t *l)
{
	status.accept_input = false;
	memset(video_buffer + 56320, 0, 2560); 
}

void ao_accept_input(vm_t *l)
{
	status.accept_input = true;
	text_new_key(0);	/* Redraw input line */
}

void ao_print(vm_t *l) 
{
	UINT8 arg;
	int i;
	
	arg = vm_pop();

	message_erasebox();

	if(!FLAG[15]) {
		memcpy(textmodebuffer, video_buffer, 64000); /* save screen contents */
		message_mbox(vm_parsemessage(l->res, arg), 0, 0, 0);
		driver_showbuffer(video_buffer);
		while(event_waitforkey() != 0x0D);
		message_erasebox();
		memcpy(video_buffer, textmodebuffer, 64000); /* restore screen contents */
	} else {
		message_mbox(vm_parsemessage(l->res, arg), 0, 0, 0);
		driver_showbuffer(video_buffer);
		FLAG[15] = false;
	}
	/* Flush Keytable */
	for(i = 0; i < 128; i++) 
		keytable[i] = 0; 
}

void ao_print_v(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	vm_push(VAR[arg]);
	ao_print(l);	/* redirect to print() */
}

void ao_display(vm_t * l)
{
	UINT8 arg0, arg1, arg2;
	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();
	graph_drawstring(video_buffer, arg1 * 8, arg0 * 8, vm_parsemessage(l->res, arg2), status.fgcolor, status.bgcolor);
	driver_showbuffer(video_buffer);
}

void ao_display_v(vm_t * l)
{
	UINT8 arg0, arg1, arg2;
	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();
	vm_push(VAR[arg0]);
	vm_push(VAR[arg1]);
	vm_push(VAR[arg2]);
	ao_display(l);	/* redirect to display() */
}

void ao_print_at(vm_t *l)
{
	UINT8 arg0, arg1, arg2, arg3;
	int i;
	
	if(action_opcode[0x97].args == 4) arg3 = vm_pop(); else arg3 = 0;
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	message_erasebox();
	
	if(!FLAG[15]) {
		memcpy(textmodebuffer, video_buffer, 64000); /* save screen contents */
		message_mbox(vm_parsemessage(l->res, arg0), (arg2 - 1) * 8, arg1 * 8, arg3);
		driver_showbuffer(video_buffer);
		while(event_waitforkey() != 0x0D);
		message_erasebox();
		memcpy(video_buffer, textmodebuffer, 64000); /* restore screen contents */
	} else {
		message_mbox(vm_parsemessage(l->res, arg0), (arg2 - 1) * 8, arg1 * 8, arg3);
		driver_showbuffer(workvisbuffer);
		FLAG[15] = false;
	}
	/* Flush Keytable */
	for(i = 0; i < 128; i++) 
		keytable[i] = 0; 
}

void ao_print_at_v(vm_t *l)
{
	UINT8 arg0, arg1, arg2, arg3;

	if(action_opcode[0x98].args == 4) arg3 = vm_pop(); else arg3 = 0;
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	vm_push(VAR[arg0]);
	vm_push(arg1);
	vm_push(arg2);
	if(action_opcode[0x97].args == 4) vm_push(arg3);
	
	ao_print_at(l);	/* redirect to print() */
}

void ao_version(vm_t *l) 
{
	char msg[100];
	
	sprintf(msg, "        %s %s\n (C) 2001, 2002 RaDIaT1oN\n\nAdventure Game Interpreter\n      Version %s", TITLE, VERSION, id->agiver);
	memcpy(textmodebuffer, video_buffer, 64000); /* save screen contents */
	message_mbox(msg, 0, 0, 0);
	driver_showbuffer(video_buffer);
	event_waitforkey();
	message_erasebox();
	memcpy(video_buffer, textmodebuffer, 64000); /* restore screen contents */
}

void ao_text_screen(vm_t *l)
{
	memset(textmodebuffer, status.bgcolor, 64000);
	video_buffer = textmodebuffer;
	driver_showbuffer(video_buffer);
}

void ao_graphics(vm_t *l)
{
	video_buffer = workvisbuffer;
	if(status.show_status_line)
		update_status();

	text_new_key(0);
}

void ao_set_cursor_char(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	
	status.cursor_char = l->res->message[arg - 1][0];
}

void ao_set_text_attribute(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	status.fgcolor = arg0;
	status.bgcolor = arg1;
}

void ao_clear_lines (vm_t *l)
{
	UINT8 arg0, arg1, arg2;
	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	memset(video_buffer + (2560 * arg0), arg2, 2560 * (arg1 - arg0 + 1));
}

void ao_clear_text_rect(vm_t *l)
{
	UINT8 arg0, arg1, arg2, arg3, arg4;
	int x, y, w, h;
	
	arg4 = vm_pop();
	arg3 = vm_pop();
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	x = (int)arg1 * 8;
	y = (int)arg0 * 8;
	w = (int)(arg3 - arg1 + 1) * 8;
	h = (int)(arg2 - arg0 + 1) * 8;

	graph_drawbox(video_buffer, x, y, w, h, arg4, arg4);
}

void ao_status_line_on(vm_t *l)
{
	status.show_status_line = true;
	update_status();
}

void ao_status_line_off(vm_t *l)
{
	status.show_status_line = false;
	memset(video_buffer, 0, 2560);
}


/* AGI STRING MANAGEMENT COMMANDS */
void ao_set_string(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();	

	strcpy(STRING[arg0], vm_parsemessage(l->res, arg1));
}

void ao_word_to_string(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();	

	strcpy(STRING[arg0], parsedwords[arg1 - 1]);
}

void ao_get_string(vm_t *l)
{
	UINT8 arg0, arg1, arg2, arg3, arg4;
	
	arg4 = vm_pop();
	arg3 = vm_pop();
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	text_input(vm_parsemessage(l->res, arg1), arg3, arg2, arg4, "");
	
	strcpy(STRING[arg0], inputstring);
	input_pos = 0; inputstring[input_pos] = 0;
}

void ao_parse(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();

	text_parse(STRING[arg]);
}

void ao_get_num(vm_t *l)
{
	UINT8 arg0, arg1;
	int val;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	text_input(vm_parsemessage(l->res, arg0), 0, 23, 3, "");
	
	if(sscanf(inputstring, "%d", &val) && (val >= 0) && (val <= 255))
		VAR[arg1] = val;
	else
		VAR[arg1] = 0;

	input_pos = 0; inputstring[input_pos] = 0;
}

/* AGI INITIALIZATION COMMANDS */
void ao_set_key(vm_t *l)
{
	UINT8 arg0, arg1, arg2;
	int scancode;
	
	arg2 = vm_pop();
	arg1 = vm_pop();	
	arg0 = vm_pop();
	
	scancode = ((UINT16)arg1 << 8) + arg0;
	event_addkey(scancode, arg2);
}

void ao_set_menu(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	menu_add(l->res->message[arg - 1]);
}

void ao_set_menu_item(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	menu_additem(l->res->message[arg0 - 1], arg1);
}

void ao_submit_menu(vm_t *l)
{
	menu_submitted = true;
}

void ao_enable_item(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	menu_enable(arg);
}

void ao_disable_item(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	menu_disable(arg);
}

void ao_menu_input(vm_t *l)
{
	menu_input();
}

/* OTHER AGI COMMANDS */
void ao_show_pri_screen(vm_t *l)
{
	driver_showbuffer(workpribuffer);
	event_waitforkey();
	driver_showbuffer(video_buffer);
}

void ao_show_obj(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	view_show_obj(arg);
}

void ao_show_obj_v(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	view_show_obj(VAR[arg]);
}

void ao_echo_line(vm_t *l)
{
	text_echoline();
	text_new_key(0);	/* update input line */
}

void ao_cancel_line(vm_t *l)
{
	input_pos = 0;
	inputstring[input_pos] = 0;
	text_new_key(0);	/* update input line */
}

void ao_close_window(vm_t *l)
{
/*	memcpy(workvisbuffer + 2560, backvisbuffer, 320 * 168);
	memcpy(workpribuffer + 2560, backpribuffer, 320 * 168);*/
	message_erasebox();
}


void ao_open_dialogue(vm_t *l) 
{
	status.dialogue_open = true;
}

void ao_close_dialogue(vm_t *l) 
{
	status.dialogue_open = false;
}

void ao_restart_game(vm_t *l)
{
	int i, k;
	UINT8 maxscore;

	if(1) {
		memcpy(textmodebuffer, video_buffer, 64000); /* save screen contents */
		message_mbox("Press ENTER to restart\nthe game.\n\nPress ESC to continue\nthis game.", 0, 0, 0);
		driver_showbuffer(video_buffer);
	

		while(1) {
			k = event_waitforkey();
			if((k == 0x0D) || (k == '\n')) break;
			if(k == 27) {
				message_erasebox();
				memcpy(video_buffer, textmodebuffer, 64000); /* restore screen contents */
				return;
			}
		}
		message_erasebox();
		memcpy(video_buffer, textmodebuffer, 64000); /* restore screen contents */
	}

	ao_stop_sound(l);
	
	memset(object, 0, sizeof(view_obj) * 255); /* Reset view table */

	for(i = 0; i < 256; i++) { /* Unload all resources except LOGIC.0 */
		if(i && LOGIC[i] && LOGIC[i]->data) res_unload(LOGIC[i]);
		
		if(PICTURE[i] && PICTURE[i]->data) res_unload(PICTURE[i]);
		if(VIEW[i] && VIEW[i]->data) res_unload(VIEW[i]);
		if(SOUND[i] && SOUND[i]->data) res_unload(SOUND[i]);
	}
	res_load_objects(); /* Reset inventory items */

	maxscore = VAR[7];
	vm_init();	/* Reset flags, vars, and strings */
	VAR[7] = maxscore;
	
	/* some initial values */
	FLAG[5] = true;
	FLAG[6] = true;
	FLAG[9] = true;
	FLAG[11] = true;

	vm_reset = true; /* exit vm on return */
}

void ao_save_game(vm_t *l) 
{
	save_savedialogue();
}

void ao_restore_game(vm_t *l) 
{
	save_restoredialogue();
	FLAG[9] = true;
	vm_reset = true;
}

void ao_pause(vm_t *l)
{
	memcpy(textmodebuffer, video_buffer, 64000); /* save screen contents */
	message_mbox("      Game Paused.\nPress Enter to continue.", 0, 0, 0);
	driver_showbuffer(video_buffer);
	event_waitforkey();
	message_erasebox();
	memcpy(video_buffer, textmodebuffer, 64000); /* restore screen contents */
}

void ao_quit(vm_t *l)
{
	UINT8 arg;
	int k;

	if(action_opcode[0x86].args == 1) arg = vm_pop(); else arg = 0;
	if(arg) exit(1);
	
	memcpy(textmodebuffer, video_buffer, 64000); /* save screen contents */
	message_mbox("Press ENTER to quit.\nPress ESC to keep playing.", 0, 0, 0);
	driver_showbuffer(video_buffer);

	while((k = event_waitforkey()) != 27) 
		if((k == 0x0D) || (k == '\n')) exit(1);
	
	message_erasebox();
	memcpy(video_buffer, textmodebuffer, 64000); /* restore screen contents */
}


struct opcode_s action_opcode[0xB6] = {
   	{"return", 		0, /* This needs no handler */ NULL},
	{"increment", 		1, ao_increment},
	{"decrement", 		1, ao_decrement},
	{"assignn", 		2, ao_assignn},
	{"assignv", 		2, ao_assignv},
	{"addn", 		2, ao_addn},
	{"addv", 		2, ao_addv},
	{"subn", 		2, ao_subn},
	{"subv", 		2, ao_subv},
	{"lindirectv", 		2, ao_lindirectv},
	{"rindirect", 		2, ao_rindirect},
	{"lindirectn", 		2, ao_lindirectn},
	{"set", 		1, ao_set},
	{"reset",		1, ao_reset},
	{"toggle",		1, ao_toggle},
	{"set.v",		1, ao_setv},
	{"reset.v",		1, ao_resetv},
	{"toggle.v",		1, ao_togglev},
	{"new.room",		1, ao_new_room},
	{"new.room.v",		1, ao_new_room_v},
	{"load.logics",		1, ao_load_logic},
	{"load.logics.v",	1, ao_load_logic_v},
	{"call",		1, ao_call},
	{"call.v",		1, ao_call_v},
	{"load.pic",		1, ao_load_pic},
	{"draw.pic",		1, ao_draw_pic},
	{"show.pic",		0, ao_show_pic},
	{"discard.pic",		1, ao_discard_pic},
	{"overlay.pic",		1, ao_overlay_pic},
	{"show.pri.screen",	0, ao_show_pri_screen},
	{"load.view",		1, ao_load_view},
	{"load.view.v",		1, ao_load_view_v},
	{"discard.view",	1, ao_discard_view},
	{"animate.obj",		1, ao_animate_obj},
	{"unanimate.all",	0, ao_unanimate_all},
	{"draw",		1, ao_draw},
	{"erase",		1, ao_erase},
	{"position",		3, ao_position},
	{"position.v",		3, ao_position_v},
	{"get.posn",		3, ao_get_posn},
	{"reposition",		3, ao_reposition},
	{"set.view",		2, ao_set_view},
	{"set.view.v",		2, ao_set_view_v},
	{"set.loop",		2, ao_set_loop},
	{"set.loop.v",		2, ao_set_loop_v},
	{"fix.loop",		1, ao_fix_loop},
	{"release.loop",	1, ao_release_loop},
	{"set.cel",		2, ao_set_cel},
	{"set.cel.v",		2, ao_set_cel_v},
	{"last.cel",		2, ao_last_cel},
	{"current.cel",		2, ao_current_cel},
	{"current.loop",	2, ao_current_loop},
	{"current.view",	2, ao_current_view},
	{"number.of.loops",	2, ao_number_of_loops},
	{"set.priority",	2, ao_set_priority},
	{"set.priority.v",	2, ao_set_priority_v},
	{"release.priority",	1, ao_release_priority},
	{"get.priority",	2, ao_get_priority},
	{"stop.update",		1, ao_stop_update},
	{"start.update",	1, ao_start_update},
	{"force.update",	1, NULL},
	{"ignore.horizon",	1, ao_ignore_horizon},
	{"observe.horizon",	1, ao_observe_horizon},
	{"set.horizon",		1, ao_set_horizon},
	{"object.on.water",	1, ao_object_on_water},
	{"object.on.land",	1, ao_object_on_land},
	{"object.on.anything",	1, ao_object_on_anything},
	{"ignore.objs",		1, ao_ignore_objs},
	{"observe.objs",	1, ao_observe_objs},
	{"distance",		3, ao_distance},
	{"stop.cycling",	1, ao_stop_cycling},
	{"start.cycling",	1, ao_start_cycling},
	{"normal.cycle",	1, ao_normal_cycle},
	{"end.of.loop",		2, ao_end_of_loop},
	{"reverse.cycle",	1, ao_reverse_cycle},
	{"reverse.loop",	2, ao_reverse_loop},
	{"cycle.time",		2, ao_cycle_time},
	{"stop.motion",		1, ao_start_motion},
	{"start.motion",	1, ao_stop_motion},
	{"step.size",		2, ao_step_size},
	{"step.time",		2, ao_step_time},
	{"move.obj",		5, ao_move_obj},
	{"move.obj.v",		5, ao_move_obj_v},
	{"follow.ego",		3, ao_follow_ego},
	{"wander",		1, ao_wander},
	{"normal.motion",	1, ao_normal_motion},
	{"set.dir",		2, ao_set_dir},
	{"get.dir",		2, ao_get_dir},
	{"ignore.blocks",	1, ao_ignore_blocks},
	{"observe.blocks",	1, ao_observe_blocks},
	{"block",		4, NULL},
	{"unblock",		0, NULL},
	{"get",			1, ao_get},
	{"get.v",		1, ao_get_v},
	{"drop",		1, ao_drop},
	{"put",			2, ao_put},
	{"put.v",		2, ao_put_v},
	{"get.room.v",		2, ao_get_room_v},
	{"load.sound",		1, ao_load_sound},
	{"sound",		2, ao_sound},
	{"stop.sound",		0, ao_stop_sound},
	{"print",		1, ao_print},
	{"print.v",		1, ao_print_v},
	{"display",		3, ao_display},
	{"display.v",		3, ao_display_v},
	{"clear.lines",		3, ao_clear_lines},
	{"text.screen",		0, ao_text_screen},
	{"graphics",		0, ao_graphics},
	{"set.cursor.char",	1, ao_set_cursor_char},
	{"set.text.attribute",	2, ao_set_text_attribute},
	{"shake.screen",	1, NULL},
	{"configure.screen",	3, NULL},
	{"status.line.on",	0, ao_status_line_on},
	{"status.line.off",	0, ao_status_line_off},
	{"set.string",		2, ao_set_string},
	{"get.string",		5, ao_get_string},
	{"word.to.string",	2, ao_word_to_string},
	{"parse",		1, ao_parse},
	{"get.num",		2, ao_get_num},
	{"prevent.input",	0, ao_prevent_input},
	{"accept.input",	0, ao_accept_input},
	{"set.key",		3, ao_set_key},
	{"add.to.pic",		7, ao_add_to_pic},
	{"add.to.pic.v",	7, ao_add_to_pic_v},
	{"status",		0, ao_status},
	{"save.game",		0, ao_save_game},
	{"restore.game",	0, ao_restore_game},
	{"init.disk",		0, NULL},
	{"restart.game",	0, ao_restart_game},
	{"show.obj",		1, ao_show_obj},
	{"random",		3, ao_random},
	{"program.control",	0, ao_program_control},
	{"player.control",	0, ao_player_control},
	{"obj.status.v",	1, NULL},
	{"quit",		1, ao_quit},
	{"show.mem",		0, NULL},
	{"pause",		0, ao_pause},
	{"echo.line",		0, ao_echo_line},
	{"cancel.line",		0, ao_cancel_line},
	{"init.joy",		0, NULL},
	{"toggle.monitor",	0, NULL},
	{"version",		0, ao_version},
	{"script.size",		1, NULL},
	{"set.game.id",		1, NULL},
	{"log",			1, NULL},
	{"set.scan.start",	0, ao_set_scan_start},
	{"reset.scan.start",	0, ao_reset_scan_start},
	{"reposition.to",	3, /* What makes these different */ao_position},
	{"reposition.to.v",	3, /* from position & position.v */ao_position_v},
	{"trace.on",		0, NULL},
	{"trace.info",		3, NULL},
	{"print.at",		4, ao_print_at},
	{"print.at.v",		4, ao_print_at_v},
	{"discard.view.v",	1, ao_discard_view_v},
	{"clear.text.rect",	5, ao_clear_text_rect},
	{"set.upper.left",	2, NULL},
	{"set.menu",		1, ao_set_menu},
	{"set.menu.item",	2, ao_set_menu_item},
	{"submit.menu",		0, ao_submit_menu},
	{"enable.item",		1, ao_enable_item},
	{"disable.item",	1, ao_disable_item},
	{"menu.input",		0, ao_menu_input},
	{"show.obj.v",		1, ao_show_obj_v},
	{"open.dialogue",	0, ao_open_dialogue},
	{"close.dialogue",	0, ao_close_dialogue},
	{"mul.n",		2, ao_muln},
	{"mul.v",		2, ao_mulv},
	{"div.n",		2, ao_divn},
	{"div.v",		2, ao_divv},
	{"close.window",	0, ao_close_window},
	{"unknown170",		1, NULL},
	{"unknown171",		0, NULL},
	{"unknown172",		0, NULL},
	{"unknown173",		0, NULL},
	{"unknown174",		1, NULL},
	{"unknown175",		1, NULL},
	{"unknown176",		0, NULL},
	{"unknown177",		1, NULL},
	{"unknown178",		0, NULL},
	{"unknown179",		4, NULL},
	{"unknown180",		2, NULL},
	{"unknown181",		0, NULL}
};


