#include <stdio.h>
#include <stdlib.h>

#include "general.h"
#include "resource.h"
#include "vm.h"
#include "text.h"
#include "gameid.h"

static FILE * dumpfile = NULL;
static UINT16 codesize = 0;
static UINT8 * c = NULL;
static UINT16 IP = 0;

struct indent_s {
	int level;
	UINT16 endIP;
	struct indent_s * next;
};

static struct indent_s *curr_indent = NULL;

static void decompile(void);
static void handleif(void);


void decomp_main(unsigned int logicno)
{
	char ss[10];
	resource_t *l;

	sprintf(ss, "%d", logicno);
	text_input("Logic to decompile: ", 0, 24, 3, ss);
	sscanf(inputstring, "%d", (int *)&logicno);
	input_pos = 0; inputstring[input_pos] = 0;
	
	l = LOGIC[logicno];

	if(l == NULL) return;
	if(l->data == NULL) return;
		
	sprintf(ss, "LOGIC.%d", logicno);
	
	dumpfile = fopen(ss, "w");
	if(!dumpfile) return;

	fprintf(dumpfile, "/* DECOMPILE OF LOGIC %d\n", l->id);
	fprintf(dumpfile, " * FROM \"%s\" AGI VERSION %s\n", id->name, id->agiver);
	fprintf(dumpfile, " *\n");
	fprintf(dumpfile, " * Generated by: %s %s, (C) 2001, 2002 RaDIaT1oN\n", TITLE, VERSION);
	fprintf(dumpfile, " */\n\n");

	codesize = ((UINT16)l->data[1] << 8) + l->data[0];
	c = l->data + 2;
	IP = 0;

	decompile();
	
	fprintf(dumpfile, "\n");
	fclose(dumpfile);
}

static void decompile(void)
{
	int i;
	struct indent_s *tmp;
	
	for(IP = 0; IP < codesize; IP++) {
	
		if(curr_indent) 
			while(curr_indent->endIP == IP) {
				tmp = curr_indent->next;
				free(curr_indent);
				curr_indent = tmp;
				if(tmp)
					for(i = 0; i < curr_indent->level; i++) 
						fprintf(dumpfile, "\t");
				fprintf(dumpfile, "}\n");
				if(!tmp) break;
			}

		if(c[IP] == 0xFE) {
			UINT16 jumpsize;
			jumpsize = ((UINT16)c[IP + 2] << 8) + c[IP + 1];
			if(curr_indent && (curr_indent->endIP == (IP + 3))) {
				tmp = curr_indent->next;
				free(curr_indent);
				curr_indent = (struct indent_s *)malloc(sizeof(struct indent_s *));
				curr_indent->next = tmp;
				curr_indent->endIP = IP + jumpsize + 3;

				if(tmp) {
					curr_indent->level = tmp->level + 1;
					for(i = 0; i < tmp->level; i++) 
						fprintf(dumpfile, "\t");
				} else curr_indent->level = 1;
				fprintf(dumpfile, "} else {\n");
			} else  {
				if(curr_indent)
					for(i = 0; i < curr_indent->level; i++) 
						fprintf(dumpfile, "\t");
				fprintf(dumpfile, "JMP 0x%04X\n", jumpsize);
			}
			IP += 2;
			continue;
		}

		if(curr_indent)
			for(i = 0; i < curr_indent->level; i++) 
				fprintf(dumpfile, "\t");
		
		if(c[IP] == 0xFF) {
			handleif();
			continue;
		}
		fprintf(dumpfile, "%s (", action_opcode[c[IP]].name);
		for(i = 0; i < action_opcode[c[IP]].args; i++) {
			if(i) fprintf(dumpfile, ", ");
			fprintf(dumpfile, "%d", c[IP + i + 1]);
		}
		IP += action_opcode[c[IP]].args;
		fprintf(dumpfile, ");\n");
	}
}

static void handleif(void)
{
	int i;
	boolean or = false;
	boolean roundone = true;
	UINT16 jumpsize;	
	struct indent_s *tmp;

	fprintf(dumpfile, "if (");
	for(IP++; c[IP] != 0xFF; IP++) {
		if(!roundone && !or) {
			fprintf(dumpfile, "  &&\n");
			if(curr_indent)
				for(i = 0; i < curr_indent->level; i++) 
					fprintf(dumpfile, "\t");
			fprintf(dumpfile, "    ");
		}
			
		if(c[IP] == 0xFC) {
			or = !or;
			if(or) {
				fprintf(dumpfile, "(");
				roundone = true;
			} else 	fprintf(dumpfile, ")");
			continue;
		}
		
		if(!roundone && or) {
			fprintf(dumpfile, " ||\n");
			if(curr_indent)
				for(i = 0; i < curr_indent->level; i++) 
					fprintf(dumpfile, "\t");
			fprintf(dumpfile, "     ");
		} else roundone = false;

		if(c[IP] == 0xFD) {
			fprintf(dumpfile, "!");
			roundone = true;
			continue;
		}

		if(c[IP] == 0x0E) {	/* said() handler */
			UINT16 saidarg;

			fprintf(dumpfile, "said (");
			IP++;
			for(i = 0; i < c[IP]; i++) {
				saidarg = ((UINT16)c[IP + (i * 2) + 2] << 8) + c[IP + (i * 2) + 1];

				if(i) fprintf(dumpfile, ", ");
				fprintf(dumpfile, "0x%04X", saidarg);
			}
			
			IP += i << 1;
			fprintf(dumpfile, ")");
			continue;
		}

		fprintf(dumpfile, "%s (", test_opcode[c[IP]].name);
		for(i = 0; i < test_opcode[c[IP]].args; i++) {
			if(i) fprintf(dumpfile, ", ");
			fprintf(dumpfile, "%d", c[IP + i + 1]);
		}
		IP += test_opcode[c[IP]].args;
		fprintf(dumpfile, ")");
	}
	fprintf(dumpfile, ") {\n");

	jumpsize = ((UINT16)c[IP + 2] << 8) + c[IP + 1];
	IP += 2;

	tmp = curr_indent;
	curr_indent = (struct indent_s *)malloc(sizeof(struct indent_s));
	curr_indent->next = tmp;
	curr_indent->endIP = IP + jumpsize + 1;

	if(tmp) curr_indent->level = tmp->level + 1;
	else curr_indent->level = 1;
}

