/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 *
 * Allegro Console Driver for DOS port.  May be fixed later to work with
 * other plotforms supported by Allegro.
 */

#ifndef DJGPP
#error
#error The Allegro console driver is only for the DOS 
#error port, and must be compiled with DJGPP.
#error
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <allegro.h>
#include <sys/nearptr.h>

#include "driver.h"
#include "event.h"
#include "general.h"
#include "sound.h"


UINT8 *physicalscreen;
UINT32 screen_base_addr;

#define KEYBOARD_BUFFER_SIZE 10

volatile int keyboardbuffer[KEYBOARD_BUFFER_SIZE];
volatile int kbbufferpos_end = 0;
int kbbufferpos_start = 0;

void alleg_kb_handler(int key);

int driver_init (void)
{
        allegro_init();
        set_gfx_mode(GFX_VGA, 320, 200, 0, 0);

	/* This hack is DJGPP specific, copied from Allegro docs. */
        __djgpp_nearptr_enable();

        __dpmi_get_segment_base_address(screen->seg, &screen_base_addr);

        physicalscreen = (unsigned char *)(screen_base_addr + 
                                    screen->line[0] - 
                                    __djgpp_base_address);
	/* pointer physicalscreen now points to start of video memory */

        install_keyboard();

        keyboard_lowlevel_callback = alleg_kb_handler;

	sound_enabled = false;
	
        printf("driver: ALLEGRO: Console driver initialized\n");
	return 0;
}



void driver_shutdown (void)
{
        remove_keyboard();
        allegro_exit();
        printf("driver: ALLEGRO: Console driver shut down\n\n");
}


void driver_showbuffer (unsigned char *buf)
{
        memcpy(physicalscreen, buf, 64000);
}


void driver_kb_update (void)
{
        int key;
        int press;
        int scancode;
        int i;
        struct mapping { /* Map Allegro keysymbols to XT scancodes */
                int in, out;
        } key_mapping[] = {
                {KEY_ESC, AGIL_KEY_ESCAPE},
	
                {KEY_1, AGIL_KEY_1}, {KEY_2, AGIL_KEY_2}, {KEY_3, AGIL_KEY_3},
                {KEY_4, AGIL_KEY_4}, {KEY_5, AGIL_KEY_5}, {KEY_6, AGIL_KEY_6},
                {KEY_7, AGIL_KEY_7}, {KEY_8, AGIL_KEY_8}, {KEY_9, AGIL_KEY_9},
                {KEY_0, AGIL_KEY_0},

                {KEY_MINUS, AGIL_KEY_MINUS},

                {KEY_EQUALS, AGIL_KEY_EQUAL},

                {KEY_BACKSPACE, AGIL_KEY_BACKSPACE}, {KEY_TAB, AGIL_KEY_TAB},

                {KEY_Q, AGIL_KEY_Q}, {KEY_W, AGIL_KEY_W}, {KEY_E, AGIL_KEY_E},
                {KEY_R, AGIL_KEY_R}, {KEY_T, AGIL_KEY_T}, {KEY_Y, AGIL_KEY_Y},
                {KEY_U, AGIL_KEY_U}, {KEY_I, AGIL_KEY_I}, {KEY_O, AGIL_KEY_O},
                {KEY_P, AGIL_KEY_P},

                {KEY_OPENBRACE, AGIL_KEY_BRACKET_LEFT},
                {KEY_CLOSEBRACE, AGIL_KEY_BRACKET_RIGHT},
                {KEY_ENTER, AGIL_KEY_ENTER}, {KEY_LCONTROL, AGIL_KEY_CONTROL}, 

                {KEY_A, AGIL_KEY_A}, {KEY_S, AGIL_KEY_S}, {KEY_D, AGIL_KEY_D},
                {KEY_F, AGIL_KEY_F}, {KEY_G, AGIL_KEY_G}, {KEY_H, AGIL_KEY_H},
                {KEY_J, AGIL_KEY_J}, {KEY_K, AGIL_KEY_K}, {KEY_L, AGIL_KEY_L},

                {KEY_COLON, AGIL_KEY_SEMICOLON},
                {KEY_QUOTE, AGIL_KEY_APOSTROPHE}, {KEY_TILDE, AGIL_KEY_GRAVE},
                {KEY_LSHIFT, AGIL_KEY_LEFTSHIFT},
                {KEY_BACKSLASH, AGIL_KEY_BACKSLASH},

                {KEY_Z, AGIL_KEY_Z}, {KEY_X, AGIL_KEY_X}, {KEY_C, AGIL_KEY_C},
                {KEY_V, AGIL_KEY_V}, {KEY_B, AGIL_KEY_B}, {KEY_N, AGIL_KEY_N},
                {KEY_M, AGIL_KEY_M},

                {KEY_COMMA, AGIL_KEY_COMMA}, {KEY_STOP, AGIL_KEY_PERIOD},
                {KEY_SLASH, AGIL_KEY_SLASH}, {KEY_RSHIFT, AGIL_KEY_RIGHTSHIFT},
                {KEY_ASTERISK, AGIL_KEY_KEYPADMULTIPLY},
                {KEY_ALT, AGIL_KEY_ALT}, {KEY_SPACE, AGIL_KEY_SPACE},
                {KEY_CAPSLOCK, AGIL_KEY_CAPSLOCK},

                {KEY_F1, AGIL_KEY_F1}, {KEY_F2, AGIL_KEY_F2}, {KEY_F3, AGIL_KEY_F3},
                {KEY_F4, AGIL_KEY_F4}, {KEY_F5, AGIL_KEY_F5}, {KEY_F6, AGIL_KEY_F6},
                {KEY_F7, AGIL_KEY_F7}, {KEY_F8, AGIL_KEY_F8}, {KEY_F9, AGIL_KEY_F9},
                {KEY_F10, AGIL_KEY_F10}, 

                {KEY_NUMLOCK, AGIL_KEY_NUMLOCK},
                {KEY_SCRLOCK, AGIL_KEY_SCROLLLOCK},

                {KEY_7_PAD, AGIL_KEY_CURSORUPLEFT},
                {KEY_8_PAD, AGIL_KEY_CURSORUP},
                {KEY_9_PAD, AGIL_KEY_CURSORUPRIGHT},
                {KEY_4_PAD, AGIL_KEY_CURSORLEFT},
                {KEY_6_PAD, AGIL_KEY_CURSORRIGHT},
                {KEY_1_PAD, AGIL_KEY_CURSORDOWNLEFT},
                {KEY_2_PAD, AGIL_KEY_CURSORDOWN},
                {KEY_3_PAD, AGIL_KEY_CURSORDOWNRIGHT},

                {KEY_HOME, AGIL_KEY_CURSORUPLEFT},
                {KEY_UP, AGIL_KEY_CURSORUP},
                {KEY_PGUP, AGIL_KEY_CURSORUPRIGHT},
                {KEY_LEFT, AGIL_KEY_CURSORLEFT},
                {KEY_RIGHT, AGIL_KEY_CURSORRIGHT},
                {KEY_END, AGIL_KEY_CURSORDOWNLEFT},
                {KEY_DOWN, AGIL_KEY_CURSORDOWN}       ,
                {KEY_PGDN, AGIL_KEY_CURSORDOWNRIGHT}, 

                {KEY_F11, AGIL_KEY_F11}, {KEY_F12, AGIL_KEY_F12},
                {KEY_ENTER_PAD, AGIL_KEY_ENTER},
                {KEY_RCONTROL, AGIL_KEY_CONTROL},
                {KEY_SLASH_PAD, AGIL_KEY_KEYPADDIVIDE},
                {KEY_PAUSE, AGIL_KEY_BREAK},

		{0, 0}
	};

	
        while(kbbufferpos_start != kbbufferpos_end) {
                key = keyboardbuffer[++kbbufferpos_start];
		kbbufferpos_start %= KEYBOARD_BUFFER_SIZE;
		
                press = !(key & 0x80);
                scancode = key & 0x7F;
		
                if(scancode == KEY_F12) {
                        printf("driver: ALLEGRO: Emergency exit called!\n");
                        exit(-1);
                }
		
                for(i = 0; key_mapping[i].in; i++) {
                        if(key_mapping[i].in == scancode) {
                                scancode = key_mapping[i].out;
                                break;
                        }
                }

                event_kb_handler(scancode, press);
       }
}

void alleg_kb_handler(int key)
{
        keyboardbuffer[++kbbufferpos_end] = key;
	kbbufferpos_end %= KEYBOARD_BUFFER_SIZE;
}



