/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */
 
/*
 * SDL driver for Win32/Linux/Whatever OS SDL has a target for ;)
 * (c) 2001 LollypopMan (lee@the-mothership.fsnet.co.uk)
 * based on the allegro driver (c) 2001 RaDIaT1oN
 *
 */
 
/* 28-12-2001 Added a threaded sound driver that emulates the
 *            SN764968 PSG
 *            - LollypopMan
 */

/*
 * RaDIaT1oN moved sound code to sound.c to implement new global
 * sound interface for drivers. (Don't know when, for version 0.1.3.)
 *
 * Also keyboard code fixed.  Was a problem with -fshort_enums in Makefile.
 * SDL in not compiled with -fshort_enums, and therefore var sizes don't match.
 */
  
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "SDL.h"
#include "SDL_thread.h"
#include "SDL_audio.h"

#include "event.h"
#include "general.h"
#include "driver.h"
#include "vm.h"
#include "gameid.h"
#include "sound.h"

static Uint32 mode_flags;
static SDL_Surface *thescreen;
void DrawPixel(SDL_Surface *screen, int x, int y, Uint8 R, Uint8 G, Uint8 B);
void driver_sound_callback(void *unused, Uint8 *stream, int len);

typedef struct {
	Uint8 r, g, b;
} rgbtriple;

rgbtriple vgapalette[] = 
{
	{   0,   0,   0},
	{   0,   0,  42},
	{   0,  42,   0},
	{   0,  42,  42},
	{  42,   0,   0},
	{  42,   0,  42},
	{  42,  21,   0},
	{  42,  42,  42},
	{  21,  21,  21},
	{  21,  21,  63},
	{  21,  63,  21},
	{  21,  63,  63},
	{  63,  21,  21},
	{  63,  21,  63},
	{  63,  63,  21},
	{  63,  63,  63}
};


int driver_init (void)
{ 
	SDL_AudioSpec fmt;
	if(atexit(driver_shutdown)) return -1;
	
	if ( SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO|SDL_INIT_TIMER) < 0 ) {
        	fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
        	exit(1);
	}
	
	atexit(SDL_Quit);

	SDL_WM_SetCaption(id->name, id->name);
	mode_flags = SDL_SWSURFACE;
	thescreen = SDL_SetVideoMode(320, 200, 32, mode_flags);
	if ( thescreen == NULL ) {
        	fprintf(stderr, "Unable to set 320x200 video: %s\n", SDL_GetError());
		exit(1);
	}
	                    
	/* Set 16-bit signed mono audio at 22Khz */
	fmt.freq = 22050;	/* 22KHz */
	fmt.format = AUDIO_S16; /* Sound data from sound.c */
	fmt.channels = 1;	/* Mono */
	fmt.samples = 512;      /* Must be greater than 400 */
	fmt.callback = driver_sound_callback;
	fmt.userdata = NULL;

	/* Open the audio device and start playing sound! */
	if ( SDL_OpenAudio(&fmt, NULL) < 0 ) {
		fprintf(stderr, "Unable to open audio: %s\n", SDL_GetError());
		sound_enabled = false;
	} else SDL_PauseAudio(0); 
	
        printf("driver: SDL: Console driver initialized\n");
	return 0;
}


void driver_shutdown (void)
{
	SDL_PauseAudio(1);
	SDL_Quit();
        printf("driver: SDL: Console driver shut down\n\n");
}

void driver_showbuffer (unsigned char *buf)
{
	int x,y;
	Uint8 col,r,g,b;
	
	if ( SDL_MUSTLOCK(thescreen) ) {
        	if ( SDL_LockSurface(thescreen) < 0 ) {
            	return;
        	}
   	 }
	for(y=0;y<200;y++)
		for(x=0;x<320;x++)
		{
        		col = buf[x + (y * 320)];
        		r = vgapalette[col].r << 2;
        		g = vgapalette[col].g << 2;
        		b = vgapalette[col].b << 2;
        		DrawPixel(thescreen, x, y, r, g, b);
/*        		DrawPixel(thescreen, (x*2) + 1, (y*2)    , r, g, b);
        		DrawPixel(thescreen, (x*2)    , (y*2) + 1, r, g, b);
        		DrawPixel(thescreen, (x*2) + 1, (y*2) + 1, r, g, b);*/
        	}
        	
    	if ( SDL_MUSTLOCK(thescreen) ) {
       		 SDL_UnlockSurface(thescreen);
    	}
    	
    	SDL_UpdateRect(thescreen, 0, 0, 320, 200);
}


void driver_kb_update (void)
{
	SDL_Event event;
        int i;
	int scancode = 0;
	int press = 0;
	
        struct mapping { 
                int in, out;
        } key_mapping[] = {
                {SDLK_ESCAPE, AGIL_KEY_ESCAPE},
	
                {SDLK_1, AGIL_KEY_1}, {SDLK_2, AGIL_KEY_2}, {SDLK_3, AGIL_KEY_3},
                {SDLK_4, AGIL_KEY_4}, {SDLK_5, AGIL_KEY_5}, {SDLK_6, AGIL_KEY_6},
                {SDLK_7, AGIL_KEY_7}, {SDLK_8, AGIL_KEY_8}, {SDLK_9, AGIL_KEY_9},
                {SDLK_0, AGIL_KEY_0},

                {SDLK_MINUS, AGIL_KEY_MINUS},

                {SDLK_EQUALS, AGIL_KEY_EQUAL},

                {SDLK_BACKSPACE, AGIL_KEY_BACKSPACE}, {SDLK_TAB, AGIL_KEY_TAB},

                {SDLK_q, AGIL_KEY_Q}, {SDLK_w, AGIL_KEY_W}, {SDLK_e, AGIL_KEY_E},
                {SDLK_r, AGIL_KEY_R}, {SDLK_t, AGIL_KEY_T}, {SDLK_y, AGIL_KEY_Y},
                {SDLK_u, AGIL_KEY_U}, {SDLK_i, AGIL_KEY_I}, {SDLK_o, AGIL_KEY_O},
                {SDLK_p, AGIL_KEY_P},

                {SDLK_LEFTPAREN, AGIL_KEY_BRACKET_LEFT},
                {SDLK_RIGHTPAREN, AGIL_KEY_BRACKET_RIGHT},
                {SDLK_RETURN, AGIL_KEY_ENTER}, {SDLK_LCTRL, AGIL_KEY_CONTROL}, 

                {SDLK_a, AGIL_KEY_A}, {SDLK_s, AGIL_KEY_S}, {SDLK_d, AGIL_KEY_D},
                {SDLK_f, AGIL_KEY_F}, {SDLK_g, AGIL_KEY_G}, {SDLK_h, AGIL_KEY_H},
                {SDLK_j, AGIL_KEY_J}, {SDLK_k, AGIL_KEY_K}, {SDLK_l, AGIL_KEY_L},

                {SDLK_SEMICOLON, AGIL_KEY_SEMICOLON},
                {SDLK_QUOTE, AGIL_KEY_APOSTROPHE}, {SDLK_BACKQUOTE, AGIL_KEY_GRAVE},
                {SDLK_LSHIFT, AGIL_KEY_LEFTSHIFT},
                {SDLK_BACKSLASH, AGIL_KEY_BACKSLASH},

                {SDLK_z, AGIL_KEY_Z}, {SDLK_x, AGIL_KEY_X}, {SDLK_c, AGIL_KEY_C},
                {SDLK_v, AGIL_KEY_V}, {SDLK_b, AGIL_KEY_B}, {SDLK_n, AGIL_KEY_N},
                {SDLK_m, AGIL_KEY_M},

                {SDLK_COMMA, AGIL_KEY_COMMA}, {SDLK_PERIOD, AGIL_KEY_PERIOD},
                {SDLK_SLASH, AGIL_KEY_SLASH}, {SDLK_RSHIFT, AGIL_KEY_RIGHTSHIFT},
                {SDLK_ASTERISK, AGIL_KEY_KEYPADMULTIPLY},
                {SDLK_LALT, AGIL_KEY_ALT}, {SDLK_SPACE, AGIL_KEY_SPACE},
                {SDLK_CAPSLOCK, AGIL_KEY_CAPSLOCK},

                {SDLK_F1, AGIL_KEY_F1}, {SDLK_F2, AGIL_KEY_F2}, {SDLK_F3, AGIL_KEY_F3},
                {SDLK_F4, AGIL_KEY_F4}, {SDLK_F5, AGIL_KEY_F5}, {SDLK_F6, AGIL_KEY_F6},
                {SDLK_F7, AGIL_KEY_F7}, {SDLK_F8, AGIL_KEY_F8}, {SDLK_F9, AGIL_KEY_F9},
                {SDLK_F10, AGIL_KEY_F10}, 

                {SDLK_NUMLOCK, AGIL_KEY_NUMLOCK},
                {SDLK_SCROLLOCK, AGIL_KEY_SCROLLLOCK},

                {SDLK_KP7, AGIL_KEY_CURSORUPLEFT},
                {SDLK_KP8, AGIL_KEY_CURSORUP},
                {SDLK_KP9, AGIL_KEY_CURSORUPRIGHT},
                {SDLK_KP4, AGIL_KEY_CURSORLEFT},
                {SDLK_KP6, AGIL_KEY_CURSORRIGHT},
                {SDLK_KP1, AGIL_KEY_CURSORDOWNLEFT},
                {SDLK_KP2, AGIL_KEY_CURSORDOWN},
                {SDLK_KP3, AGIL_KEY_CURSORDOWNRIGHT},

                {SDLK_HOME, AGIL_KEY_CURSORUPLEFT},
                {SDLK_UP, AGIL_KEY_CURSORUP},
                {SDLK_PAGEUP, AGIL_KEY_CURSORUPRIGHT},
                {SDLK_LEFT, AGIL_KEY_CURSORLEFT},
                {SDLK_RIGHT, AGIL_KEY_CURSORRIGHT},
                {SDLK_END, AGIL_KEY_CURSORDOWNLEFT},
                {SDLK_DOWN, AGIL_KEY_CURSORDOWN}       ,
                {SDLK_PAGEDOWN, AGIL_KEY_CURSORDOWNRIGHT}, 

                {SDLK_F11, AGIL_KEY_F11}, {SDLK_F12, AGIL_KEY_F12},
                {SDLK_KP_ENTER, AGIL_KEY_ENTER},
                {SDLK_RCTRL, AGIL_KEY_CONTROL},
                {SDLK_KP_DIVIDE, AGIL_KEY_KEYPADDIVIDE},
                {SDLK_PAUSE, AGIL_KEY_BREAK},

                {SDLK_DELETE, AGIL_KEY_BREAK},

		{0, 0}
	};

	while(SDL_PollEvent(&event) > 0) {
		if(event.type == SDL_QUIT) {
			printf("driver: SDL: FATAL: SDL_QUIT event occured\n");
			exit(0);
		}
		
		if((event.type != SDL_KEYUP) && (event.type != SDL_KEYDOWN)) 
			continue; 

		if(event.type == SDL_KEYDOWN) press = 1;
		else press = 0;

		scancode = event.key.keysym.sym;

		/* F12 == Emergency exit */
	        if(scancode == SDLK_F12) {
	                printf("driver: SDL: Emergency exit called!\n");
	                exit(-1);
	        }
		
		/* F11 == Full screen toggle key */
		if((scancode == SDLK_F11) && press) {
			mode_flags ^= SDL_FULLSCREEN;
			thescreen = SDL_SetVideoMode(320, 200, 32, mode_flags);
			if ( thescreen == NULL ) {
		        	fprintf(stderr, "Unable to set 320x200 video: %s\n", SDL_GetError());
				mode_flags ^= SDL_FULLSCREEN;
				thescreen = SDL_SetVideoMode(320, 200, 32, mode_flags);
			}
			continue;
		}
		
		for(i = 0; key_mapping[i].in; i++) { 
			if(scancode == key_mapping[i].in) {
				scancode = key_mapping[i].out;
				break;
			}
		}
		
/*		printf("Key event : %d, %d\n", scancode, press);*/
	
		event_kb_handler(scancode, press);
	}

}


void DrawPixel(SDL_Surface *screen, int x, int y, Uint8 R, Uint8 G, Uint8 B)
{
	Uint32 color = SDL_MapRGB(screen->format, R, G, B);

	switch (screen->format->BytesPerPixel) {
	        case 1: { /* Assuming 8-bpp */
	        		Uint8 *bufp;

				bufp = (Uint8 *)screen->pixels + y*screen->pitch + x;
				*bufp = color;
			}
		        break;

	        case 2: { /* Probably 15-bpp or 16-bpp */
				Uint16 *bufp;

				bufp = (Uint16 *)screen->pixels + y*screen->pitch/2 + x;
				*bufp = color;
			}
			break;

	        case 3: { /* Slow 24-bpp mode, usually not used */
        			Uint8 *bufp;

				bufp = (Uint8 *)screen->pixels + y*screen->pitch + x;
				*(bufp+screen->format->Rshift/8) = R;
				*(bufp+screen->format->Gshift/8) = G;
				*(bufp+screen->format->Bshift/8) = B;
			}
			break;

		case 4: { /* Probably 32-bpp */
				Uint32 *bufp;

				bufp = (Uint32 *)screen->pixels + y*screen->pitch/4 + x;
				*bufp = color;
			}
			break;
	}
}


void driver_sound_callback(void *unused, Uint8 *stream, int len)
{
	static SINT16 remainder[BUFFER_SAMPLE_SIZE];
	static int remlen = 0;
	SINT16 * buffer = NULL;
	int i = 0;
    
	if(remlen) {
		memcpy(stream, remainder, remlen);
		i += remlen;
	}
	
	while((len - i) > 0) {
		buffer = sound_getbuffer();
		if((len - i) < BUFFER_BYTE_SIZE) {
			memcpy(stream + i, buffer, len - i);
			remlen = BUFFER_BYTE_SIZE - (len - i);
			memcpy(remainder, buffer + ((len - i) >> 1), remlen);
			i = len;
		} else {
			memcpy(stream + i, buffer, BUFFER_BYTE_SIZE);
			i += BUFFER_BYTE_SIZE;
			remlen = 0;
		}
	}
}

