/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

/*
 *  Console driver for Linux/svgalib.
 */
 
#include <stdlib.h>
#include <stdio.h>
#include <vga.h>
#include <unistd.h>
#include <vgakeyboard.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <pthread.h>

#include "driver.h"
#include "event.h"
#include "sound.h"

/* My stupid screen card dosen't work with standard VGA modes in SVGALIB.
 * CUSTOMMODE must be 256 colors (8 bit), else fix driver_init() 
 * and driver_showbuffer() */
/*#define CUSTOMMODE G800x600x256 */

#ifndef CUSTOMMODE
#define GRAPHICSMODE G320x200x256
#else
#include <vgagl.h>
#define GRAPHICSMODE CUSTOMMODE
#endif

#define FRAGNUM		(32 - 1)
#define FRAGSIZE	(9)


void svgalib_kb_event_handler(int scancode, int press);

static int audio_fd;
static pthread_t sound_thread;
void * driver_sound_thread(void * arg);


/* General console driver functions */
int driver_init(void) 
{
	int i;
	
	/* Save ourselves many reboots (hopefully) */
	if(atexit(driver_shutdown)) return -1;

	/* Initialize graphics */
	if(vga_init()) return -2;

	if(vga_hasmode(GRAPHICSMODE)) 
		vga_setmode(GRAPHICSMODE);
	else 
		return -3;

#ifdef CUSTOMMODE	
	if(gl_setcontextvga(GRAPHICSMODE)) return -4;
	gl_clearscreen(0);
#endif
	
	/* Initialize keyboard */
	if(keyboard_init()) return -5;
	keyboard_seteventhandler(svgalib_kb_event_handler);
	keyboard_clearstate();
	
	/* Initialize sound driver */
        if ((audio_fd = open ("/dev/dsp", O_WRONLY)) != -1) {
		/* Fragment stuff copied from Sarien */
		i = (FRAGNUM << 16 | FRAGSIZE);
		ioctl (audio_fd, SNDCTL_DSP_SETFRAGMENT, &i);

	        i = AFMT_S16_LE;
	        ioctl (audio_fd, SNDCTL_DSP_SETFMT, &i);
	        i = 0;
	        ioctl (audio_fd, SNDCTL_DSP_STEREO, &i);
	        i = 22050;
	        ioctl (audio_fd, SNDCTL_DSP_SPEED, &i);
	
		pthread_create(&sound_thread, NULL, driver_sound_thread, NULL);
		pthread_detach(sound_thread);
		printf("driver: OSS: Sound initialized\n");
	} else sound_enabled = false;

	printf("driver: SVGALIB: Console driver initialized\n");
	return 0;
}


void driver_shutdown(void) 
{
	printf("driver: SVGALIB: Console driver shutdown\n\n");

	if(audio_fd != -1) {
		pthread_cancel(sound_thread);
		close(audio_fd);
	}
	
	if(vga_getcurrentmode() == TEXT) return; /* already shut down */

	/* Shutdown keyboard */
	keyboard_setdefaulteventhandler();
	keyboard_clearstate();
	keyboard_translatekeys(0);
	keyboard_close();
	
	/* Shutdown Graphics */
	vga_setmode(TEXT);
}


/* Graphics driver functions */
void driver_showbuffer(unsigned char * buffer)
{
#ifndef CUSTOMMODE
	/* Fast method for 320x200 */
	vga_drawscansegment(buffer, 0, 0, 64000); 
#else	
	/* Slow method for custom graphics modes */
	/* WIDTH and HEIGHT are defined by libvgagl */
	char * bigbuffer = NULL;

	if(!(bigbuffer = (unsigned char *)malloc(WIDTH * HEIGHT))) return;
	gl_scalebox(320, 200, buffer, WIDTH, HEIGHT, bigbuffer);
	gl_putbox(0, 0, WIDTH, HEIGHT, bigbuffer);
	free(bigbuffer);
#endif	
}


/* Keyboard driver functions */
void driver_kb_update(void)
{
	keyboard_update();
}


/* Not public driver function, but required for private use 
 *
 * This function must convert scancodes to XT equivalents before
 * Passing control to the core keyboard event handler */
void svgalib_kb_event_handler(int scancode, int press)
{
	int i;
	
	struct mapping {	/* Translation table of scancodes */
		int in, out;
	} keycodemap[] = {

		{SCANCODE_KEYPADENTER, AGIL_KEY_ENTER},
		
		{SCANCODE_INSERT, AGIL_KEY_KEYPAD0},
		{SCANCODE_HOME, AGIL_KEY_KEYPAD7},
		{SCANCODE_PAGEUP, AGIL_KEY_KEYPAD9},
		{SCANCODE_REMOVE, AGIL_KEY_KEYPADPERIOD},
		{SCANCODE_END, AGIL_KEY_KEYPAD1},
		{SCANCODE_PAGEDOWN, AGIL_KEY_KEYPAD3},
		
		{SCANCODE_CURSORBLOCKUP, AGIL_KEY_KEYPAD8},
		{SCANCODE_CURSORBLOCKLEFT, AGIL_KEY_KEYPAD4},
		{SCANCODE_CURSORBLOCKDOWN, AGIL_KEY_KEYPAD2},
		{SCANCODE_CURSORBLOCKRIGHT, AGIL_KEY_KEYPAD6},
		
		{SCANCODE_BREAK_ALTERNATIVE, AGIL_KEY_BREAK},

		{SCANCODE_RIGHTCONTROL, AGIL_KEY_CONTROL},
		{SCANCODE_RIGHTALT, AGIL_KEY_ALT},

		{0, 0} /* terminate array */
	};

	/* Translate all keys to their XT equivalents */
	for(i = 0; keycodemap[i].in != 0; i++) {
		if(keycodemap[i].in == scancode) 
			scancode = keycodemap[i].out;
	}

	if(scancode == SCANCODE_F12) { /* Ditch on F12 */
		printf("driver: EMERGENCY EXIT CALLED!\n");
		exit(-2);
	}
		
	/* call core keyboard event handler */
	event_kb_handler(scancode & 0x7F, press);
}


/* Sound driver stuff */
void * driver_sound_thread(void * arg)
{
	int i;
	SINT16 * buffer;

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	while(1) {
		if(!playing) {
			usleep(SOUND_DUR_USEC);
			continue;
		}
		buffer = sound_getbuffer();
	        for (i = BUFFER_BYTE_SIZE; i -= write (audio_fd, buffer, i); );
	}
}

