/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 * 
 * Mostly from LAGII, (C) 1999 XoXus
 */

/*
 *  Console driver for Linux/X11
 *  From LAGII by XoXus
 *
 *  Fixed for AGIL by me, also now works with TrueColor visuals,
 *  	but much slower.  If somebody want to take over this driver
 *	please let me know.
 */

#include <stdlib.h> 
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <pthread.h>
#include <X11/keysym.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "general.h"
#include "event.h"
#include "driver.h"
#include "gameid.h"
#include "sound.h"

#define FRAGNUM		(32 - 1)
#define FRAGSIZE	(9)

void x11_death(int sig); /* private driver function */


Display *x_display;
Window x_root, x_window;
int x_screen;
Visual *x_visual;
Colormap x_colormap;
XSetWindowAttributes x_attributes;
XImage *ximg;
GC x_gc;
UINT32 pixelpalette[16];

static int audio_fd;
static pthread_t sound_thread;
void * driver_sound_thread(void * arg);


int driver_init (void)
{
	int i;
	XColor xcol;
	Status err;
	int depth;

	unsigned int x_ega_palette[16][3] = { /* RGB values of EGA palette */
		{0, 0, 0},
		{0, 0, 43690},
		{0, 43690, 0},
		{0, 43690, 43690},
		{43690, 0, 0},
		{43690, 0, 43690},
		{43690, 21845, 0},
		{43690, 43690, 43690},
		{21845, 21845, 21845},
		{21845, 21845, 65535},
		{21845, 65535, 21845},
		{21845, 65535, 65535},
		{65535, 21845, 21845},
		{65535, 21845, 65535},
		{65535, 65535, 21845},
		{65535, 65535, 65535}
	};

	XWMHints hints;
	XSizeHints sizehints;
	XClassHint classhint;
	XTextProperty appname, iconname;
	char *apptext;	/* = "AGIL";	LAGII_TITLE " " LAGII_VERSION;*/
	char *icontext;	/* = "AGIL";	LAGII_TITLE;*/

	if(atexit(driver_shutdown)) return -1;

	x_display = XOpenDisplay (NULL);
	if (!x_display) {
		printf("driver: X11: FATAL: Can't open X display!\n");
		return -1;
	}

	x_root = DefaultRootWindow (x_display);
	x_screen = DefaultScreen (x_display);
	x_visual = DefaultVisual (x_display, x_screen);


	/* Set EGA palette */
/*	x_colormap = XCreateColormap (x_display, x_root, x_visual, AllocNone);*/
	x_colormap = DefaultColormap(x_display, x_screen);
	for (i=0; i < 16; i++) {

		xcol.red = x_ega_palette[i][0];
		xcol.green = x_ega_palette[i][1];
		xcol.blue = x_ega_palette[i][2];
		err = XAllocColor (x_display, x_colormap, &xcol);
		if (!err) {
			printf("driver: X11: FATAL: Couldn't alloc color %i!\n", i);
/*			XFreeColormap(x_colormap);*/
			XCloseDisplay(x_display);
			return -1;
		}
		pixelpalette[i] = xcol.pixel;
	}

	/* Create window */
/*	x_attributes.event_mask = (KeyPressMask | KeyReleaseMask);
	x_window = XCreateWindow (x_display, x_root, 0, 0, 320, 200, 0,
			CopyFromParent, InputOutput, CopyFromParent,
			CWEventMask, &x_attributes);*/

	x_window = XCreateSimpleWindow(x_display, x_root, 0, 0, 320, 200, 0,
		pixelpalette[0], pixelpalette[0]);
	XSelectInput(x_display, x_window, (KeyPressMask | KeyReleaseMask));

	/* Set-up hints */
	apptext = id->name;
	icontext = id->name;
	XStringListToTextProperty (&apptext, 1, &appname);
	XStringListToTextProperty (&icontext, 1, &iconname);
	sizehints.flags = (PSize | PMinSize | PMaxSize);
	sizehints.min_width = 320;
	sizehints.max_width = 320;
	sizehints.min_height = 200;
	sizehints.max_height = 200;
	hints.flags = (StateHint | IconPixmapHint | InputHint);
	hints.initial_state = NormalState;
	hints.input = 1;
	classhint.res_name = "AGIL";
	classhint.res_class = "AGIL";
	XSetWMProperties (x_display, x_window, &appname, &iconname,
			NULL, 0, &sizehints, &hints, &classhint);

	x_gc = DefaultGC (x_display, x_screen);
	depth = DefaultDepth (x_display, x_screen);

	ximg = XCreateImage (x_display, x_visual, depth, ZPixmap, 0,
			NULL, 320, 200, 8, 0);
	ximg->data = (void *)malloc(ximg->bytes_per_line * 200);
	
	/* Initialize sound driver */
        if ((audio_fd = open ("/dev/audio", O_WRONLY)) != -1) {
		/* Frament stuff copied from Sarien */
		i = (FRAGNUM << 16 | FRAGSIZE);
		ioctl (audio_fd, SNDCTL_DSP_SETFRAGMENT, &i);

	        i = AFMT_S16_LE;
	        ioctl (audio_fd, SNDCTL_DSP_SETFMT, &i);
	        i = 0;
	        ioctl (audio_fd, SNDCTL_DSP_STEREO, &i);
	        i = 22000;
	        ioctl (audio_fd, SNDCTL_DSP_SPEED, &i);

		pthread_create(&sound_thread, NULL, driver_sound_thread, NULL);
		pthread_detach(sound_thread);
		printf("driver: OSS: Sound initialized\n");
	} else sound_enabled = false;

	/* Map window */
	XMapRaised (x_display, x_window);

	/* Synchronise with X server */
	XSync (x_display, False);

	/* Just in case I get a nasty signal */
	signal (SIGHUP, x11_death);
	signal (SIGINT, x11_death);
	signal (SIGQUIT, x11_death);
	signal (SIGILL, x11_death);
	signal (SIGTERM, x11_death);

	printf("driver: X11: driver initialized\n");
	return 0;
}



void driver_shutdown (void)
{
	if(!x_display) return;

	if(audio_fd != -1) {
		pthread_cancel(sound_thread);
		close(audio_fd);
	}

	XDestroyImage(ximg);
	XFreeColormap (x_display, x_colormap);
	XUnmapWindow (x_display, x_window);
	XDestroyWindow (x_display, x_window);
	XCloseDisplay (x_display);
	printf("driver: X11: driver shut down\n\n");
}


void driver_showbuffer (unsigned char *buf)
{
	int x, y;
	
	/* Draw buffer on ximg */
	for(x = 0; x < 320; x++) 
		for(y = 0; y < 200; y++) 
			XPutPixel(ximg, x, y, pixelpalette[(int)buf[(y * 320) + x]]);
			
	XPutImage (x_display, x_window, x_gc, ximg, 0, 0, 0, 0, 640, 400);
	XSync (x_display, False);
}


void driver_kb_update (void)
{
	XEvent ev;
	int i, scancode, press;
	KeySym key;
	char buf[2];

	struct x_key_mapping_s { /* Map X keysymbols to XT scancodes */
		KeySym keysym;
		int scancode;
	} x_key_mapping[] = {
		{XK_Escape, AGIL_KEY_ESCAPE},
	
		{XK_1, AGIL_KEY_1}, {XK_2, AGIL_KEY_2}, {XK_3, AGIL_KEY_3},
		{XK_4, AGIL_KEY_4}, {XK_5, AGIL_KEY_5}, {XK_6, AGIL_KEY_6},
		{XK_7, AGIL_KEY_7}, {XK_8, AGIL_KEY_8}, {XK_9, AGIL_KEY_9},
		{XK_0, AGIL_KEY_0},

		{XK_KP_Subtract, AGIL_KEY_MINUS},

		{XK_KP_Equal, AGIL_KEY_EQUAL},

		{XK_BackSpace, AGIL_KEY_BACKSPACE}, {XK_Tab, AGIL_KEY_TAB},

		{XK_q, AGIL_KEY_Q}, {XK_w, AGIL_KEY_W}, {XK_e, AGIL_KEY_E},
		{XK_r, AGIL_KEY_R}, {XK_t, AGIL_KEY_T}, {XK_y, AGIL_KEY_Y},
		{XK_u, AGIL_KEY_U}, {XK_i, AGIL_KEY_I}, {XK_o, AGIL_KEY_O},
		{XK_p, AGIL_KEY_P},

		{XK_bracketleft, AGIL_KEY_BRACKET_LEFT},
		{XK_bracketright, AGIL_KEY_BRACKET_RIGHT},
		{XK_Return, AGIL_KEY_ENTER}, {XK_Control_L, AGIL_KEY_CONTROL},	

		{XK_a, AGIL_KEY_A}, {XK_s, AGIL_KEY_S}, {XK_d, AGIL_KEY_D},
		{XK_f, AGIL_KEY_F}, {XK_g, AGIL_KEY_G}, {XK_h, AGIL_KEY_H},
		{XK_j, AGIL_KEY_J}, {XK_k, AGIL_KEY_K}, {XK_l, AGIL_KEY_L},

		{XK_semicolon, AGIL_KEY_SEMICOLON},
		{XK_apostrophe, AGIL_KEY_APOSTROPHE}, {XK_grave, AGIL_KEY_GRAVE},
		{XK_Shift_L, AGIL_KEY_LEFTSHIFT},
		{XK_backslash, AGIL_KEY_BACKSLASH},

		{XK_z, AGIL_KEY_Z}, {XK_x, AGIL_KEY_X}, {XK_c, AGIL_KEY_C},
		{XK_v, AGIL_KEY_V}, {XK_b, AGIL_KEY_B}, {XK_n, AGIL_KEY_N},
		{XK_m, AGIL_KEY_M},

		{XK_comma, AGIL_KEY_COMMA}, {XK_period, AGIL_KEY_PERIOD},
		{XK_slash, AGIL_KEY_SLASH}, {XK_Shift_R, AGIL_KEY_RIGHTSHIFT},
		{XK_KP_Multiply, AGIL_KEY_KEYPADMULTIPLY},
		{XK_Alt_L, AGIL_KEY_ALT}, {XK_space, AGIL_KEY_SPACE},
		{XK_Caps_Lock, AGIL_KEY_CAPSLOCK},

		{XK_F1, AGIL_KEY_F1}, {XK_F2, AGIL_KEY_F2}, {XK_F3, AGIL_KEY_F3},
		{XK_F4, AGIL_KEY_F4}, {XK_F5, AGIL_KEY_F5}, {XK_F6, AGIL_KEY_F6},
		{XK_F7, AGIL_KEY_F7}, {XK_F8, AGIL_KEY_F8}, {XK_F9, AGIL_KEY_F9},
		{XK_F10, AGIL_KEY_F10},	

		{XK_Num_Lock, AGIL_KEY_NUMLOCK},
		{XK_Scroll_Lock, AGIL_KEY_SCROLLLOCK},

		/* NumPad with NumLock off */
		{XK_KP_Home, AGIL_KEY_CURSORUPLEFT},
		{XK_KP_Up, AGIL_KEY_CURSORUP},
		{XK_KP_Page_Up, AGIL_KEY_CURSORUPRIGHT},
		{XK_KP_Left, AGIL_KEY_CURSORLEFT},
		{XK_KP_Right, AGIL_KEY_CURSORRIGHT},
		{XK_KP_End, AGIL_KEY_CURSORDOWNLEFT},
		{XK_KP_Down, AGIL_KEY_CURSORDOWN},
		{XK_KP_Page_Down, AGIL_KEY_CURSORDOWNRIGHT},

		/* NumPad with NumLock on */
		{XK_KP_7, AGIL_KEY_CURSORUPLEFT}, {XK_KP_8, AGIL_KEY_CURSORUP},
		{XK_KP_9, AGIL_KEY_CURSORUPRIGHT}, {XK_KP_4, AGIL_KEY_CURSORLEFT},
		{XK_KP_5, AGIL_KEY_KEYPAD5}, {XK_KP_6, AGIL_KEY_CURSORRIGHT},
		{XK_KP_1, AGIL_KEY_CURSORDOWNLEFT},
		{XK_KP_2, AGIL_KEY_CURSORDOWN},
		{XK_KP_3, AGIL_KEY_CURSORDOWNRIGHT}, {XK_KP_0, AGIL_KEY_KEYPAD0},

		/* Cursor block keys added by me */
		{XK_Home, AGIL_KEY_CURSORUPLEFT},
		{XK_Up, AGIL_KEY_CURSORUP},
		{XK_Page_Up, AGIL_KEY_CURSORUPRIGHT},
		{XK_Left, AGIL_KEY_CURSORLEFT},
		{XK_Right, AGIL_KEY_CURSORRIGHT},
		{XK_End, AGIL_KEY_CURSORDOWNLEFT},
		{XK_Down, AGIL_KEY_CURSORDOWN}	,
		{XK_Page_Down, AGIL_KEY_CURSORDOWNRIGHT},	

		{XK_F11, AGIL_KEY_F11}, {XK_F12, AGIL_KEY_F12},
		{XK_KP_Enter, AGIL_KEY_ENTER},
		{XK_Control_R, AGIL_KEY_CONTROL},
		{XK_KP_Divide, AGIL_KEY_KEYPADDIVIDE},
		{XK_Sys_Req, AGIL_KEY_PRINTSCREEN}, {XK_Alt_R, AGIL_KEY_ALT},
		{XK_Pause, AGIL_KEY_BREAK},

		{0, 0}
	};

	while (XEventsQueued (x_display, QueuedAfterReading)) {
		XNextEvent (x_display, &ev);
		if ((ev.type != KeyPress) && (ev.type != KeyRelease))
			continue;
			
		if (ev.type == KeyPress) press = 1;
		else press = 0;
		
		XLookupString (&ev.xkey, buf, 1, &key, NULL);

		/* Kill shifted alpha keys */
		if ((key >= XK_A) && (key <= XK_Z))
			key += (XK_a - XK_A);

		/* Map this to a scancode */
		scancode = 0;
		for (i = 0; x_key_mapping[i].keysym; i++) {
/*			if (x_key_mapping[i].keysym == 0)
				break;*/
			if (x_key_mapping[i].keysym == key) {
				scancode = x_key_mapping[i].scancode;
				break;
			}
		}
		if (scancode == 0)
			continue;

		/* Give key to main event handler */
		event_kb_handler (scancode, press);
	}
}


void x11_death (int sig)
{
	printf("driver: X11: FATAL: Fatal signal recieved\n");
	exit (-1);
}


/* Sound driver stuff */
void * driver_sound_thread(void * arg)
{
	int i;
	SINT16 * buffer;

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	while(1) {
		if(!playing) {
			usleep(SOUND_DUR_USEC);
			continue;
		}
		buffer = sound_getbuffer();
	        for (i = BUFFER_BYTE_SIZE; i -= write (audio_fd, buffer, i); );
	}
}
