/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 *
 * Portions from LAGII, (C) 1999 XoXus
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "general.h"
#include "event.h"
#include "driver.h"
#include "text.h"
#include "object.h"
#include "status.h"
#include "vm.h"
#include "resource.h"
#include "decomp.h"

unsigned char keytable[128];
boolean events_initialized = false;
boolean event_catchkeys = false;
int event_caughtkey = 0;

struct event {	/* linked list: event queue */
	UINT8 code;
	struct event *next;
} *firstevent = NULL;

struct set_key { /* linked list: special keys, set with set.key() */
	int scancode;
	UINT8 eventcode;
	struct set_key *next;
} *first_setkey = NULL;


void event_movement_key(int scancode);

void event_init(void)
{
	if(events_initialized) return;
	
	atexit(event_flushqueue);
	atexit(event_flushkeys);

	memset(keytable, 0, 128);
	events_initialized = true;
	
	printf("event: Events initialized\n");
	return;
}


void event_flushqueue(void)
{
	struct event * current;
	
	while(firstevent) {
		current = firstevent->next;
		free(firstevent);
		firstevent = current;
	}
	printf("event: Event queue flushed\n");
}


void event_add(UINT8 code) /* making endofqueue global would speed this up */
{
	struct event * endofqueue = NULL;

	if(!firstevent) {
		firstevent = (struct event *)malloc(sizeof(struct event));
		endofqueue = firstevent;
	} else {
		endofqueue = firstevent;
		while(endofqueue->next)
			endofqueue = endofqueue->next;
		endofqueue->next = (struct event *)malloc(sizeof(struct event));
		endofqueue = endofqueue->next;
	}
	endofqueue->next = NULL;
	endofqueue->code = code;
}


boolean event_check(UINT8 code)
{
	struct event * current = firstevent;
	struct event * tmp;
	
	if(!firstevent) return false;
	
	if(firstevent->code == code) {
		current = firstevent->next;
		free(firstevent);
		firstevent = current;
		return true;
	}
	
	while(current->next) {
		if(current->next->code == code) {
			tmp = current->next;
			current->next = tmp->next;
			free(tmp);
			return true;
		}
		current = current->next;
	}
	
	return false;
}


void event_flushkeys(void)
{
	struct set_key * current;
	
	while(first_setkey) {
		current = first_setkey->next;
		free(first_setkey);
		first_setkey = current;
	}
	printf("event: Special keys flushed\n");
}


void event_addkey(int scancode, UINT8 eventcode)
{
	struct set_key * endofqueue = NULL;

	if(!first_setkey) {
		first_setkey = (struct set_key *)malloc(sizeof(struct set_key));
		endofqueue = first_setkey;
	} else {
		endofqueue = first_setkey;
		while(endofqueue->next)
			endofqueue = endofqueue->next;
		endofqueue->next = (struct set_key *)malloc(sizeof(struct set_key));
		endofqueue = endofqueue->next;
	}
	endofqueue->next = NULL;
	endofqueue->scancode = scancode;
	endofqueue->eventcode = eventcode;
}


boolean event_checkkey(int scancode)
{
	struct set_key * current = first_setkey;
	
	while(current) {
		if(current->scancode == scancode) {
			event_add(current->eventcode);
			return true;
		}
		current = current->next;
	}
	
	return false;
}


boolean event_checkkey_noexec(int scancode)
{
	struct set_key * current = first_setkey;
	
	while(current) {
		if(current->scancode == scancode) return true;
		current = current->next;
	}
	
	return false;
}


int event_waitforkey(void)
{
	event_caughtkey = 0;
	event_catchkeys = true;

	while(!event_caughtkey) driver_kb_update();

	event_catchkeys = false;
		
	return event_caughtkey;
}


void event_kb_handler(int scancode, int press)
{
	int i;
	
	struct mapping { 	/* Table to lookup ASCII values */
		int in, out;
	} asciimap[] = {
		{AGIL_KEY_1, '1'}, {AGIL_KEY_2, '2'}, {AGIL_KEY_3, '3'},
		{AGIL_KEY_4, '4'}, {AGIL_KEY_5, '5'}, {AGIL_KEY_6, '6'},
		{AGIL_KEY_7, '7'}, {AGIL_KEY_8, '8'}, {AGIL_KEY_9, '9'},
		{AGIL_KEY_0, '0'},
		
		{AGIL_KEY_A, 'a'}, {AGIL_KEY_B, 'b'}, {AGIL_KEY_C, 'c'},
		{AGIL_KEY_D, 'd'}, {AGIL_KEY_E, 'e'}, {AGIL_KEY_F, 'f'},
		{AGIL_KEY_G, 'g'}, {AGIL_KEY_H, 'h'}, {AGIL_KEY_I, 'i'},
		{AGIL_KEY_J, 'j'}, {AGIL_KEY_K, 'k'}, {AGIL_KEY_L, 'l'},
		{AGIL_KEY_M, 'm'}, {AGIL_KEY_N, 'n'}, {AGIL_KEY_O, 'o'},
		{AGIL_KEY_P, 'p'}, {AGIL_KEY_Q, 'q'}, {AGIL_KEY_R, 'r'},
		{AGIL_KEY_S, 's'}, {AGIL_KEY_T, 't'}, {AGIL_KEY_U, 'u'},
		{AGIL_KEY_V, 'v'}, {AGIL_KEY_W, 'w'}, {AGIL_KEY_X, 'x'},
		{AGIL_KEY_Y, 'y'}, {AGIL_KEY_Z, 'z'},
		
		{AGIL_KEY_COMMA, ','}, {AGIL_KEY_PERIOD, '.'},
		{AGIL_KEY_SLASH, '/'}, {AGIL_KEY_MINUS, '-'},
		{AGIL_KEY_EQUAL, '='}, {AGIL_KEY_SEMICOLON, ';'},
		{AGIL_KEY_GRAVE, '`'}, {AGIL_KEY_BRACKET_LEFT, '['},
		{AGIL_KEY_BRACKET_RIGHT, ']'}, {AGIL_KEY_APOSTROPHE, '\''},
		
		{AGIL_KEY_ENTER, 0x0D}, {AGIL_KEY_BACKSPACE, '\b'},
		{AGIL_KEY_TAB, '\t'}, {AGIL_KEY_BACKSLASH, '\\'},
		{AGIL_KEY_SPACE, ' '}, {AGIL_KEY_ESCAPE, 27}, 
		
		{0, 0}
	};

	struct mapping shiftmap[] = {	/* Table to lookup shifted values */
		{'`', '~'}, {'1', '!'}, {'2', '@'}, {'3', '#'}, {'4', '$'},
		{'5', '%'}, {'6', '^'}, {'7', '&'}, {'8', '*'}, {'9', '('},
		{'0', ')'}, {'-', '_'}, {'=', '+'},
		
		{'q', 'Q'}, {'w', 'W'}, {'e', 'E'}, {'r', 'R'}, {'t', 'T'},
		{'y', 'Y'}, {'u', 'U'}, {'i', 'I'}, {'o', 'O'}, {'p', 'P'}, 
		{'[', '{'}, {']', '}'}, {'\\', '|'},
		
		{'a', 'A'}, {'s', 'S'}, {'d', 'D'}, {'f', 'F'}, {'g', 'G'}, 
		{'h', 'H'}, {'j', 'J'}, {'k', 'K'}, {'l', 'L'}, {';', ':'}, 
		{'\'', '"'},
		
		{'z', 'Z'}, {'x', 'X'}, {'c', 'C'}, {'v', 'V'}, {'b', 'B'}, 
		{'n', 'N'}, {'m', 'M'}, {',', '<'}, {'.', '>'}, {'/', '?'},

		{0, 0}
	};	/* This shift table may well be overkill, I don't know of
		   any AGI games that require these characters */

	if(!events_initialized) return;
	if(keytable[scancode] == press) return; /* keys held down */
	keytable[scancode] = press;
	if(!press) return;			/* ignore releases */

	if((scancode == AGIL_KEY_LEFTSHIFT) ||
	   (scancode == AGIL_KEY_RIGHTSHIFT) ||
	   (scancode == AGIL_KEY_CONTROL) ||
	   (scancode == AGIL_KEY_ALT)) return;
	   
	if(scancode == AGIL_KEY_BREAK) {
	   	decomp_main(VAR[0]);
		return;
	}
	   
	/* shift ALT+key pairs and function keys up 1 byte */
	if(((scancode >= AGIL_KEY_F1) && (scancode <= AGIL_KEY_F10)) ||
	   (keytable[AGIL_KEY_ALT])) scancode <<= 8;
	   
	/* map to ascii values */
	for(i = 0; asciimap[i].out != 0; i++) 
		if(scancode == asciimap[i].in) {
			scancode = asciimap[i].out;
			break;
		}
	
	/* offset CTRL+key pairs */
	if((keytable[AGIL_KEY_CONTROL]) &&
	   (scancode >= 'a') && (scancode <= 'z')) 
	   	scancode -= 'a' - 1; 
	
	VAR[19] = scancode & 0xFF;
	
	/* keys wanted by another function */
	if(event_catchkeys) {
		event_caughtkey = scancode;
		return;
	}
	
	/* All translations done. now what to do with scancode */

	if(!text_special_input)
		if(event_checkkey(scancode)) return; /* key is set with set.key() */
	
	if((scancode == AGIL_KEY_CURSORUP) ||
	   (scancode == AGIL_KEY_CURSORUPRIGHT) ||
	   (scancode == AGIL_KEY_CURSORRIGHT) ||
	   (scancode == AGIL_KEY_CURSORDOWNRIGHT) ||
	   (scancode == AGIL_KEY_CURSORDOWN) ||
	   (scancode == AGIL_KEY_CURSORDOWNLEFT) ||
	   (scancode == AGIL_KEY_CURSORLEFT) ||
	   (scancode == AGIL_KEY_CURSORUPLEFT) ||
	   (scancode == AGIL_KEY_KEYPAD5)) {
		event_movement_key(scancode);
		return;
	}

	/* if shift held then map shifted values */
	if(keytable[AGIL_KEY_LEFTSHIFT] || keytable[AGIL_KEY_RIGHTSHIFT])
		for(i = 0; shiftmap[i].out; i++)
			if(scancode == shiftmap[i].in) {
				scancode = shiftmap[i].out;
				break;
			}

	text_new_key(scancode);	/* send to text handler */
	
	return;
}


void event_movement_key(int scancode) 
{
	int newdir = 0;
	
	if(status.program_control) return;
	
	switch(scancode) {
		case AGIL_KEY_CURSORUP:
			newdir = 1;
			break;
		case AGIL_KEY_CURSORUPRIGHT:
			newdir = 2;
			break;
		case AGIL_KEY_CURSORRIGHT:
			newdir = 3;
			break;
		case AGIL_KEY_CURSORDOWNRIGHT:
			newdir = 4;
			break;
		case AGIL_KEY_CURSORDOWN:
			newdir = 5;
			break;
		case AGIL_KEY_CURSORDOWNLEFT:
			newdir = 6;
			break;
		case AGIL_KEY_CURSORLEFT:
			newdir = 7;
			break;
		case AGIL_KEY_CURSORUPLEFT:
			newdir = 8;
			break;
		case AGIL_KEY_KEYPAD5:
			newdir = 0;
			break;
	}

	if(newdir == object[0].direction) object[0].direction = 0;
	else object[0].direction = newdir;
	VAR[6] = object[0].direction;
	object[0].flags |= VO_MOTION;
}
