/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 *
 * Portions from AGI spec.
 * Portions from LAGII, (C) 1999 XoXus
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "general.h"
#include "graphics.h"
#include "font.h"

UINT8 *backvisbuffer = NULL, *workvisbuffer = NULL;
UINT8 *backpribuffer = NULL, *workpribuffer = NULL;
UINT8 *textmodebuffer = NULL; 
	/* can be used as a temp buffer when in graphics mode */
	
UINT8 *video_buffer = NULL; /* point this to workvisbuffer or textmodebuffer */
UINT8 font[96][8];

boolean pripen = false, vispen = false;
UINT8 pricolor = 4, viscolor = 15;


int graph_malloc_screens(void)
{
	if(atexit(graph_free_screens)) return -1;

	if(!(backvisbuffer = (UINT8 *)malloc(320 * 168))) return -2;
	if(!(backpribuffer = (UINT8 *)malloc(320 * 168))) return -2;
	
	if(!(workpribuffer = (UINT8 *)malloc(320 * 200))) return -2;
	if(!(workvisbuffer = (UINT8 *)malloc(320 * 200))) return -2;
	
	if(!(textmodebuffer = (UINT8 *)malloc(320 * 200))) return -2;
	
	memset(backvisbuffer, 0, 320 * 168);
	memset(backpribuffer, 4, 320 * 168);
	memset(workvisbuffer, 0, 320 * 200);
	memset(workpribuffer, 4, 320 * 200);
	memset(textmodebuffer, 0, 320 * 200);	
	
	printf("graphics: Allocated screens\n");
	return 0;
}


void graph_free_screens(void)
{
	if(backvisbuffer) free(backvisbuffer);	backvisbuffer = NULL;
	if(backpribuffer) free(backpribuffer);	backpribuffer = NULL;
	if(workvisbuffer) free(workvisbuffer);	workvisbuffer = NULL;
	if(workpribuffer) free(workpribuffer);	workpribuffer = NULL;
	if(textmodebuffer) free(textmodebuffer);	textmodebuffer = NULL;

	printf("graphics: Released screens\n");	
	return;
}


void graph_loadfont()
{
	int i;
	
	for(i = 0; i < 96; i++)
		memcpy(font[i], &agi_font[i * 8], 8);
		
	return;
}


void graph_drawchar(UINT8 * buffer, int x, int y, char ch, UINT8 fg, SINT8 bg)
{
	int i, j;
	UINT8 c;
	
	if((ch < 1) || (ch < 32)) return; /* illegal char */
	
	for(j = 0; j < 8; j++) {
		c = font[ch - 32][j];
		for(i = 0; i < 8; i++) {
			if(c & (1 << i))
				buffer[((y + j) * 320) + x + i] = fg;
			else
				if(bg >= 0)
					buffer[((y + j) * 320) + x + i] = bg;
		}
	}
}


void graph_drawstring(UINT8 * buffer, int x, int y, char * str, UINT8 fg, SINT8 bg)
{
	int i;
	
	for(i = 0; str[i]; i++) {
		if(str[i] == '\n') {
			x = 0;
			y += 8;
			continue;
		}
		graph_drawchar(buffer, x, y, str[i], fg, bg);
		x += 8;
	}
}


void graph_drawbox(UINT8 * buffer, int x, int y, int w, int h, UINT8 bg, UINT8 border)
{
	int i;

	for(i = 0; i < h; i++) {
		memset(buffer + (320 * (y + i) + x), bg, w);
/*		memset(workpribuffer + (320 * (y + i) + x), 16, w);*/
	}

	if(bg == border) return;
		
	memset(buffer + (320 * (y + 1) + x + 2), border, w - 4);
	memset(buffer + (320 * (y + h - 2) + x + 2), border, w - 4);

	for(i = 2; i < h - 2; i++) {
		buffer[320 * (y + i) + x + 2] = border;
		buffer[320 * (y + i) + x + 3] = border;
		buffer[320 * (y + i) + x + w - 3] = border;
		buffer[320 * (y + i) + x + w - 4] = border;
	}
}

/* Message box moved to message.c */


/* Picture drawing primitives */

/* pset() is #define'd in graphics.h */
void psetp(int x, int y)
{
	if(pripen) {
		if((x < 0) || (x > 159)) return;
		if((y < 0) || (y > 167)) return;

		pixel(backpribuffer, x * 2, y) = pricolor;
		pixel(backpribuffer, (x * 2) + 1, y) = pricolor;
	}
}


void psetv(int x, int y)
{
	if(vispen) {
		if((x < 0) || (x > 159)) return;
		if((y < 0) || (y > 167)) return;

		pixel(backvisbuffer, x * 2, y) = viscolor;
		pixel(backvisbuffer, (x * 2) + 1, y) = viscolor;
	}
}


/* The #define efforts of these didn't work?? Beats me! These do! */
UINT8 pgetv(int x, int y) {return pixel(backvisbuffer, x * 2, y);}
UINT8 pgetp(int x, int y) {return pixel(backpribuffer, x * 2, y);}


/* Dumb fill stack stuff */
struct point_s {
	int x, y;
} fill_point[1001];
int ci = 0, ei = 0;

void fill_push(int x, int y) 
{
	fill_point[ei].x = x;
	fill_point[ei].y = y;
	ei++;
	ei %= 1000;
}

void fill_pop(int *x, int *y) 
{
	*x = fill_point[ci].x;
	*y = fill_point[ci].y;
	ci++;
	ci %= 1000;
}

void fillv(int x, int y)
{
	/* Sanity checks */
	if(!vispen) return;
	if(viscolor == 15) return;
	
	fill_push(x, y);
		
	while(ci != ei) {
		fill_pop(&x, &y);
				
		if((x < 0) || (x > 159)) continue;
		if((y < 0) || (y > 167)) continue;
		if(pgetv(x, y) != 15) continue;
	
		psetv(x, y);
		psetp(x, y);

		if(pgetv(x - 1, y) == 15)
			fill_push(x - 1, y);
		if(pgetv(x, y - 1) == 15)
			fill_push(x, y - 1);

		if(pgetv(x + 1, y) == 15)
			fill_push(x + 1, y);
		if(pgetv(x, y + 1) == 15)
			fill_push(x, y + 1);
	}
	ci = ei = 0;
}


void fillp(int x, int y)
{
	/* Sanity checks */
	if(!pripen) return;
	if(pricolor == 4) return;
	
	fill_push(x, y);
		
	while(ci != ei) {
		fill_pop(&x, &y);
				
		if((x < 0) || (x > 159)) continue;
		if((y < 0) || (y > 167)) continue;
		if(pgetp(x, y) != 4) continue;
	
		psetp(x, y);

		if(pgetp(x - 1, y) == 4)
			fill_push(x - 1, y);
		if(pgetp(x, y - 1) == 4)
			fill_push(x, y - 1);

		if(pgetp(x + 1, y) == 4)
			fill_push(x + 1, y);
		if(pgetp(x, y + 1) == 4)
			fill_push(x, y + 1);
	}
	ci = ei = 0;
}


/* Sierra style round(): from AGI spec */
int sround(float aNumber, float dirn)
{
	if (dirn < 0)
		return ((aNumber - floor(aNumber) <= 0.501)? floor(aNumber) : ceil(aNumber));
	return ((aNumber - floor(aNumber) < 0.499)? floor(aNumber) : ceil(aNumber));
}


/* Sierra line() clone: from AGI spec */
void line(int x1, int y1, int x2, int y2)
{
	int height, width;
	float x, y, addX, addY;

	height = (y2 - y1);
	width = (x2 - x1);
	addX = (height == 0 ? height:(float)width/abs(height));
	addY = (width == 0 ? width:(float)height/abs(width));

	if (abs(width) > abs(height)) {
		y = y1;
		addX = (width == 0? 0 : (width/abs(width)));
		for (x = x1; x != x2; x += addX) {
			pset(sround(x, addX), sround(y, addY));
			y += addY;
		}
		pset(x2, y2);
	} else {
		x = x1;
		addY = (height == 0? 0 : (height/abs(height)));
		for (y = y1; y != y2; y += addY) {
			pset(sround(x, addX), sround(y, addY));
			x += addX;
		}
		pset(x2, y2);
	}
}


/* Get real priority of AGI pixel : useful function */
UINT8 graph_get_real_pri(int x, int y)
{
	UINT8 *p;
	UINT8 tmp;
	
	p = workpribuffer + 2560;
	if(y <= 0) return 4;


	do {
		tmp = pixel(p, x * 2, y);
		if(tmp < 3) y++;
		if(tmp == 3) tmp = 4;
	} while(tmp < 4);
	
	return tmp;
}
