/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

#ifndef __VM_H__
#define __VM_H__

#include "general.h"
#include "resource.h"

extern boolean vm_reset;
extern boolean new_room;

void vm_init(void);
void vm_exec(UINT8 logicno);
char * vm_parsemessage(resource_t *l, UINT8 messageno);

void vm_fix_opcode_table(void);

void vm_push(UINT8 data);
UINT8 vm_pop(void);

extern UINT8 VAR[];
volatile extern boolean FLAG[];
extern char STRING[24][40];

typedef struct vm_s {	/* similar to LAGII logic_vm_t */
	resource_t * res;
	UINT16 IP;
	UINT8 * code;	
} vm_t;

struct opcode_s {
	char name[20];
	UINT8 args;
	void (*execute)(vm_t *); 
};

extern struct opcode_s action_opcode[];
extern struct opcode_s test_opcode[];

#define AGI_OPCODE(x)	\
	void x (vm_t *)

/* Test opcodes */
AGI_OPCODE(to_equaln);
AGI_OPCODE(to_equalv);
AGI_OPCODE(to_lessn);
AGI_OPCODE(to_lessv);
AGI_OPCODE(to_greatern);
AGI_OPCODE(to_greaterv);
AGI_OPCODE(to_isset);
AGI_OPCODE(to_issetv);
AGI_OPCODE(to_has);
AGI_OPCODE(to_obj_in_room);
AGI_OPCODE(to_has);
AGI_OPCODE(to_controller);
AGI_OPCODE(to_havekey);
AGI_OPCODE(to_obj_in_box);
AGI_OPCODE(to_center_posn);
AGI_OPCODE(to_right_posn);
	
/* Action opcodes */
AGI_OPCODE(ao_increment);
AGI_OPCODE(ao_decrement);
AGI_OPCODE(ao_assignn);
AGI_OPCODE(ao_assignv);
AGI_OPCODE(ao_addn);
AGI_OPCODE(ao_addv);
AGI_OPCODE(ao_subn);
AGI_OPCODE(ao_subv);
AGI_OPCODE(ao_lindirectv);
AGI_OPCODE(ao_rindirect);
AGI_OPCODE(ao_lindirectn);
AGI_OPCODE(ao_set);
AGI_OPCODE(ao_reset);
AGI_OPCODE(ao_toggle);
AGI_OPCODE(ao_setv);
AGI_OPCODE(ao_resetv);
AGI_OPCODE(ao_togglev);
AGI_OPCODE(ao_new_room);
AGI_OPCODE(ao_new_room_v);
AGI_OPCODE(ao_load_logic);
AGI_OPCODE(ao_load_logic_v);
AGI_OPCODE(ao_call);
AGI_OPCODE(ao_call_v);
AGI_OPCODE(ao_load_pic);
AGI_OPCODE(ao_draw_pic);
AGI_OPCODE(ao_show_pic);
AGI_OPCODE(ao_discard_pic);
AGI_OPCODE(ao_overlay_pic);
AGI_OPCODE(ao_show_pri_screen);
AGI_OPCODE(ao_load_view);
AGI_OPCODE(ao_load_view_v);
AGI_OPCODE(ao_discard_view);
AGI_OPCODE(ao_animate_obj);
AGI_OPCODE(ao_unanimate_all);
AGI_OPCODE(ao_draw);
AGI_OPCODE(ao_erase);
AGI_OPCODE(ao_position);
AGI_OPCODE(ao_position_v);
AGI_OPCODE(ao_get_posn);
AGI_OPCODE(ao_reposition);
AGI_OPCODE(ao_set_view);
AGI_OPCODE(ao_set_view_v);
AGI_OPCODE(ao_set_loop);
AGI_OPCODE(ao_set_loop_v);
AGI_OPCODE(ao_fix_loop);
AGI_OPCODE(ao_release_loop);
AGI_OPCODE(ao_set_cel);
AGI_OPCODE(ao_set_cel_v);
AGI_OPCODE(ao_last_cel);
AGI_OPCODE(ao_current_cel);
AGI_OPCODE(ao_current_loop);
AGI_OPCODE(ao_current_view);
AGI_OPCODE(ao_number_of_loops);
AGI_OPCODE(ao_set_priority);
AGI_OPCODE(ao_set_priority_v);
AGI_OPCODE(ao_release_priority);
AGI_OPCODE(ao_get_priority);
AGI_OPCODE(ao_stop_update);
AGI_OPCODE(ao_start_update);
AGI_OPCODE(ao_force_update);
AGI_OPCODE(ao_ignore_horizon);
AGI_OPCODE(ao_observe_horizon);
AGI_OPCODE(ao_set_horizon);
AGI_OPCODE(ao_object_on_water);
AGI_OPCODE(ao_object_on_land);
AGI_OPCODE(ao_object_on_anything);
AGI_OPCODE(ao_ignore_objs);
AGI_OPCODE(ao_observe_objs);
AGI_OPCODE(ao_distance);
AGI_OPCODE(ao_stop_cycling);
AGI_OPCODE(ao_start_cycling);
AGI_OPCODE(ao_normal_cycle);
AGI_OPCODE(ao_end_of_loop);
AGI_OPCODE(ao_reverse_cycle);
AGI_OPCODE(ao_reverse_loop);
AGI_OPCODE(ao_cycle_time);
AGI_OPCODE(ao_stop_motion);
AGI_OPCODE(ao_start_motion);
AGI_OPCODE(ao_step_size);
AGI_OPCODE(ao_step_time);
AGI_OPCODE(ao_move_obj);
AGI_OPCODE(ao_move_obj_v);
AGI_OPCODE(ao_follow_ego);
AGI_OPCODE(ao_wander);
AGI_OPCODE(ao_normal_motion);
AGI_OPCODE(ao_get_dir);
AGI_OPCODE(ao_set_dir);
AGI_OPCODE(ao_ignore_blocks);
AGI_OPCODE(ao_observe_blocks);
AGI_OPCODE(ao_block);
AGI_OPCODE(ao_unblock);
AGI_OPCODE(ao_get);
AGI_OPCODE(ao_get_v);
AGI_OPCODE(ao_drop);
AGI_OPCODE(ao_put);
AGI_OPCODE(ao_put_v);
AGI_OPCODE(ao_get_room);
AGI_OPCODE(ao_load_sound);
AGI_OPCODE(ao_sound);
AGI_OPCODE(ao_stop_sound);
AGI_OPCODE(ao_print);
AGI_OPCODE(ao_print_v);
AGI_OPCODE(ao_display);
AGI_OPCODE(ao_display_v);
AGI_OPCODE(ao_clear_lines);
AGI_OPCODE(ao_test_screen);
AGI_OPCODE(ao_graphics);
AGI_OPCODE(ao_set_cursor_char);
AGI_OPCODE(ao_set_text_attribute);
AGI_OPCODE(ao_shake_screen);
AGI_OPCODE(ao_configure_screen);
AGI_OPCODE(ao_status_line_on);
AGI_OPCODE(ao_status_line_off);
AGI_OPCODE(ao_get_string);
AGI_OPCODE(ao_set_string);
AGI_OPCODE(ao_word_to_string);
AGI_OPCODE(ao_parse);
AGI_OPCODE(ao_get_num);
AGI_OPCODE(ao_prevent_input);
AGI_OPCODE(ao_accept_input);
AGI_OPCODE(ao_set_key);
AGI_OPCODE(ao_add_to_pic);
AGI_OPCODE(ao_add_to_pic_v);
AGI_OPCODE(ao_status);
AGI_OPCODE(ao_save_game);
AGI_OPCODE(ao_restore_game);
AGI_OPCODE(ao_init_disk);
AGI_OPCODE(ao_restart_game);
AGI_OPCODE(ao_show_obj);
AGI_OPCODE(ao_random);
AGI_OPCODE(ao_program_control);
AGI_OPCODE(ao_player_control);
AGI_OPCODE(ao_obj_status_v);
AGI_OPCODE(ao_quit);
AGI_OPCODE(ao_show_mem);
AGI_OPCODE(ao_pause);
AGI_OPCODE(ao_echo_line);
AGI_OPCODE(ao_cancel_line);
AGI_OPCODE(ao_init_joy);
AGI_OPCODE(ao_toggle_monitor);
AGI_OPCODE(ao_version);
AGI_OPCODE(ao_script_size);
AGI_OPCODE(ao_set_game_id);
AGI_OPCODE(ao_log);
AGI_OPCODE(ao_set_scan_start);
AGI_OPCODE(ao_reset_scan_start);
AGI_OPCODE(ao_trace_on);
AGI_OPCODE(ao_trace_info);
AGI_OPCODE(ao_print_at);
AGI_OPCODE(ao_print_at_v);
AGI_OPCODE(ao_discard_view_v);
AGI_OPCODE(ao_clear_text_rect);
AGI_OPCODE(ao_set_upper_left);
AGI_OPCODE(ao_set_menu);
AGI_OPCODE(ao_set_menu_item);
AGI_OPCODE(ao_submit_menu);
AGI_OPCODE(ao_enable_item);
AGI_OPCODE(ao_disable_item);
AGI_OPCODE(ao_menu_input);
AGI_OPCODE(ao_show_obj_v);
AGI_OPCODE(ao_open_dialogue);
AGI_OPCODE(ao_close_dialogue);
AGI_OPCODE(ao_muln);
AGI_OPCODE(ao_mulv);
AGI_OPCODE(ao_divn);
AGI_OPCODE(ao_divv);
AGI_OPCODE(ao_close_window);

#endif
