/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 *
 * I wrote this myself, but apparently the algorithm is under patent.
 * Send me more info.
 */

#include <stdio.h>
#include <stdlib.h>

#include "general.h"
#include "lzw.h"

static int lzw_wordlen = 9;
static UINT32 lzw_bitqueue = 0;
static int lzw_queuepos = 0;

static UINT8 lzw_stack[100];	/* 100 should be ample */
static int lzw_SP = 0;

static struct lzw_word_s {
	int prefixcode;
	UINT8 append;
} lzw_word[2048];	/* does not allow for 12 bit words */
static int lzw_nextword = 258;


int lzw_getword(FILE * stream)
{
	unsigned char ch;
	int ret;

	/* Read in a sufficient number of bits */
	while(lzw_queuepos < lzw_wordlen) { 
		ch = fgetc(stream);
		lzw_bitqueue |= ch << lzw_queuepos;
		lzw_queuepos += 8;
	}
	
	/* Return last (lzw_wordlen) bits */
	ret = lzw_bitqueue % (1 << lzw_wordlen);
	lzw_bitqueue >>= lzw_wordlen;
	lzw_queuepos -= lzw_wordlen;

	return ret;
}


/* Put translated code on lzw_stack */
int lzw_translatecode(int codeword)
{
	if(codeword >= lzw_nextword) {
		printf("lzw: FATAL: Illegal code word encountered\n");
		exit(-1);
	}
	
	while(codeword > 257) {
		lzw_stack[lzw_SP] = lzw_word[codeword].append; lzw_SP++;
		codeword = lzw_word[codeword].prefixcode;
	}
	lzw_stack[lzw_SP] = codeword; lzw_SP++; /* PUSH char */

	if(lzw_SP == 100) {
		printf("lzw: FATAL: Stack overflow\n");
		exit(-1);
	}

	return 0;
}


void lzw_inflate(UINT8 * dest, FILE * stream) 
{
	int oldcode, newcode;
	int outpos = 0;
	UINT8 ch = 0;

	/* clean up globals */
	lzw_bitqueue = 0;
	lzw_queuepos = 0;
	lzw_SP = 0;
	lzw_wordlen = 9;
	lzw_nextword = 258;
	
	oldcode = 2048; /* No oldcode */

	while(1) {
		newcode = lzw_getword(stream);
		if(newcode == 256) {	/* Reset LZW */
			lzw_wordlen = 9;
			lzw_nextword = 258;
			oldcode = 2048;
			continue;
		}
		if(newcode == 257) return; /* End of stream */
		
		if(newcode == lzw_nextword) { /* EXCEPTION */
			lzw_stack[lzw_SP] = ch; lzw_SP++;
			if(lzw_translatecode(oldcode)) return;
		} else	if(lzw_translatecode(newcode)) return;

		ch = lzw_stack[lzw_SP - 1]; /* get first char of newcode */

		do {		/* Output string from stack */
			lzw_SP--;
			dest[outpos] = lzw_stack[lzw_SP]; outpos++;
		} while (lzw_SP > 0);

		if(oldcode != 2048) { /* Oldcode exists, not reset */
			/* Add new code */
			lzw_word[lzw_nextword].prefixcode = oldcode;
			lzw_word[lzw_nextword].append = ch;
			/* Increase word size if required && allowed */
			if(((lzw_nextword++) + 1) >> lzw_wordlen) {
				if(lzw_wordlen < 11) lzw_wordlen++;
			}
		}

		oldcode = newcode;
	}
}
