/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

#ifndef TARGET_SDL
#define CYCLE_TIME_LIMIT 30
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#ifdef TARGET_SDL
#include "SDL.h"
#endif

#include "general.h"
#include "driver.h"
#include "event.h"
#include "graphics.h"
#include "menu.h"
#include "text.h"
#include "resource.h"
#include "gameid.h"
#include "picture.h"
#include "vm.h"
#include "status.h"
#include "object.h"
#include "view.h"
#include "sound.h"
#include "save.h"
#include "sound.h"

void update_status(void);
void update_clock(void);

#ifdef CYCLE_TIME_LIMIT
#include <signal.h>

void cycle_timeup(int sig) 
{
	printf("main: FATAL: Cycle time limit exceeded\n");
	exit(-2);
}
#endif


int initialize(void) /* everything called here should register with atexit() */
{
#ifdef WIN32
	/* Initialize console on Windoze */
	freopen("CON", "w", stdout);
	freopen("CON", "w", stderr);
#endif
	printf("\n======== %s %s, (C) 2001, 2002 RaDIaT1oN ========\n", TITLE, VERSION);

	id_game();
	if(!id) {
		printf("main: init: Game ID failed, this should not be possible\n");
		return -1;
	}

	if(driver_init()) {
		printf("main: init: Failed to initialize console driver!\n");
		return -1;
	}

	vm_fix_opcode_table();
	
	event_init();

	if(graph_malloc_screens()) {
		printf("main: init: failed to allocate memory for screens\n");
		return -1;
	}
	video_buffer = workvisbuffer;
	graph_loadfont();

	res_init(); /* res_init() exits if fatal errors */
	if(!LOGIC[0]) {
		printf("main: init: LOGIC.0 nonexistant\n");
		return -1;
	}
	res_load(LOGIC[0]);
	if(!LOGIC[0]->data) {
		printf("main: init: failed to load LOGIC.0\n");
		return -1;
	}

	text_init();	
	res_load_words();
	res_load_objects();

	sound_init();
	
	atexit(menu_flush); /* menu has no init function */

#ifdef CYCLE_TIME_LIMIT
	signal(SIGALRM, cycle_timeup);
#endif

#ifdef UNIX
	{
		char dir[31];
		snprintf(dir, 30, "%s/.agil", (char *)getenv("HOME"));
		mkdir(dir, 0x1ED);
		snprintf(savedir, 30, "%s/.agil/%s", (char *)getenv("HOME"), id->id);
		mkdir(savedir, 0x1ED);
	}
#else
	strcpy(savedir, gamedir);
#endif
	srand(time(NULL));
	return 0;
}


int main(int argc, char ** argv) 
{
	int i;
	int time, oldtime, interval, delaytime;
	boolean obj_drawn = true;
	int oldscore = 255;
	boolean oldsound = true;
	
	if(argc > 2) {
		printf("USAGE: %s [gamedir]\n", argv[0]);
		return -2;
	}
	if(argc == 2) {
		int i;
		strncpy(gamedir, argv[1], 38);
		for(i = 0; i < strlen(gamedir); i++)
			if(gamedir[i] == '\\') gamedir[i] = '/';
		if(gamedir[strlen(gamedir) - 1] != '/') {
			gamedir[strlen(gamedir) + 1] = 0;
			gamedir[strlen(gamedir)] = '/';
		}
	}
	
	if(initialize()) {
		printf("main: FATAL: initialization failed\n\n");
		return -1;
	}

	memset(object, 0, sizeof(view_obj) * 255);
	
	vm_init();	/* Reset FLAGS, VARS, and STRINGS */
	
	oldtime = (int)clock();
	
	while(1) {	/* INTERPRETER WORK CYCLE */
/*		printf("main: START OF INTERPRETER WORK CYCLE\n");*/
		if(VAR[10] == 0) VAR[10] = 1;
		time = (int)clock();
		interval = time - oldtime;
#ifndef TARGET_SDL
		delaytime = (50000 * VAR[10]) - (interval * (1000000 / CLOCKS_PER_SEC)) ;
		if(delaytime > 0)
			usleep(delaytime);
#else
		delaytime = (50 * VAR[10]) - (interval * (1000 / CLOCKS_PER_SEC)) ;
		if(delaytime > 0)
			SDL_Delay(delaytime);
#endif
		oldtime = (int)clock();
		update_clock();

		
#ifdef CYCLE_TIME_LIMIT
		alarm(CYCLE_TIME_LIMIT);
#endif

		FLAG[2] = false;
		FLAG[4] = false;
		VAR[9] = 0;

		/* update view objects */
		for(i = 254; i >= 0; i--) {
			if(object[i].flags & VO_ANIMATE)
				if((object[i].flags & (VO_UPDATE | VO_DRAWN)) ==
				   (VO_UPDATE | VO_DRAWN))
					object_update(&object[i]);
		}


		/* update status line */
		if(status.show_status_line && ((oldscore != VAR[3]) || (oldsound != FLAG[9]))) {
			update_status();
			oldscore = VAR[3];
			oldsound = FLAG[9];
		}
		text_new_key(0); /* redraw input line */

		driver_kb_update();

		do {
			new_room = false;
			vm_exec(0); /* execute LOGIC.0 */
		} while(new_room);

		object[0].direction = VAR[6];
		
		VAR[5] = 0;
		VAR[4] = 0;

		if(FLAG[5] && (~object[0].flags & VO_AUTO_MOVE)) status.program_control = false;
		FLAG[5] = false;

		driver_showbuffer(video_buffer);

		/* erase view objects */
		for(i = 0; i < 255; i++) {
			if(object[i].flags & VO_ON_SCREEN)
				view_erase(&object[i]);
		}

		FLAG[1] = false;
		for(i = 254; i >= 0; i--) {
			if((object[i].flags & VO_DRAWN) &&
			   (object[i].flags & VO_ANIMATE)) {
				obj_drawn = view_draw(&object[i]);
				if((i == 0) && !obj_drawn) FLAG[1] = true;
			}
		}
	}
}

void update_status(void)
{
		char tmp[25];

		memset(video_buffer, 15, 320 * 8);
		sprintf(tmp, "Score:%d of %d", VAR[3], VAR[7]);
		graph_drawstring(video_buffer, 8, 0, tmp, 0, 15);
		if(FLAG[9] && sound_enabled)
			graph_drawstring(video_buffer, 240, 0, "Sound:on", 0, 15);
		else
			graph_drawstring(video_buffer, 240, 0, "Sound:off", 0, 15);
}

void update_clock(void)
{
	static time_t ticks = 0;

	if(!ticks) {
		time(&ticks); 
		return;
	}
	
	if(time(NULL) != ticks) {
		VAR[11]++;
		VAR[12] += VAR[11] / 60;
		VAR[11] %= 60;
		VAR[13] += VAR[12] / 60;
		VAR[12] %= 60;
		VAR[14] += VAR[13] / 24;
		VAR[13] %= 24;
		time(&ticks);
	}	
}

