/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "general.h"
#include "event.h"
#include "graphics.h"
#include "driver.h"
#include "vm.h"

boolean menu_submitted = false;


struct menu_item {
	char name[25];
	UINT8 eventcode;
	boolean enabled;
	
	struct menu_item *next;
	struct menu_item *prev;
};

struct menu {
	char name[10];
	UINT8 maxlength;
	UINT8 numitems;
		
	struct menu *next;
	struct menu *prev;
	
	struct menu_item *firstitem;
	struct menu_item *lastitem;
	struct menu_item *citem;
};


struct menu *firstmenu = NULL;
struct menu *lastmenu = NULL;


void menu_flush(void)
{
	struct menu *currentmenu;
	struct menu_item *currentitem;
	
	while(firstmenu) { /* flush menus */
		while(firstmenu->firstitem) { /* flush menu items */
			currentitem = firstmenu->firstitem->next;
			free(firstmenu->firstitem);
			firstmenu->firstitem = currentitem;
		}
		currentmenu = firstmenu->next;
		free(firstmenu);
		firstmenu = currentmenu;
	}
	firstmenu = NULL;
	lastmenu = NULL;
}


void menu_disable(UINT8 code)
{
	struct menu *currentmenu;
	struct menu_item *currentitem;
	
	currentmenu = firstmenu;
	while(currentmenu) {
		currentitem = currentmenu->firstitem;
		while(currentitem) {
			if(currentitem->eventcode == code) currentitem->enabled = false;
			currentitem = currentitem->next;
		}
		currentmenu = currentmenu->next;
	}
}


void menu_enable(UINT8 code)
{
	struct menu *currentmenu;
	struct menu_item *currentitem;
	
	currentmenu = firstmenu;
	while(currentmenu) {
		currentitem = currentmenu->firstitem;
		while(currentitem) {
			if(currentitem->eventcode == code) currentitem->enabled = true;
			currentitem = currentitem->next;
		}
		currentmenu = currentmenu->next;
	}
}


void menu_add(char * name) 
{
	if(menu_submitted) return;
	
	if(!lastmenu) {
		firstmenu = (struct menu *)malloc(sizeof(struct menu));
		lastmenu = firstmenu;
		lastmenu->prev = NULL;
	} else {
		lastmenu->next = (struct menu *)malloc(sizeof(struct menu));
		lastmenu->next->prev = lastmenu;
		lastmenu = lastmenu->next;
	}
	lastmenu->next = NULL;
	strncpy(lastmenu->name, name, 9);
	lastmenu->firstitem = NULL;
	lastmenu->lastitem = NULL;
	lastmenu->maxlength = 0;
	lastmenu->numitems = 0;
}


void menu_additem(char * name, UINT8 event) 
{
	struct menu *l;

	if(menu_submitted) return;
	if(!lastmenu) return;

	l = lastmenu;
		
	if(!l->lastitem) {
		l->firstitem = (struct menu_item *)malloc(sizeof(struct menu_item));
		l->lastitem = l->firstitem;
		l->lastitem->prev = NULL;
		l->citem = l->firstitem;
	} else {
		l->lastitem->next = (struct menu_item *)malloc(sizeof(struct menu_item));
		l->lastitem->next->prev = l->lastitem;
		l->lastitem = l->lastitem->next;
	}
	l->lastitem->next = NULL;
	strncpy(l->lastitem->name, name, 24);
	l->lastitem->eventcode = event;
	l->lastitem->enabled = true;

	l->numitems++;
	if(strlen(name) > l->maxlength) l->maxlength = strlen(name);
}


void menu_input(void) /* run menu */
{
	UINT8 * tempbuf;

	struct menu *tmenu;
	struct menu_item *titem;
	
	static struct menu *cmenu = NULL;
	
	int xpos = 8, ypos = 16, bxpos = 8;
	int key = 0;
	int i;
	
	if(!menu_submitted) return;
	if(!FLAG[14]) return;
	if(!firstmenu) return;
	
	if(!cmenu) cmenu = firstmenu; /* first call */

	tempbuf = (UINT8 *)malloc(320 * 200); /* temporary graphics buffer */
	
	while(key != 27) {	/* <ESC> pressed */
		xpos = 8;
		memcpy(tempbuf, video_buffer, 320 * 200);
		memset(tempbuf, 15, 320 * 8); /* clear top line */

		/* run through menus */
		for(tmenu = firstmenu; tmenu != NULL; tmenu = tmenu->next) {
			if(tmenu == cmenu) { /* if current menu */
				bxpos = xpos;
				if(bxpos + ((cmenu->maxlength * 8) + 8) >= 320)
					bxpos = 320 - ((cmenu->maxlength * 8) + 8);
				ypos = 16;
				graph_drawstring(tempbuf, xpos, 0, tmenu->name, 15, 0);
				graph_drawbox(tempbuf, bxpos - 8, ypos - 8, (cmenu->maxlength * 8) + 16, (cmenu->numitems * 8) + 16, 15, 0);
				/* run through items */
				for(titem = cmenu->firstitem; titem != NULL; titem = titem->next) {
					/* if current item */
					if(titem == cmenu->citem) 
						graph_drawstring(tempbuf, bxpos, ypos, titem->name, 15, titem->enabled?0:7);
					else
						graph_drawstring(tempbuf, bxpos, ypos, titem->name, titem->enabled?0:7, 15);
					ypos += 8;
				}
			} else graph_drawstring(tempbuf, xpos, 0, tmenu->name, 0, 15);
				
			xpos += (strlen(tmenu->name) + 1) * 8;
		}
		
		driver_showbuffer(tempbuf);
		key = event_waitforkey();
		switch(key) {
			case AGIL_KEY_CURSORRIGHT:
				if(cmenu->next) cmenu = cmenu->next;
				else cmenu = firstmenu;
					
				break;
			case AGIL_KEY_CURSORLEFT:
				if(cmenu->prev) cmenu = cmenu->prev;
				else cmenu = lastmenu;
				break;
			case AGIL_KEY_CURSORUP:
				if(cmenu->citem->prev) cmenu->citem = cmenu->citem->prev;
				else cmenu->citem = cmenu->lastitem;
				break;
			case AGIL_KEY_CURSORDOWN:
				if(cmenu->citem->next) cmenu->citem = cmenu->citem->next;
				else cmenu->citem = cmenu->firstitem;
				break;
			case '\n': case 0x0D:
				if(cmenu->citem->enabled) {
					event_add(cmenu->citem->eventcode);

					/* flush event.c's keytable */
					for(i = 0; i < 128; i++) 
						keytable[i] = 0; 

					free(tempbuf);
					return;
				}
		}
	}
	
	free(tempbuf);
}

