/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "general.h"
#include "resource.h"
#include "graphics.h"
#include "object.h"
#include "check.h"
#include "status.h"
#include "view.h"
#include "vm.h"

view_obj object[255];

void object_update_posn(view_obj *o);
void object_update_cycle(view_obj *o);
void object_recalc_dirs(view_obj *o);


void object_follow_ego(view_obj *o)
{
	o->tx = object[0].x;
	o->ty = object[0].y;
}

void object_wander(view_obj *o)
{
	int tmp;
	if(!o->steps) {
		tmp = rand() % 3;
		if(tmp < 1) o->direction--;
		if(tmp > 1) o->direction++;
		o->direction %= 9;
		if(o->direction == 0) o->direction = 8;
		o->steps = rand() % 10;
	}
}


void object_update(view_obj *o)
{
	if(!(o->flags & VO_UPDATE)) return;

	if(o->flags & (VO_MOTION | VO_WANDER | VO_AUTO_MOVE | VO_FOLLOW_EGO)) {
		if(o->flags & VO_WANDER) object_wander(o);
		if(o->flags & VO_FOLLOW_EGO) object_follow_ego(o);
		if(o->flags & VO_AUTO_MOVE) object_recalc_dirs(o); 

		o->step_count--;
		if(o->step_count == 0) {
			object_update_posn(o);
			o->step_count = o->step_time;
		} 
	}

	if(o->flags & VO_CYCLE) {
		o->cycle_count--;
		if(o->cycle_count == 0) {
			object_update_cycle(o); 
			o->cycle_count = o->cycle_time;
		} 
	}

	if((o - object) == 0) {
		int i;
		VAR[6] = o->direction;

		FLAG[0] = true;
		FLAG[3] = false;

		for(i = 0; i < o->width; i++) {
			if(pixel(backpribuffer, ((o->x + i) * 2), o->y) != 3) FLAG[0] = false;
			if(pixel(backpribuffer, ((o->x + i) * 2), o->y) == 2) FLAG[3] = true;
		}
		
	}
}


void object_recalc_dirs(view_obj *o)
{
	UINT8 dirs[3][3] = {{8, 1, 2}, {7, 0, 3}, {6, 5, 4}};
	int i = 1, j = 1;

	if(abs((int)o->x - o->tx) < o->step_size) {
		o->x = o->tx;
		check_fix_position(o);
	}
	if(abs((int)o->y - o->ty) < o->step_size) {
		o->y = o->ty;
		check_fix_position(o);	
	}

	if(((abs((int)o->y - o->ty) < o->step_size) && 
	    (abs((int)o->x - o->tx) < o->step_size)) ||
	    ((o->flags & VO_NOT_MOVED) && !(o->flags & (VO_FOLLOW_EGO | VO_WANDER)))) {
		o->x = o->tx; o->y = o->ty;
		check_fix_position(o);
		o->flags &= ~(VO_AUTO_MOVE | VO_NOT_MOVED);
		FLAG[o->target_flag] = true;
		if(o == &object[0]) ao_player_control(NULL);
		o->direction = 0;
		return;
	}

	if(o->flags & VO_NOT_MOVED) { 
		o->direction = rand() % 9;
		o->steps = rand() % 10;
		return;
	}
	if(o->steps != 0) return;

	if(o->y > o->ty) i = 0;	
	if(o->y < o->ty) i = 2;	
	if(o->x > o->tx) j = 0;
	if(o->x < o->tx) j = 2;
	
	o->direction = dirs[i][j];
}


void object_update_posn(view_obj *o)
{
	int x, y;
	int i;

	x = o->x;
	y = o->y;
	
	switch(o->direction) {	/* Vertical movement */
		case 1: case 2: case 8:
			y -= o->step_size;
			break;
		case 4: case 5: case 6:
			y += o->step_size;
			break;
	}

	switch(o->direction) { /* Horizontal movement */
		case 6: case 7: case 8:
			x -= o->step_size;
			break;
		case 2: case 3: case 4:
			x += o->step_size;
			break;
	}

	if((y <= status.horizon) || ((y - o->height) < -1)) {
		if(o->flags & VO_HORIZON) {
			if(o == &object[0]) VAR[2] = 1;
			else { VAR[4] = (o - &object[0]); VAR[5] = 1; }
			o->flags |= VO_NOT_MOVED;
			return;
		}
	}
	if(y >= 168) {
		if(o == &object[0]) VAR[2] = 3;
		else { VAR[4] = (o - &object[0]); VAR[5] = 3; }
		o->flags |= VO_NOT_MOVED;
		return;
	}
	if(x < 0) {
		if(o == &object[0]) VAR[2] = 4;
		else { VAR[4] = (o - &object[0]); VAR[5] = 4; }
		o->flags |= VO_NOT_MOVED;
		return;
	}
	if(x + o->width > 160) {
		if(o == &object[0]) VAR[2] = 2;
		else { VAR[4] = (o - &object[0]); VAR[5] = 2; }
		o->flags |= VO_NOT_MOVED;
		return;
	}

	if((o - object) == 0) VAR[2] = 0;
	
	/* Suspect something wrong here with handling of NOLAND/NOWATER */
	/* Check for boundries */
	for(i = 0; i < o->width; i++) {
		if(pixel(backpribuffer, ((x + i) * 2), y) == 0) {
			o->flags |= VO_NOT_MOVED;
			return;
		}
		if((o->flags & VO_BLOCK) &&
		   (pixel(backpribuffer, ((x + i) * 2), y) == 1)) {
		   	o->flags |= VO_NOT_MOVED;
			return;
		}
		if((o->flags & VO_NOLAND) &&
		   (pixel(backpribuffer, ((x + i) * 2), y) != 3)) {
		   	o->flags |= VO_NOT_MOVED;
			return;
		}
		if((o->flags & VO_NOWATER) &&
		   (pixel(backpribuffer, ((x + i) * 2), y) == 3)) {
		   	o->flags |= VO_NOT_MOVED;
			return;
		}
	}
	
	if(o->flags & VO_OBJECTS)
		for(i = 254; i >= 0; i--) {
			if(o == &object[i]) continue;
			if(~object[i].flags & (VO_ANIMATE | VO_DRAWN)) continue;
			if(y == (object[i].y + 1)) {
				int lx, rx, ox;
			
				lx = object[i].x;
				rx = object[i].x + object[i].width - 1;
				ox = x + o->width - 1;
						
				if(((lx < x) && (rx > x)) ||
				  ((lx < ox) && (rx > ox))) {
					o->flags |= VO_NOT_MOVED;
					return;
				}
			}
		}
/*
	if(o == &object[0]) { 
		FLAG[0] = true;
		FLAG[3] = false;

		for(i = 0; i < o->width; i++) {
			if(pixel(backpribuffer, ((x + i) * 2), y) != 3) FLAG[0] = false;
			if(pixel(backpribuffer, ((x + i) * 2), y) == 2) FLAG[3] = true;
		}
		
	}
*/

	o->flags &= ~VO_NOT_MOVED;
	if(o->steps) o->steps--;
		
	o->x = x;
	o->y = y;
}


void object_update_cycle(view_obj *o)
{
	if(o->do_once_flag) {
		if(o->flags & VO_REVERSE) {
			if(o->cel == 0) {
				FLAG[o->do_once_flag] = true;
				o->do_once_flag = 0;
				o->flags &= ~(VO_CYCLE | VO_REVERSE);
				return;
			}
		} else {
			if(o->cel == (o->nocels - 1)) {
				FLAG[o->do_once_flag] = true;
				o->do_once_flag = 0;
				o->flags &= ~(VO_CYCLE);
				return;
			}
		}
	}

	if(o->flags & VO_REVERSE) {
		if(o->cel == 0) o->cel = o->nocels - 1;
		else o->cel--;
	} else {
		o->cel++;
		if(o->cel == o->nocels) o->cel = 0;
	}

	view_recalc(o);
}	


void object_setdefaultpri(view_obj *o)
{
	o->priority = (o->y / 12) + 1;
	if(o->priority < 4) o->priority = 4;
}


void object_setdefaultloop(view_obj *o)
{
	const UINT8 dirloops2[] = {255, 255, 0, 0, 0, 255, 1, 1, 1};
	const UINT8 dirloops4[] = {255,   3, 0, 0, 0,   2, 1, 1, 1};

	if((o->noloops > 1) && (o->noloops < 4))
		if(dirloops2[o->direction] != 255)
				o->loop = dirloops2[o->direction];

	if(o->noloops >= 4)
		if(dirloops4[o->direction] != 255)
				o->loop = dirloops4[o->direction];

	view_recalc(o);
}


