/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 *
 * Portions by Lance Ewing. No copyright info.
 */

#include <stdio.h>
#include <math.h>

#include "general.h"
#include "graphics.h"
#include "pic_op.h"

UINT8 patCode, patNum;

/* These implement the AGI drawing instructions */
UINT16 enablevisdraw(UINT8 * pic) 
{
	vispen = true;
	viscolor = pic[1];
	return 2;
}


UINT16 disablevisdraw(UINT8 * pic)
{
	vispen = false;
	return 1;
}


UINT16 enablepridraw(UINT8 * pic)
{
	pripen = true;
	pricolor = pic[1];
	return 2;
}


UINT16 disablepridraw(UINT8 * pic)
{
	pripen = false;
	return 1;
}


UINT16 drawcorner(UINT8 * pic)
{
	UINT16 index;
	int x = pic[1], y = pic[2];
	boolean dir;
	
	dir = !(pic[0] == 0xF4);

	pset(x, y);
	
	for(index = 3; pic[index] < 0xF0; index++) {
		if(dir) {
			line(x, y, pic[index], y);
			x = pic[index];
		} else {
			line(x, y, x, pic[index]);
			y = pic[index];
		}
		dir = !dir;
	}
	return index;
}


UINT16 drawabsline(UINT8 * pic)
{
	UINT16 index = 3;
	unsigned char x, y;

	x = pic[1];
	y = pic[2];

	pset(x, y);

	while(pic[index] < 0xF0) {	
		line(x, y, pic[index], pic[index + 1]);
		x = pic[index];
		y = pic[index + 1];
		index += 2;
	};

	return index;
}


UINT16 drawrelativeline(UINT8 * pic)
{
	int x1 = pic[1], y1 = pic[2];
	int x2, y2;
	UINT16 index = 3;

	pset(x1, y1);

	while(pic[index] < 0xF0) {
		x2 = x1 + (pic[index] & 0x80 ? -((pic[index] >> 4) & 7) : (pic[index] >> 4) & 7);
		y2 = y1 + (pic[index] & 0x08 ? -(pic[index] & 7) : pic[index] & 7);

		line(x1, y1, x2, y2);
		x1 = x2;
		y1 = y2;
		index++;
	} 
	return index;
}


UINT16 floodfill(UINT8 * pic)
{
	UINT16 index = 1;
	int x, y;

	for(index = 1; pic[index] < 0xF0; index += 2) {
		x = pic[index];
		y = pic[index + 1];

		if(vispen) fillv(x, y);
		if(pripen && !vispen) fillp(x, y);
	}

	return index;
}


UINT16 setbrushstyle(UINT8 * pic)
{
	patCode = pic[1];
	return 2;
}

/* The following couple from AGI spec.  showpic.c by Lance Ewing */
#define plotPatternPoint() \
   if (patCode & 0x20) { \
      if ((splatterMap[bitPos>>3] >> (7-(bitPos&7))) & 1) pset(x1, y1); \
      bitPos++; \
      if (bitPos == 0xff) bitPos=0; \
   } else pset(x1, y1)

void plotPattern(UINT8 x, UINT8 y)
{ 
  static UINT8 circles[][15] = { /* agi circle bitmaps */
    {0x80},
    {0xfc},
    {0x5f, 0xf4},
    {0x66, 0xff, 0xf6, 0x60},
    {0x23, 0xbf, 0xff, 0xff, 0xee, 0x20},
    {0x31, 0xe7, 0x9e, 0xff, 0xff, 0xde, 0x79, 0xe3, 0x00},
    {0x38, 0xf9, 0xf3, 0xef, 0xff, 0xff, 0xff, 0xfe, 0xf9, 0xf3, 0xe3, 0x80},
    {0x18, 0x3c, 0x7e, 0x7e, 0x7e, 0xff, 0xff, 0xff, 0xff, 0xff, 0x7e, 0x7e,
     0x7e, 0x3c, 0x18}
  };

  static UINT8 splatterMap[32] = { /* splatter brush bitmaps */
    0x20, 0x94, 0x02, 0x24, 0x90, 0x82, 0xa4, 0xa2,
    0x82, 0x09, 0x0a, 0x22, 0x12, 0x10, 0x42, 0x14,
    0x91, 0x4a, 0x91, 0x11, 0x08, 0x12, 0x25, 0x10,
    0x22, 0xa8, 0x14, 0x24, 0x00, 0x50, 0x24, 0x04
  };

  static UINT8 splatterStart[128] = { /* starting bit position */
    0x00, 0x18, 0x30, 0xc4, 0xdc, 0x65, 0xeb, 0x48,
    0x60, 0xbd, 0x89, 0x05, 0x0a, 0xf4, 0x7d, 0x7d,
    0x85, 0xb0, 0x8e, 0x95, 0x1f, 0x22, 0x0d, 0xdf,
    0x2a, 0x78, 0xd5, 0x73, 0x1c, 0xb4, 0x40, 0xa1,
    0xb9, 0x3c, 0xca, 0x58, 0x92, 0x34, 0xcc, 0xce,
    0xd7, 0x42, 0x90, 0x0f, 0x8b, 0x7f, 0x32, 0xed,
    0x5c, 0x9d, 0xc8, 0x99, 0xad, 0x4e, 0x56, 0xa6,
    0xf7, 0x68, 0xb7, 0x25, 0x82, 0x37, 0x3a, 0x51,
    0x69, 0x26, 0x38, 0x52, 0x9e, 0x9a, 0x4f, 0xa7,
    0x43, 0x10, 0x80, 0xee, 0x3d, 0x59, 0x35, 0xcf,
    0x79, 0x74, 0xb5, 0xa2, 0xb1, 0x96, 0x23, 0xe0,
    0xbe, 0x05, 0xf5, 0x6e, 0x19, 0xc5, 0x66, 0x49,
    0xf0, 0xd1, 0x54, 0xa9, 0x70, 0x4b, 0xa4, 0xe2,
    0xe6, 0xe5, 0xab, 0xe4, 0xd2, 0xaa, 0x4c, 0xe3,
    0x06, 0x6f, 0xc6, 0x4a, 0xa4, 0x75, 0x97, 0xe1
  };

  int circlePos = 0;
  UINT8 x1, y1, penSize, bitPos = splatterStart[patNum];

  penSize = (patCode&7);

  if (x<((penSize/2)+1)) x=((penSize/2)+1);
  else if (x>160-((penSize/2)+1)) x=160-((penSize/2)+1);
  if (y<penSize) y = penSize;
  else if (y>=168-penSize) y=167-penSize;

  for (y1=y-penSize; y1<=y+penSize; y1++) {
    for (x1=x-(ceil((float)penSize/2)); x1<=x+(floor((float)penSize/2)); x1++) 
{
      if (patCode & 0x10) { /* Square */
        plotPatternPoint();
      }
      else { /* Circle */
        if ((circles[patCode&7][circlePos>>3] >> (7-(circlePos&7)))&1) {
          plotPatternPoint();
        }
        circlePos++;
      }
    }
  }

} 


UINT16 plotBrush(UINT8 *pic)
{
   UINT16 index;
   UINT8 x1, y1;

   for (index = 1; pic[index] < 0xF0; index += 2) {
     if (patCode & 0x20) {
        patNum = pic[index];
        patNum = ((patNum >> 1) & 0x7f);
	index++;
     }
     x1 = pic[index];
     y1 = pic[index + 1];
     plotPattern(x1, y1);
   }
   return index;
}


