/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 *
 * I may have copied some of this from elsewhere, but don't remember.
 */

/*
 * I'm not at all happy with this file.
 * When I have a good idea it will be completely rewritten.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "general.h"
#include "resource.h"
#include "text.h"
#include "gameid.h"
#include "picture.h"
#include "lzw.h"

resource_t *LOGIC[256];
resource_t *PICTURE[256];
resource_t *VIEW[256];
resource_t *SOUND[256];

inv_item_t inv_item[256];

FILE *vol[15];

char gamedir[40] = "./";

#define initentry(x, ix)					\
	x[ix] = (resource_t *)malloc(sizeof(resource_t));	\
	x[ix]->id = ix;						\
	x[ix]->data = NULL;					\
	x[ix]->size = 0;					\
	x[ix]->message = NULL;					\
	x[ix]->compressed = false;

void v2loaddirs(void);
void v3loaddirs(void);
void decodemessages(resource_t *l);


int res_init(void)
{
	char volfilename[50];
	int i;
	
	atexit(res_shutdown);	

	if(!id) id_game();
	
	for(i = 0; i < 256; i++) { /* Zero all resources */
		LOGIC[i] = NULL;
		PICTURE[i] = NULL;
		VIEW[i] = NULL;
		SOUND[i] = NULL;
		inv_item[i].name[0] = 0;
		inv_item[i].room = 0;
	}

	for(i = 0; i < 15; i++) { /* Open all existant vol's */
		vol[i] = NULL;
		
		if(id->agiver[0] == '3')
			sprintf(volfilename, "%s%svol.%d", gamedir, id->id, i);
		else
			sprintf(volfilename, "%svol.%d", gamedir, i);
			
		vol[i] = fopen(volfilename, "rb");
	}

	if(id->agiver[0] == '3')	/* load *dir */
		v3loaddirs();
	else
		v2loaddirs();
	
	printf("resource: Resources initalized\n");
	return 0;
}


void res_load(resource_t *res)
{
	UINT8 volheader[7];
	int tempsize;
	
	if(!res) {
		printf("resource: WARNING: attempt to load nonexistant resource\n");
		return;
	}
	if(!vol[res->vol]) {
		printf("resource: FATAL: resource load request from NULL vol\n");
		exit(-1);
	}

	if(res->data) return;
		
	fseek(vol[res->vol], res->offset, SEEK_SET);

	if(id->agiver[0] == '3') {
		fread(volheader, 1, 7, vol[res->vol]);
		res->size = ((UINT16)volheader[4] << 8) | volheader[3];
		res->data = (UINT8 *)malloc(res->size);
		tempsize = ((UINT16)volheader[6] << 8) | volheader[5];

		/* somebody said if(volheader[2] >> 4) but this breaks VIEWs */
		if(tempsize != res->size) { /* compressed resource */
		
			if(res->resource_type == picture) 
				pic_inflate(res->data, vol[res->vol], tempsize);
			else	lzw_inflate(res->data, vol[res->vol]);
			
			res->compressed = true; 
				/* Needed for logic message decoding */
			
		} else 	fread(res->data, 1, res->size, vol[res->vol]);
	} else {
		fread(volheader, 1, 5, vol[res->vol]);	
		res->size = ((UINT16)volheader[4] << 8) | volheader[3];
		res->data = (UINT8 *)malloc(res->size);
		fread(res->data, 1, res->size, vol[res->vol]);
	}

	if(res->resource_type == logic) decodemessages(res);
}


void res_unload(resource_t *res)
{
	if(!res) {
		printf("resource: WARNING: Attempt to unload nonexistant resource\n");
		return;
	}
	
	if(res->data) {
		free(res->data);
		res->data = NULL;
		if((res->resource_type == logic) && (res->message)) {
			free(res->message);
			res->message = NULL;
		}
	} else {
		printf("resource: WARNING: Attempt to unload resource not loaded\n");
	}
}


void res_shutdown(void)
{
	int i;

	for(i = 0; i < 15; i++) {
		if(vol[i]) fclose(vol[i]);
		vol[i] = NULL;
	}

	for(i = 0; i < 256; i++) {
		if(LOGIC[i]) {
			if(LOGIC[i]->data) free(LOGIC[i]->data);
			free(LOGIC[i]);
			LOGIC[i] = NULL;
		}
		if(PICTURE[i]) {
			if(PICTURE[i]->data) free(PICTURE[i]->data);
			free(PICTURE[i]);
			PICTURE[i] = NULL;
		}
		if(VIEW[i]) {
			if(VIEW[i]->data) free(VIEW[i]->data);
			free(VIEW[i]);
			VIEW[i] = NULL;
		}
		if(SOUND[i]) {
			if(SOUND[i]->data) free(SOUND[i]->data);
			free(SOUND[i]);
			SOUND[i] = NULL;
		}
	}
	printf("resource: Resources shutdown\n");
}


void res_load_words(void)
{
	FILE * words;
	char wordsfilename[50];
	char word[40] = "";
	int i = 0;
	int c;
	int number;
	
	snprintf(wordsfilename, 49, "%swords.tok", gamedir);
	if(!(words = fopen(wordsfilename, "rb"))) {
		printf("resource: CRITICAL: Error opening 'words.tok'\n");
		return;
	}
	
	number = (fgetc(words) << 8) + fgetc(words);
	
	fseek(words, number, SEEK_SET);
	
	while(!feof(words)) {
		c = fgetc(words);
		if(feof(words)) break;
		i = c;
		while(c < 0x80) {
			c = fgetc(words);
			if(feof(words)) break;
			word[i] = ~c & 0x7F;
			i++;
		}
		word[i] = 0;
		/* do these input past end of file?? */
		number = fgetc(words) << 8;
		number |= fgetc(words);
		if(word[0]) /* We get a zero word at end of file ?? */
			text_addword(word, number);
	}
	
	fclose(words);
	
	printf("resource: Dictionaries loaded\n");
}


void res_load_objects(void)
{
	FILE * objfile;
	char objfilename[50];
	int filesize = 0;
	int i;
	int textstart;
	int offset;
	int entries;
	
	unsigned char *objdata;
	char key[] = "Avis Durgan";	

	snprintf(objfilename, 49, "%sobject", gamedir);
	if(!(objfile = fopen(objfilename, "rb"))) {
		printf("resource: CRITICAL: Error opening 'object'\n");
		return;
	}
	
	fseek(objfile, 0, SEEK_END);
	filesize = ftell(objfile);
		
	fseek(objfile, 0, SEEK_SET);
	objdata = (unsigned char *)alloca(filesize);
	fread(objdata, 1, filesize, objfile);
	fclose(objfile);
	
	if((((int)objdata[1] << 8) | objdata[0]) >= filesize)
		for(i = 0; i <= filesize; i++)  
			objdata[i] ^= key[i % 11];
	
	textstart = ((int)objdata[1] << 8) | objdata[0];
	entries = (textstart / 3) - 1;
	objdata += 3;
	for(i = 0; i <= entries; i++) {
		offset = ((int)objdata[(i * 3) + 1] << 8) | objdata[i * 3];
		strncpy(inv_item[i].name, (char *)&objdata[offset], 24);
		inv_item[i].room = objdata[(i * 3) + 2];
/*		printf("inv object %d: %s %d %.4X\n", i, inv_item[i].name, inv_item[i].room, offset);*/
	}
	printf("resource: Inventory objects loaded\n");
}


void v2loaddirs(void)
{
	FILE * dirfile;
	int i;
	char dirfilename[50];
	UINT8 direntry[3];

	snprintf(dirfilename, 49, "%slogdir", gamedir);
	if(!(dirfile = fopen(dirfilename, "rb"))) {
		printf("resource: FATAL: Error opening logdir\n");
		exit(-1);
	}
	for(i = 0; !feof(dirfile) && (i < 256); i++) {
		fread(direntry, 1, 3, dirfile);
		if(direntry[0] == 0xFF) continue;
		initentry(LOGIC, i);
		LOGIC[i]->vol = direntry[0] >> 4;
		LOGIC[i]->offset = (((int)direntry[0] & 0x0F) << 16) | ((int)direntry[1] << 8) | direntry[2];
		LOGIC[i]->resource_type = logic;
		LOGIC[i]->ssIP = 0;
	}
	fclose(dirfile);

	snprintf(dirfilename, 49, "%spicdir", gamedir);
	if(!(dirfile = fopen(dirfilename, "rb"))) {
		printf("resource: FATAL: Error opening picdir\n");
		exit(-1);
	}
	for(i = 0; !feof(dirfile) && (i < 256); i++) {
		fread(direntry, 1, 3, dirfile);
		if(direntry[0] == 0xFF) continue;
		initentry(PICTURE, i);
		PICTURE[i]->vol = direntry[0] >> 4;
		PICTURE[i]->offset = (((int)direntry[0] & 0x0F) << 16) | ((int)direntry[1] << 8) | direntry[2];
		PICTURE[i]->resource_type = picture;
	}
	fclose(dirfile);


	snprintf(dirfilename, 49, "%sviewdir", gamedir);	
	if(!(dirfile = fopen(dirfilename, "rb"))) {
		printf("resource: FATAL: Error opening viewdir\n");
		exit(-1);
	}
	for(i = 0; !feof(dirfile) && (i < 256); i++) {
		fread(direntry, 1, 3, dirfile);
		if(direntry[0] == 0xFF) continue;
		initentry(VIEW, i);
		VIEW[i]->vol = direntry[0] >> 4;
		VIEW[i]->offset = (((int)direntry[0] & 0x0F) << 16) | ((int)direntry[1] << 8) | direntry[2];
		VIEW[i]->resource_type = view;
	}
	fclose(dirfile);

	snprintf(dirfilename, 49, "%ssnddir", gamedir);
	if(!(dirfile = fopen(dirfilename, "rb"))) {
		printf("resource: Error opening snddir (Who cares?)\n");
		return;
	}
	for(i = 0; !feof(dirfile) && (i < 256); i++) {
		fread(direntry, 1, 3, dirfile);
		if(direntry[0] == 0xFF) continue;
		initentry(SOUND, i);
		SOUND[i]->vol = direntry[0] >> 4;
		SOUND[i]->offset = (((int)direntry[0] & 0x0F) << 16) | ((int)direntry[1] << 8) | direntry[2];
		SOUND[i]->resource_type = sound;
	}
	fclose(dirfile);
}


void v3loaddirs(void)
{
	UINT16 header[4];
	char dirfilename[50];
	FILE * dirfile;
	UINT8 direntry[3];
	int i;
	
	sprintf(dirfilename, "%s%sdir", gamedir, id->id);
	if(!(dirfile = fopen(dirfilename, "rb"))) {
		printf("resource: FATAL: Error opening %sdir\n", id->id);
		exit(-1);
	}
	fread(header, 2, 4, dirfile); /* Warning : This may be bad practice */
			/* UINT16 might be different on other systems */
			
	for(i = 0; i < (header[1] - header[0]) / 3; i++) { /* read logdir */
		if(feof(dirfile)) return;
		fread(direntry, 1, 3, dirfile);
		if(direntry[0] == 0xFF) continue;
		initentry(LOGIC, i);
		LOGIC[i]->vol = direntry[0] >> 4;
		LOGIC[i]->offset = (((int)direntry[0] & 0x0F) << 16) | ((int)direntry[1] << 8) | direntry[2];
		LOGIC[i]->resource_type = logic;
		LOGIC[i]->ssIP = 0;
	}
	
	for(i = 0; i < (header[2] - header[1]) / 3; i++) { /* read picdir */
		if(feof(dirfile)) return;
		fread(direntry, 1, 3, dirfile);
		if(direntry[0] == 0xFF) continue;
		initentry(PICTURE, i);
		PICTURE[i]->vol = direntry[0] >> 4;
		PICTURE[i]->offset = (((int)direntry[0] & 0x0F) << 16) | ((int)direntry[1] << 8) | direntry[2];
		PICTURE[i]->resource_type = picture;
	}

	for(i = 0; i < (header[3] - header[2]) / 3; i++) { /* read viewdir */
		if(feof(dirfile)) return;
		fread(direntry, 1, 3, dirfile);
		if(direntry[0] == 0xFF) continue;
		initentry(VIEW, i);
		VIEW[i]->vol = direntry[0] >> 4;
		VIEW[i]->offset = (((int)direntry[0] & 0x0F) << 16) | ((int)direntry[1] << 8) | direntry[2];
		VIEW[i]->resource_type = view;
	}

	for(i = 0; !feof(dirfile); i++) {		/* read snddir */
		if(feof(dirfile)) return;
		fread(direntry, 1, 3, dirfile);
		if(direntry[0] == 0xFF) continue;
		initentry(SOUND, i);
		SOUND[i]->vol = direntry[0] >> 4;
		SOUND[i]->offset = (((int)direntry[0] & 0x0F) << 16) | ((int)direntry[1] << 8) | direntry[2];
		SOUND[i]->resource_type = sound;
	}
	
	fclose(dirfile);
}


void decodemessages(resource_t * l) {
	UINT16 messageblock = 0;
	UINT8 messagecount = 0;
	UINT16 messageoffset = 0;
	UINT16 start, end;
	UINT16 index = 0;
	const char key[] = "Avis Durgan";

	messageblock = (UINT16)(l->data[1] << 8) + l->data[0] + 2;
	messagecount = l->data[messageblock];

	if(!l->compressed) {	/* Decrypt with "Avis Durgan" */
/*		printf("resource: Decrypting v2 messages\n");*/
/*		start = (UINT16)(l->data[messageblock + 4] << 8) + l->data[messageblock + 3] + 1;*/
		start = (messagecount * 2) + 3;
		end = (UINT16)(l->data[messageblock + 2] << 8) + l->data[messageblock + 1];
		for(index = 0; index <= end - start; index++) 
			l->data[messageblock + index + start] ^= key[index % 11];
	}
	
	/* set up message pointers */

	l->message = (char **)malloc(messagecount * sizeof(char *));
	for(index = 0; index < messagecount; index++) {
		messageoffset = (UINT16)(l->data[messageblock + (index << 1) + 4] << 8) + l->data[messageblock + (index << 1) + 3] + 1;
		l->message[index] = (char *)&l->data[messageblock + messageoffset];
/*		printf("%s\n", l->message[index]);*/
	}
}

