/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>

#include "general.h"
#include "save.h"
#include "driver.h"
#include "gameid.h"
#include "vm.h"
#include "object.h"
#include "status.h"
#include "graphics.h"
#include "resource.h"
#include "event.h"
#include "message.h"
#include "text.h"


char savedir[31];

void save_savedialogue(void)
{

	char sgname[42];
	char disc[12][35];
	char mesg[4000] = "";
	FILE * temp;
	int i, p = 0;
	int k;
	boolean olddialog;
	
	olddialog = status.dialogue_open;
	status.dialogue_open = true;
	
	memcpy(textmodebuffer, video_buffer, 64000);
	message_erasebox();
	message_mbox("           SAVE GAME\n\nWhere would you like to save this game?\n\n", 0, 0, 0);
	driver_showbuffer(video_buffer);

	text_input("", 4, 13, 30, savedir);
	strcpy(savedir, inputstring);
#ifdef UNIX
	mkdir(savedir, 0755);
#endif
	input_pos = 0; inputstring[input_pos] = 0;
	message_erasebox();
	
	if(savedir[strlen(savedir) - 1] != '/') {
		savedir[strlen(savedir) + 1] = 0;
		savedir[strlen(savedir)] = '/';
	}
	p += sprintf(mesg, "Use the arrow keys to select the slot in which you wish to save the game. Press ENTER to save in the slot, ESC to not save a game.\n");
	for(i = 0; i < 12; i++) {
		sprintf(sgname, "%s%ssg.%d", savedir, id->id, i);
		temp = fopen(sgname, "r");
		if(temp) {
			fread(disc[i], 42, 1, temp);
			fclose(temp);
		} else disc[i][0] = 0;
		p += sprintf(&mesg[p], "\n - %s", disc[i]);
	}
	message_mbox(mesg, 0, 0, 0);
	i = 0;
	
	graph_drawchar(video_buffer, 24, (7 + i) * 8 + 4, '>', 0, 15);
	driver_showbuffer(video_buffer);
	while((k = event_waitforkey()) != 0x0D) {
		if(k == 27) {
			message_erasebox();
			memcpy(video_buffer, textmodebuffer, 64000);
			status.dialogue_open = olddialog;
			return;
		}
		graph_drawchar(video_buffer, 24, (7 + i) * 8 + 4, '>', 15, 15);
		switch(k) {
			case AGIL_KEY_CURSORDOWN: 
				i++;
				break;
			case AGIL_KEY_CURSORUP:
				i--;
				break;
		}
		if(i < 0) i = 11;
		if(i > 11) i = 0;
		graph_drawchar(video_buffer, 24, (7 + i) * 8 + 4, '>', 0, 15);
		driver_showbuffer(video_buffer);	
	}	

	message_erasebox();
	memcpy(video_buffer, textmodebuffer, 64000); /* cheap workaround */
	message_mbox("Please enter a discription for your saved game.\n\n", 0, 0, 0);
	text_input("", 3, 12, 31, disc[i]);
	strcpy(disc[i], inputstring);

	input_pos = 0; inputstring[input_pos] = 0;
	message_erasebox();
	memcpy(video_buffer, textmodebuffer, 64000);

	sprintf(sgname, "%s%ssg.%d", savedir, id->id, i);	
	save_savegame(sgname, disc[i]);
	status.dialogue_open = olddialog;
}



void save_restoredialogue(void)
{
	char sgname[42];
	char disc[35];
	char mesg[4000] = "";
	FILE * temp;
	int i, p = 0;
	int k;
	int savegames[12];
	int savecount = 0;
	boolean olddialog;
	int top = 0;
	
	olddialog = status.dialogue_open;
	status.dialogue_open = true;
	
	memcpy(textmodebuffer, video_buffer, 64000);
	message_erasebox();
	message_mbox("           RESTORE GAME\n\nWhere would you like to restore a game from?\n\n", 0, 0, 0);
	driver_showbuffer(video_buffer);

	text_input("", 4, 13, 30, savedir);
	strcpy(savedir, inputstring);

	input_pos = 0; inputstring[input_pos] = 0;
	message_erasebox();
	
	if(savedir[strlen(savedir) - 1] != '/') {
		savedir[strlen(savedir) + 1] = 0;
		savedir[strlen(savedir)] = '/';
	}
	p += sprintf(mesg, "Use the arrow keys to select the game which you wish to restore.   \nPress ENTER to restore the game,\nESC to not restore a game.\n");
	for(i = 0; i < 12; i++) {
		sprintf(sgname, "%s%ssg.%d", savedir, id->id, i);
		temp = fopen(sgname, "r");
		if(temp) {
			fread(disc, 42, 1, temp);
			p += sprintf(&mesg[p], "\n - %s", disc);
			savegames[savecount] = i;
			fclose(temp);
			savecount++;
		}
	}

	if(savecount == 0) {
		message_mbox("No savegames found in specified directory.", 0, 0, 0);
		driver_showbuffer(video_buffer);
		event_waitforkey();
		message_erasebox();
		memcpy(video_buffer, textmodebuffer, 64000);
		return;
	}

	message_mbox(mesg, 0, 0, 0);
	i = 0;
	
	if(top <= 0) top = 137 - ((((savecount + 6) * 8) + 10) / 2);

	graph_drawchar(video_buffer, 24, (i * 8) + top, '>', 0, 15);
	driver_showbuffer(video_buffer);
	while((k = event_waitforkey()) != 0x0D) {
		if(k == 27) {
			message_erasebox();
			memcpy(video_buffer, textmodebuffer, 64000);
			status.dialogue_open = olddialog;
			return;
		}
		graph_drawchar(video_buffer, 24, (i * 8) + top, '>', 15, 15);
		switch(k) {
			case AGIL_KEY_CURSORDOWN: 
				i++;
				break;
			case AGIL_KEY_CURSORUP:
				i--;
				break;
		}
		if(i < 0) i = savecount - 1;
		if(i >= savecount) i = 0;
		graph_drawchar(video_buffer, 24, (i * 8) + top, '>', 0, 15);
		driver_showbuffer(video_buffer);	
	}	

	message_erasebox();
	memcpy(video_buffer, textmodebuffer, 64000);

	sprintf(sgname, "%s%ssg.%d", savedir, id->id, savegames[i]);	
	save_restoregame(sgname);
}


void save_savegame(char * filename, char * discription)
{
	FILE *savefile;
	int i;

	savefile = fopen(filename, "wb");
	if(!savefile) {
		memcpy(textmodebuffer, video_buffer, 64000);
		message_mbox("Failed to open savegame file", 0, 0, 0);
		driver_showbuffer(video_buffer);
		while(event_waitforkey() != 0x0D);
		message_erasebox();
		memcpy(video_buffer, textmodebuffer, 64000);
		return;
	}

	fwrite(discription, 35, 1, savefile);
	fwrite(&id->crc, 4, 1, savefile);
	fwrite(VAR, 1, 256, savefile);
	fwrite((void *)FLAG, sizeof(boolean), 256, savefile);
	fwrite(STRING, 40, 24, savefile);
	fwrite(object, sizeof(view_obj), 255, savefile);
	fwrite(inv_item, sizeof(inv_item_t), 255, savefile);
	fwrite(&status, sizeof(struct game_status_s), 1, savefile);
	
	fwrite(workvisbuffer, 1, 64000, savefile);
	fwrite(workpribuffer, 1, 64000, savefile);
	fwrite(backvisbuffer, 1, 53760, savefile);
	fwrite(backpribuffer, 1, 53760, savefile);

	for(i = 0; i < 256; i++) {
		if(LOGIC[i] && LOGIC[i]->data) fputc(1, savefile); else fputc(0, savefile);
		if(PICTURE[i] && PICTURE[i]->data) fputc(1, savefile); else fputc(0, savefile);
		if(VIEW[i] && VIEW[i]->data) fputc(1, savefile); else fputc(0, savefile);
		if(SOUND[i] && SOUND[i]->data) fputc(1, savefile); else fputc(0, savefile);
	}
	
	memcpy(textmodebuffer, video_buffer, 64000);
	message_mbox("Game Saved", 0, 0, 0);
	driver_showbuffer(video_buffer);
	while(event_waitforkey() != 0x0D);
	message_erasebox();
	memcpy(video_buffer, textmodebuffer, 64000);	

	fclose(savefile);
}

void save_restoregame(char * filename)
{
	FILE *savefile;
	int i;
	UINT32 savecrc;
	char discription[35];
	
	savefile = fopen(filename, "rb");
	if(!savefile) {
		memcpy(textmodebuffer, video_buffer, 64000);
		message_mbox("Failed to open savegame file", 0, 0, 0);
		driver_showbuffer(video_buffer);
		while(event_waitforkey() != 0x0D);
		message_erasebox();
		memcpy(video_buffer, textmodebuffer, 64000);	
		return;
	}
	
	fread(discription, 35, 1, savefile);
	fread(&savecrc, 4, 1, savefile);
	
	if(savecrc != id->crc) {
		memcpy(textmodebuffer, video_buffer, 64000);
		message_mbox("Savegame file is from a different game!", 0, 0, 0);
		driver_showbuffer(video_buffer);
		while(event_waitforkey() != 0x0D);
		message_erasebox();
		memcpy(video_buffer, textmodebuffer, 64000);	
		return;
	}
	
	fread(VAR, 1, 256, savefile);
	fread((void *)FLAG, sizeof(boolean), 256, savefile);
	fread(STRING, 40, 24, savefile);
	fread(object, sizeof(view_obj), 255, savefile);
	fread(inv_item, sizeof(inv_item_t), 255, savefile);
	fread(&status, sizeof(struct game_status_s), 1, savefile);
	
	object[0].direction = 0; VAR[6] = 0;

	fread(workvisbuffer, 1, 64000, savefile);
	fread(workpribuffer, 1, 64000, savefile);
	fread(backvisbuffer, 1, 53760, savefile);
	fread(backpribuffer, 1, 53760, savefile);

	for(i = 0; i < 256; i++) {
		if(fgetc(savefile)) res_load(LOGIC[i]);
		if(fgetc(savefile)) res_load(PICTURE[i]);
		if(fgetc(savefile)) res_load(VIEW[i]);
		if(fgetc(savefile)) res_load(SOUND[i]);
	}

	fclose(savefile);
}
