/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 *
 *	Basically copied from LollypopMan's threaded SDL sound driver 
 *	(original) to generalize sound support.  My effort may be found in
 *	sound.c.rad.  Portions of this file are from agiplay.c in
 *	the AGI spec, by Claudio Matsuoka.
 *
 *	THE VAST MAJORITY OF THIS FILE IS NOT MINE!
 *
 *	Only produces raw signed 16bit mono 22050 Hz data.
 *	Sound drivers must use this (or shut up: sound_enabled = false; )
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "general.h"
#include "sound.h"
#include "vm.h"

#define MAX_OUTPUT 0x7fff
#define STEP 0x10000

/* Formulas for noise generator */
/* bit0 = output */

/* noise feedback for white noise mode */
#define FB_WNOISE 0x12000	/* bit15.d(16bits) = bit0(out) ^ bit2 */
/*#define FB_WNOISE 0x14000*/	/* bit15.d(16bits) = bit0(out) ^ bit1 */
/*#define FB_WNOISE 0x28000*/	/* bit16.d(17bits) = bit0(out) ^ bit2 (same to AY-3-8910) */
/*#define FB_WNOISE 0x50000*/	/* bit17.d(18bits) = bit0(out) ^ bit2 */

/* noise feedback for periodic noise mode */
/* it is correct maybe (it was in the Megadrive sound manual) */
/*#define FB_PNOISE 0x10000*/	/* 16bit rorate */
#define FB_PNOISE 0x08000

/* noise generator start preset (for periodic noise) */
#define NG_PRESET 0x0f35

void play_note (int i, int freq, int vol);
void stop_note (int i);
void mix_channels (int s);


/* Sound driver stuff */
struct agi_note {
        UINT8 dur_lo;
        UINT8 dur_hi;
        UINT8 frq_0;
        UINT8 frq_1;
        UINT8 vol;
};

struct channel_info {
        struct agi_note *ptr;
        int end;
        int freq;
        int phase;
        int vol;
        int timer;
};




struct SN76496
{
	int Channel;
	int SampleRate;
	unsigned int UpdateStep;
	int VolTable[16];	/* volume table         */
	int Register[8];	/* registers */
	int LastRegister;	/* last register written */
	int Volume[4];		/* volume of voice 0-2 and noise */
	unsigned int RNG;		/* noise generator      */
	int NoiseFB;		/* noise feedback mask */
	int Period[4];
	int Count[4];
	int Output[4];
};


static struct SN76496 sn;

static SINT16 buffer[BUFFER_SAMPLE_SIZE];
static struct channel_info chn[4];
static UINT8 playedflag;
volatile boolean playing = false;
boolean sound_enabled = true;

void sound_init(void)
{
	int i;
	struct SN76496 *R = &sn;
	double out;
	int gain = 0xff;
	
	R->SampleRate = 22000;
	R->UpdateStep = ((double)STEP * R->SampleRate * 16) / 3579000;

	for (i = 0;i < 4;i++) R->Volume[i] = 0;

	R->LastRegister = 0;
	for (i = 0;i < 8;i+=2)
	{
		R->Register[i] = 0;
		R->Register[i + 1] = 0x0f;	/* volume = 0 */
	}

	for (i = 0;i < 4;i++)
	{
		R->Output[i] = 0;
		R->Period[i] = R->Count[i] = R->UpdateStep;
	}
	R->RNG = NG_PRESET;
	R->Output[3] = R->RNG & 1;

	/* increase max output basing on gain (0.2 dB per step) */
	out = MAX_OUTPUT / 3;
	while (gain-- > 0)
		out *= 1.023292992;	/* = (10 ^ (0.2/20)) */

	/* build volume table (2dB per step) */
	for (i = 0;i < 15;i++)
	{
		/* limit volume to avoid clipping */
		if (out > MAX_OUTPUT / 3) R->VolTable[i] = MAX_OUTPUT / 3;
		else R->VolTable[i] = out;

		out /= 1.258925412;	/* = 10 ^ (2/20) = 2dB */
	}
	R->VolTable[15] = 0;
}


void sound_play(UINT8 *data, UINT8 flg)
{
	int i;
        struct SN76496 *R = &sn;

	if(!sound_enabled) {
		FLAG[flg] = true;
		return;
	}
	
	/* Kill the Sound registers */
	for (i = 0;i < 4;i++) R->Volume[i] = 0;

	R->LastRegister = 0;
	for (i = 0;i < 8;i+=2)
	{
		R->Register[i] = 0;
		R->Register[i + 1] = 0x0f;	/* volume = 0 */
	}

	for (i = 0;i < 4;i++)
	{
		R->Output[i] = 0;
		R->Period[i] = R->Count[i] = R->UpdateStep;
	}
	R->RNG = NG_PRESET;
	R->Output[3] = R->RNG & 1;

        for (i = 0; i < 4; i++) {
                chn[i].ptr = (struct agi_note *)(data +
                        (data[i << 1] | (data[(i << 1) + 1] << 8)));
                chn[i].timer = 0;
                chn[i].end = 0;
        }
	playedflag = flg;
	playing = true;
}


void sound_stop(void)
{
	int i;
	
	if(!sound_enabled) return;
	
	playing = false;
	for(i = 0; i < 4; i++) {
		chn[i].ptr = NULL;
		chn[i].timer = 0;
		chn[i].end = true;
	}
}


SINT16 * sound_getbuffer(void)
{
	int freq, i;

	if(playing) {	/* Write out sound data */
		for (playing = i = 0; i < 4; i++) {
                        playing |= !chn[i].end;
                        
                        if (chn[i].end)
                                continue;

                        if ((--chn[i].timer) <= 0) {
                                stop_note (i);
                                chn[i].timer = ((int)chn[i].ptr->dur_hi << 8) |
                                        chn[i].ptr->dur_lo;
                                if (chn[i].timer == 0xffff) {
                                        chn[i].end = 1;
					continue;
				}
                                freq = ((chn[i].ptr->frq_0 & 0x3f) << 4)
                                        | (int)(chn[i].ptr->frq_1 & 0x0f);
                                if (freq)
                                        play_note (i, freq, chn[i].ptr->vol);
                                chn[i].ptr++;
                        }
                }
                mix_channels (BUFFER_SAMPLE_SIZE);
		if(!playing) FLAG[playedflag] = true;
        } else 	/* Write out silence */
		memset(buffer, 0, BUFFER_BYTE_SIZE);

	/* If sound is off */
	if(!FLAG[9]) memset(buffer, 0, BUFFER_BYTE_SIZE);
	
	return buffer;
}




void play_note (int i, int freq, int vol)
{
	struct SN76496 *R = &sn;
        /*if (vol >= 255 || freq < 16 || freq >= 900)
               	vol = 0;*/
        
        switch(i)
        {
        	case 0:
        		R->Register[0] = freq;
        		R->Period[0] = R->UpdateStep * R->Register[0];
        		if(R->Period[0] == 0) R->Period[0] = R->UpdateStep;
        		R->Volume[0] = R->VolTable[vol & 0x0f];
        		break;
        	
        	case 1:
        		R->Register[2] = freq;
        		R->Period[1] = R->UpdateStep * R->Register[2];
        		if(R->Period[1] == 0) R->Period[1] = R->UpdateStep;
        		R->Volume[1] = R->VolTable[vol & 0x0f];
        		break;
        		
        	case 2:
        		R->Register[4] = freq;
        		R->Period[2] = R->UpdateStep * R->Register[4];
        		if(R->Period[2] == 0) R->Period[2] = R->UpdateStep;
        		R->Volume[2] = R->VolTable[vol & 0x0f];
        		
        		/* update noise shift frequency */
			if ((R->Register[6] & 0x03) == 0x03)
				R->Period[3] = 2 * R->Period[2];
        		break;
        		
        	case 3:
        		R->Register[6] = freq;
        		{
				int n = R->Register[6];
				R->NoiseFB = (n & 4) ? FB_WNOISE : FB_PNOISE;
				n &= 3;
				/* N/512,N/1024,N/2048,Tone #3 output */
				R->Period[3] = (n == 3) ? 2 * R->Period[2] : (R->UpdateStep << (5+n));
				/* reset noise shifter */
				R->RNG = NG_PRESET;
				R->Output[3] = R->RNG & 1;
			}
			R->Volume[3] = R->VolTable[vol & 0x0f];
			break;
        }


        chn[i].freq = freq;
        chn[i].phase = 0;
        chn[i].vol = vol;

}

void stop_note (int i)
{
        chn[i].vol = 0;
}


void mix_channels (int length)
{
        int i;
	struct SN76496 *R = &sn;
	SINT16 *buf = buffer;

	/* If the volume is 0, increase the counter */
	for (i = 0;i < 4;i++)
	{
		if (R->Volume[i] == 0)
		{
			/* note that I do count += length, NOT count = length + 1. You might think */
			/* it's the same since the volume is 0, but doing the latter could cause */
			/* interferencies when the program is rapidly modulating the volume. */
			if (R->Count[i] <= length*STEP) R->Count[i] += length*STEP;
		}
	}

	while (length > 0)
	{
		int vol[4];
		unsigned int out;
		int left;


		/* vol[] keeps track of how long each square wave stays */
		/* in the 1 position during the sample period. */
		vol[0] = vol[1] = vol[2] = vol[3] = 0;

		for (i = 0;i < 3;i++)
		{
			if (R->Output[i]) vol[i] += R->Count[i];
			R->Count[i] -= STEP;
			/* Period[i] is the half period of the square wave. Here, in each */
			/* loop I add Period[i] twice, so that at the end of the loop the */
			/* square wave is in the same status (0 or 1) it was at the start. */
			/* vol[i] is also incremented by Period[i], since the wave has been 1 */
			/* exactly half of the time, regardless of the initial position. */
			/* If we exit the loop in the middle, Output[i] has to be inverted */
			/* and vol[i] incremented only if the exit status of the square */
			/* wave is 1. */
			while (R->Count[i] <= 0)
			{
				R->Count[i] += R->Period[i];
				if (R->Count[i] > 0)
				{
					R->Output[i] ^= 1;
					if (R->Output[i]) vol[i] += R->Period[i];
					break;
				}
				R->Count[i] += R->Period[i];
				vol[i] += R->Period[i];
			}
			if (R->Output[i]) vol[i] -= R->Count[i];
		}

		left = STEP;
		do
		{
			int nextevent;


			if (R->Count[3] < left) nextevent = R->Count[3];
			else nextevent = left;

			if (R->Output[3]) vol[3] += R->Count[3];
			R->Count[3] -= nextevent;
			if (R->Count[3] <= 0)
			{
				if (R->RNG & 1) R->RNG ^= R->NoiseFB;
				R->RNG >>= 1;
				R->Output[3] = R->RNG & 1;
				R->Count[3] += R->Period[3];
				if (R->Output[3]) vol[3] += R->Period[3];
			}
			if (R->Output[3]) vol[3] -= R->Count[3];

			left -= nextevent;
		} while (left > 0);

		out = vol[0] * R->Volume[0] + vol[1] * R->Volume[1] +
				vol[2] * R->Volume[2] + vol[3] * R->Volume[3];

		if (out > MAX_OUTPUT * STEP) out = MAX_OUTPUT * STEP;

		*(buf++) = out / STEP;

		length--;
	}
}
