/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

#include <stdio.h>
#include <string.h>

#include "general.h"
#include "driver.h"
#include "resource.h"
#include "object.h"
#include "event.h"
#include "vm.h"

void to_equaln(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	if(VAR[arg0] == arg1) 
		vm_push(1);
	else	vm_push(0);
}

void to_equalv(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	if(VAR[arg0] == VAR[arg1]) 
		vm_push(1);
	else	vm_push(0);
}

void to_lessn(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	if(VAR[arg0] < arg1) 
		vm_push(1);
	else	vm_push(0);
}

void to_lessv(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	if(VAR[arg0] < VAR[arg1]) 
		vm_push(1);
	else	vm_push(0);
}

void to_greatern(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	if(VAR[arg0] > arg1) 
		vm_push(1);
	else	vm_push(0);
}

void to_greaterv(vm_t *l)
{
	UINT8 arg0, arg1;

	arg1 = vm_pop();
	arg0 = vm_pop();

	if(VAR[arg0] > VAR[arg1]) 
		vm_push(1);
	else	vm_push(0);
}

void to_isset(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();

	if(FLAG[arg]) 
		vm_push(1);
	else	vm_push(0);
}

void to_issetv(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();

	if(FLAG[VAR[arg]]) 
		vm_push(1);
	else	vm_push(0);
}

void to_has(vm_t *l)
{
	UINT8 arg;
	
	arg = vm_pop();
	
	if(inv_item[arg].room == 255) vm_push(1);
	else vm_push(0);
	
}

void to_obj_in_room(vm_t *l)
{
	UINT8 arg0, arg1;
	
	arg1 = vm_pop();
	arg0 = vm_pop();

	if(inv_item[arg0].room == VAR[arg1]) vm_push(1);
	else vm_push(0);

}

void to_posn(vm_t *l)
{
	UINT8 arg0, arg1, arg2, arg3, arg4;
	view_obj *o;
	
	arg4 = vm_pop();
	arg3 = vm_pop();	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	o = &object[arg0];

	if((arg1 <= o->x) &&
	   (arg2 <= o->y) &&
	   (o->x <= arg3) &&
	   (o->y <= arg4)) 
		vm_push(1);
	else
		vm_push(0);
	 
}

void to_controller(vm_t *l)
{
	UINT8 arg;

	arg = vm_pop();

	if(event_check(arg)) 
		vm_push(1);
	else	vm_push(0);
}

void to_havekey(vm_t *l)
{
	int i;
	boolean ret = false;
	
	driver_kb_update();
	
	for(i = 0; i < 128; i++) {
		if((i == AGIL_KEY_ALT) || /* ignore modifier keys */
		   (i == AGIL_KEY_CONTROL) ||
		   (i == AGIL_KEY_LEFTSHIFT) || (i == AGIL_KEY_RIGHTSHIFT)) 
		   	continue;

		if((event_checkkey_noexec(i)) ||	/* ignore special keys */
		   (event_checkkey_noexec(i << 8))) continue;
		
		if(keytable[i]) ret = true;
		keytable[i] = 0; /* flush keytable */
	}
	
	vm_push(ret);
}

void to_compare_strings(vm_t *l)
{
	UINT8 arg0, arg1;
	boolean ret;
	
	arg1 = vm_pop();
	arg0 = vm_pop();
	
	ret = !strcmp(STRING[arg0], STRING[arg1]);
	
	vm_push(ret);
}

void to_obj_in_box(vm_t *l)
{
	UINT8 arg0, arg1, arg2, arg3, arg4;
	view_obj *o;
	
	arg4 = vm_pop();
	arg3 = vm_pop();	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	o = &object[arg0];

	if((arg1 <= o->x) &&
	   (arg2 <= o->y) &&
	   (o->x + o->width - 1 <= arg3) &&
	   (o->y <= arg4)) 
		vm_push(1);
	else
		vm_push(0);
	 
}

void to_center_posn(vm_t *l)
{
	UINT8 arg0, arg1, arg2, arg3, arg4;
	view_obj *o;
	
	arg4 = vm_pop();
	arg3 = vm_pop();	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	o = &object[arg0];

	if((arg1 <= o->x + (o->width / 2)) &&
	   (arg2 <= o->y) &&
	   (o->x + (o->width / 2) <= arg3) &&
	   (o->y <= arg4)) 
		vm_push(1);
	else
		vm_push(0);
	 
}

void to_right_posn(vm_t *l)
{
	UINT8 arg0, arg1, arg2, arg3, arg4;
	view_obj *o;
	
	arg4 = vm_pop();
	arg3 = vm_pop();	
	arg2 = vm_pop();
	arg1 = vm_pop();
	arg0 = vm_pop();

	o = &object[arg0];

	if((arg1 <= o->x + o->width - 1) &&
	   (arg2 <= o->y) &&
	   (o->x + o->width - 1 <= arg3) &&
	   (o->y <= arg4)) 
		vm_push(1);
	else
		vm_push(0);
	 
}


struct opcode_s test_opcode[0x13] = {
	{"condition00",		0, /* to fill the space in the array */ NULL},
	{"equaln", 		2, to_equaln},
	{"equalv", 		2, to_equalv},
	{"lessn", 		2, to_lessn},
	{"lessv", 		2, to_lessv},
	{"greatern", 		2, to_greatern},
	{"greaterv", 		2, to_greaterv},
	{"isset", 		1, to_isset},
	{"issetv", 		1, to_issetv},
	{"has", 		1, to_has},
	{"obj.in.room", 	2, to_obj_in_room},
	{"posn", 		5, to_posn},
	{"controller", 		1, to_controller},
	{"have.key", 		0, to_havekey},
	{"said", 		0, /* Handler is in vm.c */ NULL},
	{"compare.strings", 	2, to_compare_strings},
	{"obj.in.box", 		5, to_obj_in_box},
	{"center.posn", 	5, to_center_posn},
	{"right.posn", 		5, to_right_posn} 
};  
