/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "general.h"
#include "event.h"
#include "driver.h"
#include "graphics.h"
#include "text.h"
#include "status.h"
#include "vm.h"

struct word {
	char word[40]; /* must be big enough for all possible words */
	UINT16 id;
	
	struct word * next;
};

struct word * dic[26]; /* dictionaries A to Z */
UINT16 parsed[10]; /* the parsed input string (as examined by said() */
char parsedwords[10][40];

char oldinputstring[39] = "";
char inputstring[39] = "";
int input_pos = 0;

boolean text_input_entered = false;
boolean text_special_input = false;

void text_init(void)
{
	int i;

	for(i = 0; i < 26; i++)
		dic[i] = NULL;

	parsed[0] = 0;
	
	atexit(text_flushwords);
	printf("text: Text handler initialized\n");
}


void text_flushwords(void)
{
	int i;
	struct word * current;
	
	for(i = 0; i < 26; i++)
		while(dic[i]) {
			current = dic[i]->next;
			free(dic[i]);
			dic[i] = current;
		}
		
	printf("text: Words list flushed\n");
}


void text_addword(char * word, UINT16 wordid)
{
	int d;
	struct word * current;
	
	d = word[0] - 'a'; /* which list to use */
	
	if(!dic[d]) { /* dictionary empty */
		dic[d] = (struct word *)malloc(sizeof(struct word));
		dic[d]->next = NULL;
	} else { /* add new words to beginning of list */
		current = dic[d];
		dic[d] = (struct word *)malloc(sizeof(struct word));
		dic[d]->next = current;
	}
	strncpy(dic[d]->word, word, 39);
	dic[d]->id = wordid;
}


struct word * text_checkword(char * word) /* return NULL on unknown word */
{
	int d, i;
	struct word * current;
	
	d = word[0] - 'a';
	if((d < 0) || (d > 25)) return NULL; /* illegal dictionary */
	
	for(current = dic[d]; current != NULL; current = current->next) {
		i = 0;
		while((current->word[i] == word[i]) && word[i]) i++;
		if(current->word[i] == 0) {
			if((word[i] == 0) || (word[i] == ' ')) {
				return current;
			}
		} 
	}
	
	return NULL;
}


void text_cleanstring(char * str)
{
	int i;

	for(i = 0; str[i]; i++) {
		while((str[i] == ' ') && ((str[i + 1] == ' ') || (str[i + 1] == 0)))
			strcpy(&str[i], &str[i + 1]); /* remove whitespace */

		if((str[i] >= 'A') && (str[i] <= 'Z')) /* make lowercase */
			str[i] += 32;
			
	}
	
	while(str[0] == ' ') strcpy(str, &str[1]); /* remove prefixed w/s */
}


void text_parse(char * input)
{
	char * localcopy;
	char * tmp;
	struct word * match;
	int i = 0;
	
	if(!input[0]) {
		parsed[0] = 0;
		parsedwords[0][0] = 0;
		return;
	}
	
	localcopy = (char *)alloca(strlen(input) + 1);
	strcpy(localcopy, input);
	text_cleanstring(localcopy);
	
	while(localcopy[0]) {
		match = text_checkword(localcopy);
		if(match) {
			if(match->id) {
				parsed[i] = match->id;
				strncpy(parsedwords[i], match->word, 39);
			} else	i--; /* ignore wordid 0 */
			
			localcopy += strlen(match->word);
			if(localcopy[0] == ' ') localcopy++;
		} else {
			parsed[i] = 1; /* Anyword */
			
			if(!VAR[9]) VAR[9] = i + 1;
			
			tmp = strchr(localcopy, ' ');
			strncpy(parsedwords[i], localcopy, (int)(tmp?tmp - localcopy + 1:39));
			localcopy = tmp;
			if(localcopy == NULL) {
				i++; 
				break;
			} else {
				tmp = strchr(parsedwords[i], ' ');
				tmp[0] = 0; /* terminate the monsters */
			}
			localcopy++;
		}
		i++;
	}
	
	parsed[i] = 0; /* terminate parsed string */
	parsedwords[i][0] = 0;

	if(i != 0) {
		FLAG[2] = true;
		FLAG[4] = false;
	}
}


void text_echoline(void)
{
	if(strlen(oldinputstring) <= strlen(inputstring)) return;
	
	strcpy(&inputstring[input_pos], &oldinputstring[input_pos]);
	input_pos = strlen(inputstring);
}


void text_new_key(int code)
{
	char inputline[41];
	int i;
		
	if(!(status.accept_input || text_special_input)) return;
	
	if(code == '\b') {
		if(input_pos == 0) return;
		input_pos--;
		inputstring[input_pos] = 0;
	}
	
	if((code >= 32) && (code < 127)) {
		if(input_pos == 38) return;
		inputstring[input_pos] = code;
		input_pos++;
		inputstring[input_pos] = 0;
	}
	
	if((code == 0x0D) || (code == '\n')) {
		for(i = 0; i < 128; i++) { 
			if((i == AGIL_KEY_ALT) ||
			   (i == AGIL_KEY_CONTROL) ||
			   (i == AGIL_KEY_LEFTSHIFT) || (i == AGIL_KEY_RIGHTSHIFT)) 
				continue;

			keytable[i] = 0;
		}
		
		if(text_special_input) {
			text_input_entered = true;
		} else {
			text_parse(inputstring);

			if(parsed[0]) 
				strcpy(oldinputstring, inputstring);
			
			input_pos = 0;
			inputstring[input_pos] = 0;
		}
	}

	if(!text_special_input) {
		snprintf(inputline, 41, "%s%s%c ", STRING[0], inputstring, status.cursor_char);
		memset(video_buffer + (176 * 320), 0, 8 * 320);
		graph_drawstring(video_buffer, 0, 176, inputline, 15, 0);
	}
}


void text_input(char * prompt, int x, int y, int l, char * def)
{
	char inputline[41];
	int oldpos = 0;
	
	if(!status.accept_input && !status.dialogue_open) return;
	text_special_input = true; text_input_entered = false;
/*	input_pos = 0; inputstring[input_pos] = 0;*/
	strncpy(inputstring, def, l);
	input_pos = strlen(inputstring);
	
	snprintf(inputline, 40, "%s%s%c ", prompt, inputstring, status.cursor_char);
	graph_drawstring(video_buffer, x * 8, y * 8, inputline, status.fgcolor, status.bgcolor);
	driver_showbuffer(video_buffer);
	
	while(!text_input_entered) {
		driver_kb_update();
		
		if(oldpos == input_pos) continue;
		else oldpos = input_pos;
			
		if(input_pos > l) {
			input_pos = l;
			inputstring[input_pos] = 0;
		}

		snprintf(inputline, 40, "%s%s%c ", prompt, inputstring, status.cursor_char);
		graph_drawstring(video_buffer, x * 8, y * 8, inputline, status.fgcolor, status.bgcolor);
		driver_showbuffer(video_buffer);
	}

	graph_drawstring(video_buffer, x * 8, y * 8, inputline, status.bgcolor, status.bgcolor);
	text_special_input = false;
}
