/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "general.h"
#include "graphics.h"
#include "resource.h"
#include "object.h"
#include "view.h"
#include "event.h"
#include "driver.h"
#include "message.h"
#include "text.h"
#include "vm.h"

void view_recalc(view_obj *o)
{
	if(!VIEW[o->view]) {
		printf("object: FATAL: view object assigned nonexistant view\n");
		exit(-1);
	}
	
	if(!VIEW[o->view]->data) {
		printf("object: WARNING: view object assigned unloaded view\n");
		res_load(VIEW[o->view]);
	}
	
	/* Get VIEW info */
	o->noloops = VIEW[o->view]->data[2];
	if(o->loop >= o->noloops) o->loop = 0;
	
	/* Get LOOP info */
	o->loop_ofs = ((UINT16)VIEW[o->view]->data[6 + (o->loop << 1)] << 8) | VIEW[o->view]->data[5 + (o->loop << 1)];
	o->nocels = VIEW[o->view]->data[o->loop_ofs];
	if(o->cel >= o->nocels) o->cel = 0;
	
	/* Get CEL info */
	o->cel_ofs = o->loop_ofs + (((UINT16)VIEW[o->view]->data[o->loop_ofs + 2 + (o->cel << 1)] << 8) | VIEW[o->view]->data[o->loop_ofs + 1 + (o->cel << 1)]);
	o->width = VIEW[o->view]->data[o->cel_ofs];
	o->height = VIEW[o->view]->data[o->cel_ofs + 1];
	o->mirror = (VIEW[o->view]->data[o->cel_ofs + 2] >> 7) && (((VIEW[o->view]->data[o->cel_ofs + 2] >> 4) & 7) != o->loop);
	o->trans = VIEW[o->view]->data[o->cel_ofs + 2] & 0x0F;
}


boolean view_draw(view_obj *o)
{
	boolean ret = false;
	UINT8 * v;
	UINT8 * p;
	int X, Y, x = 0, y = 0;
	int i, j;
	
	if(!VIEW[o->view]) return false;

	if(!(o->flags & VO_FIX_PRI)) object_setdefaultpri(o);
	if(!(o->flags & VO_FIX_LOOP)) object_setdefaultloop(o);

/*	if(o == &object[0]) o->priority = 15;*/

	o->flags |= VO_ON_SCREEN;
	o->oldx = o->x; o->oldy = o->y; o->oldw = o->width; o->oldh = o->height;
	
	v = workvisbuffer + 2560;
	p = workpribuffer + 2560;
	X = o->x;
	Y = o->y - o->height + 1;
	if(Y < 0) Y = 0;

	if(o->mirror) x = o->width - 1;
	
	for(i = o->cel_ofs + 3; y < o->height; i++) {
		if(VIEW[o->view]->data[i] == 0) {
			if(o->mirror) x = o->width - 1;
			else x = 0;
			
			y++;
			continue;
		}
		for(j = 0; j < (VIEW[o->view]->data[i] & 0x0F); j++) {
			if((VIEW[o->view]->data[i] >> 4) != o->trans) {
				if(graph_get_real_pri(X + x, Y + y) <= o->priority){
					pixel(v, ((X + x) * 2), (Y + y)) = VIEW[o->view]->data[i] >> 4;
					pixel(v, (((X + x) * 2) + 1), (Y + y)) = VIEW[o->view]->data[i] >> 4;
					if(pixel(p, ((X + x) * 2), (Y + y)) >= 3) {
						pixel(p, ((X + x) * 2), (Y + y)) = o->priority;
						pixel(p, (((X + x) * 2) + 1), (Y + y)) = o->priority;
					}
					ret = true;
				}
			}
			if(o->mirror) x--; else x++;
		}
	}
	return ret;
}


void view_erase(view_obj *o)
{
	UINT8 * v;
	UINT8 * p;
	UINT8 X, Y, x = 0, y = 0;

	v = workvisbuffer + 2560;
	p = workpribuffer + 2560;
	
	o->flags &= ~VO_ON_SCREEN;
	
	X = o->oldx;
	Y = o->oldy - o->oldh;
	if(Y > 168) Y = 0;

	for(y = 0; y <= o->oldh; y++) {
		for(x = 0; x < o->oldw; x++) {
			if(pixel(p, ((X + x) * 2), (Y + y)) > 15) continue;
			
			pixel(v, ((X + x) * 2), (Y + y)) = 
				pixel(backvisbuffer, ((X + x) * 2), (Y + y));
			pixel(v, (((X + x) * 2) + 1), (Y + y)) =
				pixel(backvisbuffer, (((X + x) * 2) + 1), (Y + y));

			pixel(p, ((X + x) * 2), (Y + y)) = 
				pixel(backpribuffer, ((X + x) * 2), (Y + y));
			pixel(p, (((X + x) * 2) + 1), (Y + y)) = 
				pixel(backpribuffer, (((X + x) * 2) + 1), (Y + y));
		}
	}
}


void view_add_to_pic(int view, int loop, int cel, int x, int y, int pri, int m)
{
	view_obj tmp;
	UINT8 *realvis, *realpri;
	UINT8 startpri;
	int i;
	
	tmp.view = view;
	tmp.loop = loop;
	tmp.cel = cel;
	tmp.x = x;
	tmp.y = y;
	tmp.priority = pri;
	tmp.flags = (VO_FIX_PRI | VO_FIX_LOOP);
	if(tmp.priority < 4) tmp.flags &= ~VO_FIX_PRI;
	
	view_recalc(&tmp);
	
	realvis = workvisbuffer; workvisbuffer = backvisbuffer - 2560;
	realpri = workpribuffer; workpribuffer = backpribuffer - 2560;

	view_draw(&tmp);
	
	workvisbuffer = realvis;
	workpribuffer = realpri;

	view_draw(&tmp);	

	/* Yuckk! This is some old dirty code by me. Must fix. */
	if(m < 3) {
		tmp.height--;
		tmp.width--;
		
		for(i = tmp.x; i <= tmp.x + tmp.width; i++) {
			pixel(backpribuffer, (i * 2), tmp.y) = m;
			pixel(backpribuffer, ((i * 2) + 1), tmp.y) = m;
		}
		for(startpri = floor((float)(tmp.y) / 12) + 1; 
			(startpri == floor((float)(tmp.y - 1) / 12) + 1) && (tmp.y > (y - tmp.height));
			tmp.y--) {

			pixel(backpribuffer, (tmp.x * 2), tmp.y) = m;
			pixel(backpribuffer, ((tmp.x * 2) + 1), tmp.y) = m;
			pixel(backpribuffer, ((tmp.x + tmp.width) * 2), tmp.y) = m;
			pixel(backpribuffer, (((tmp.x + tmp.width) * 2) + 1), tmp.y) = m;
		}

		for(i = tmp.x; i <= tmp.x + tmp.width; i++) {
			pixel(backpribuffer, (i * 2), tmp.y) = m;
			pixel(backpribuffer, ((i * 2) + 1), tmp.y) = m;
		}
	}
}


void view_show_obj(int view)
{
	view_obj tmp;
	UINT16 discoffset;
	char * disc;
	boolean weloadedview = false;
	char * view_parsedisc(char * message);
	
	tmp.view = view;
	tmp.loop = 0;
	tmp.cel = 0;
	tmp.priority = 15;
	tmp.flags = (VO_FIX_PRI | VO_FIX_LOOP);
	tmp.y = 168;

	if(!VIEW[view]->data) {
		weloadedview = true;
		res_load(VIEW[view]);
	}
	view_recalc(&tmp);

	tmp.x = 80 - (tmp.width / 2);

	discoffset = ((UINT16)VIEW[view]->data[4] << 8) | VIEW[view]->data[3];
	disc = (char *)VIEW[view]->data + discoffset;
	
	memcpy(textmodebuffer, video_buffer, 64000);	
	view_draw(&tmp);
	message_mbox(view_parsedisc(disc), 0, 0, 0);
	driver_showbuffer(video_buffer);
	while(event_waitforkey() != 0x0D);
	message_erasebox();
	memcpy(video_buffer, textmodebuffer, 64000);	

	if(weloadedview) res_unload(VIEW[view]);
}


char * view_parsedisc(char * message)
{
	static char ret[1000];
	char workret[1000];
	int SI = 0, DI = 0;
	int tmp, nowidth;

	while(message[SI]) {
		if(message[SI] != '%') {
			if(message[SI] == '\\') SI++;
			workret[DI] = message[SI];
			SI++;
			DI++;
		} else {
			SI++;
			switch(message[SI]) {
				case 'v':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					if(message[SI] == '|') {
						SI++;
						sscanf(&message[SI], "%d", &nowidth);
						SI += sprintf(ret, "%d", nowidth);
						switch(nowidth) {
							case 1:
								DI += sprintf(&workret[DI], "%01d", VAR[tmp]);
								break;
							case 2:
								DI += sprintf(&workret[DI], "%02d", VAR[tmp]);
								break;
							case 3:
								DI += sprintf(&workret[DI], "%03d", VAR[tmp]);
								break;
							default:
								DI += sprintf(&workret[DI], "%d", VAR[tmp]);
						}
					} else
						DI += sprintf(&workret[DI], "%d", VAR[tmp]);
					break;
				case '0':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					DI += sprintf(&workret[DI], "%s", inv_item[tmp].name);
					break;
				case 'w':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					DI += sprintf(&workret[DI], "%s", parsedwords[tmp - 1]);
					break;
				case 's':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					DI += sprintf(&workret[DI], "%s", STRING[tmp]);
					break;
				case 'g':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					DI += sprintf(&workret[DI], "%s", vm_parsemessage(LOGIC[0], tmp));
					break;
			}
		}
	}
	workret[DI] = 0;
	
	strcpy(ret, workret);

	return ret;
}
