/*
 * AGIL: Adventure Game Interpreter for Linux, (C) 2001, 2002 RaDIaT1oN
 * 	See the file COPYING
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "general.h"
#include "text.h"
#include "gameid.h"
#include "vm.h"
#include "main.h"

UINT8 vm_stack[100];
UINT8 vm_SP = 0;

boolean new_room = false;
boolean vm_reset = false;

UINT8 VAR[256];
volatile boolean FLAG[256];
char STRING[24][40];

void vm_ifhandler(vm_t * c); /* private */


void vm_init(void)
{
	int i;
	for(i = 0; i < 256; i++) {
		VAR[i] = 0;
		FLAG[i] = false;
	}
	
/*	memset(STRING, 0, 24 * 40);*/
	
	/* Initial values */
	FLAG[5] = true;		/* new_room */
	FLAG[11] = true;	/* LOGIC.0 run for first time */
	FLAG[9] = true;		/* sound_on */
	VAR[8] = 255;		/* Free memory */
	VAR[10] = 2;		/* Normal speed */
	VAR[22] = 1;		/* Sound: PC */
	VAR[24] = 0x29;		/* Input line length */
	VAR[26] = 3;		/* EGA screen */
}


void vm_exec(UINT8 logicno) /* Execute LOGIC[logicno] */
{
	vm_t c;
	int i = 0;
	
	c.res = LOGIC[logicno];
	if(!c.res->data) {
		printf("vm: WARNING: attempt to exec unloaded logic\n");
		return;
	}

	c.IP = c.res->ssIP;
	c.code = c.res->data + 2;
	vm_reset = false;

	while(c.code[c.IP] != 0) { 		/* not return() */


		if(c.code[c.IP] == 0xFF) {	/* if structure */
			vm_ifhandler(&c);
			continue;
		}
		
		if(c.code[c.IP] == 0xFE) { 	/* JMP */
			UINT16 jumpsize;
/*			printf("JMP\n");*/
			jumpsize = ((UINT16)c.code[c.IP + 2] << 8) + c.code[c.IP + 1] + 3;
			c.IP += jumpsize;
			if(jumpsize >> 15) update_clock();
			continue;
		}

/*		printf(" %s(", action_opcode[c.code[c.IP]].name); */

		if(action_opcode[c.code[c.IP]].execute) { /* Normal opcodes */
			/* Push arguments */
			for(i = 1; i <= action_opcode[c.code[c.IP]].args; i++) {
/*				printf("%d, ", c.code[c.IP + i]);*/
				vm_push(c.code[c.IP + i]);
			}
/*			printf("\b\b");*/
			
			/* execute opcode */
			action_opcode[c.code[c.IP]].execute(&c); /* execute */		
		}
		if(vm_SP) {
/*			printf("vm: FATAL: stack corrupted\n");*/
			exit(-1);
		}
/*		printf(");\n");*/
		if(vm_reset) return;
		/* update IP */
		c.IP += action_opcode[c.code[c.IP]].args + 1;
	}

	if(logicno == 0) { /* If we clean exit from LOGIC.0 */
		FLAG[6] = false;
		FLAG[12] = false;
	}
}



void vm_ifhandler(vm_t * c) /* Should only be called from vm_exec() */
{
	int skip, i;
	boolean or = false, not = false, nor = false, ret, cond = true, orcond = false;
	boolean said;
		
/*	printf("if\n");*/
	for(c->IP++; c->code[c->IP] != 0xFF; c->IP++) {
		said = false;
		
		if(c->code[c->IP] == 0xFC) {	/* OR handler */
/*			printf("OR\n");*/
			if((or = !or)) {
				orcond = false;
				if(not) {
					nor = true;
					not = false;
				} else nor = false;
			} else {
				if(nor) orcond = !orcond;
				cond = cond && orcond;
			}
			continue;
		}
		
		if(c->code[c->IP] == 0xFD) {	/* NOT handler */
/*			printf("NOT\n");*/
			not = true;
			continue;
		}
		
/*		printf("   %s(", test_opcode[c->code[c->IP]].name);*/
		
		if(test_opcode[c->code[c->IP]].execute) { /* excute opcode */
			for(i = 1; i <= test_opcode[c->code[c->IP]].args; i++) {
/*				printf("%d, ", c->code[c->IP + i]);*/
				vm_push(c->code[c->IP + i]);
			}
/*			printf("\b\b");*/

			test_opcode[c->code[c->IP]].execute(c);
			ret = vm_pop();	/* get return value */
		} else {
			ret = false;
			if(c->code[c->IP] != 0x0E) not = false;
		}
/*		printf(");\n");*/
		
		if(c->code[c->IP] == 0x0E) {	/* said() handler */
			UINT16 saidarg;
			boolean match = false;

			said = true;
			c->IP++;
			if(FLAG[2] && !FLAG[4])	
				ret = true;
			else
				ret = false;
				
			for(i = 0; i < c->code[c->IP]; i++) {
				saidarg = ((UINT16)c->code[c->IP + (i * 2) + 2] << 8) + c->code[c->IP + (i * 2) + 1];
				if(saidarg == 1) continue;
				if(saidarg == 9999) match = true;
				if(saidarg != parsed[i] && !match) ret = false;
				if((parsed[i] == 0) && !match) ret = false;
			}
			if((parsed[i] != 0) && !match) ret = false;

			c->IP += i << 1;
			if(ret && cond) FLAG[4] = true;
		}

		if(not) {
			ret = !ret;
			not = false;
		}
		 
		if(or) { 
			orcond = orcond || ret;
		} else {
			cond = cond && ret;
		}

		if(!said)
			c->IP += test_opcode[c->code[c->IP]].args; /* update IP */
	}

/*	printf("END IF\n"); */
	c->IP++;
	skip = ((UINT16)c->code[c->IP + 1] << 8) | (c->code[c->IP]);
	c->IP += 2;

	if(!cond) c->IP += skip;	/* if condition not met */
}


void vm_push(UINT8 data)
{
	vm_stack[vm_SP] = data;
	vm_SP++;

	if(vm_SP == 100) {
		printf("vm: FATAL: stack overflow\n");
		exit(-1);
	}
}


UINT8 vm_pop(void)
{
	vm_SP--;
	return vm_stack[vm_SP];
}


/* This whole function is a shit one, having to return static char ret */
/* I will rewrite when I have a better idea */
char * vm_parsemessage(resource_t *l, UINT8 messageno)
{
	char * message;
	static char ret[1000];
	char workret[1000];
	int SI = 0, DI = 0;
	int tmp;
	int nowidth;

	messageno--;	
	message = l->message[messageno];
	
	while(message[SI]) {
		if(message[SI] != '%') {
			if(message[SI] == '\\') SI++;
			workret[DI] = message[SI];
			SI++;
			DI++;
		} else {
			SI++;
			switch(message[SI]) {
				case 'v':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					if(message[SI] == '|') {
						SI++;
						sscanf(&message[SI], "%d", &nowidth);
						SI += sprintf(ret, "%d", nowidth);
						switch(nowidth) {
							case 1:
								DI += sprintf(&workret[DI], "%01d", VAR[tmp]);
								break;
							case 2:
								DI += sprintf(&workret[DI], "%02d", VAR[tmp]);
								break;
							case 3:
								DI += sprintf(&workret[DI], "%03d", VAR[tmp]);
								break;
							default:
								DI += sprintf(&workret[DI], "%d", VAR[tmp]);
						}
					} else
						DI += sprintf(&workret[DI], "%d", VAR[tmp]);
					break;
				case 'm':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					DI += sprintf(&workret[DI], "%s", vm_parsemessage(l, tmp));
					break;
				case '0':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					DI += sprintf(&workret[DI], "%s", inv_item[tmp].name);
					break;
				case 'w':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					DI += sprintf(&workret[DI], "%s", parsedwords[tmp - 1]);
					break;
				case 's':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					DI += sprintf(&workret[DI], "%s", STRING[tmp]);
					break;
				case 'g':
					SI++;
					sscanf(&message[SI], "%d", &tmp);
					SI += sprintf(ret, "%d", tmp);
					DI += sprintf(&workret[DI], "%s", vm_parsemessage(LOGIC[0], tmp));
					break;
			}
		}
	}
	workret[DI] = 0;
	
	strcpy(ret, workret);

	return ret;
}


void vm_fix_opcode_table(void) /* from LAGII */
{

	/* 'quit' takes 0 args for 2.089 */
	if(strcmp(id->agiver, "2.089") == 0) {
		printf("vm: Fixing quit() for AGI 2.089\n");
		action_opcode[0x86].args = 0;
	}

	/* 'print.at' and 'print.at.v' take 3 args before 2.400 */

	/* THIS CHUNK ACCORDING TO AGI SPEC */
	/* CAUSES SPACE QUEST 1 (1.1A) (AGI 2.272) TO BREAK AT THE SKIMMER */
	/* SEQUENCE.  (THE FIRST PRINT.AT COMMAND).  CAN FINNISH GAME */
	/* IF TAKES FOUR ARGS */
	
/*	if(strcmp(id->agiver, "2.400") < 0) {
		printf("vm: Fixing print.at() and print.at.v() for AGI %s\n", id->agiver);
		action_opcode[0x97].args = 3;
		action_opcode[0x98].args = 3;
	}*/

	/* 'unknown176' takes 1 arg for 3.002.086 */
/*	if ((ver[0] == 3) && (ver[1] == 2) && (ver[2] == 86)) {*/
	if(strcmp(id->agiver, "3.002.086") == 0) {
		printf("vm: Fixing unknown176() for AGI 3.002.086\n");
		action_opcode[0xB0].args = 1;
	}
}
